#############################################################################
#                                                                           #
# MakeFile for BTXE/DOS by tom schlangen                        TS 961013   #
# Based on the original Bink/DOS makefile                                   #
# Suited for compilers MSC V7.0, MSC V8.00 (aka Visual C/C++), and the      #
# assemblers MS MASM 5.1 or Borland TASM 1.01                               #
#                                                                           #
#############################################################################
#                                                                           #
# Usage: nmake -f BTXEC700.MAK ["CPU=<string>"] [optional target]           #
#                                                                           #
# If no target is specified, the default will generate a non-overlayed,     #
# (or `BTBIG', if you prefer) DOS executable, patched with the english      #
# language messages and optimized for no specific processor.                #
#                                                                           #
# Defining "CPU=<string>" (<string> may be `86', `186', `286', `386')       #
# on the nmake command line will generate code optimized for the            #
# correspondingly name processor.                                           #
#                                                                           #
#############################################################################

#############################################################################
#                                                                           #
# Definitions                                                               #
#                                                                           #
#############################################################################
PROJ = BINKLEY
PROJFILE = BINKLEY.MAK

SRC_DIR  = ..\src
INCL_DIR = ..\include

#############################################################################
#                                                                           #
# Define `masm' or `tasm' according to the assembler you use                #
#                                                                           #
#############################################################################
ASM = tasm
CC  = cl

LINKER = link
LRF  = echo > NUL

#############################################################################
#                                                                           #
# Recommended optimizing command line switches:                             #
#   for MSC 4.0, instead of -Oslern, use -Ot                                #
#   for MSC 5.0,                     use -Os                                #
#   for MSC 5.1,                     use -Oxrn                              #
#   for MSC 6.0a,                    use -Olerg                             #
#   for MSC 7.0a,                    use -Olerg                             #
#   for MSC 8.0,                     use -Olerg                             #
#                                                                           #
#############################################################################

!IF "$(DEBUG)" != ""
cdebug = -Gt -Zi -Od
!ELSE
cdebug = -Gs -Gt32 -Olerg
#cdebug = -Gs -Gt -Otlergi
!ENDIF

!IF "$(WARNINGS)" != ""
cwarn = -W4
!ELSE
cwarn = -W3
!ENDIF

!IF "$(CPU)" == "86"
coptf= -G0
cpuid= -D_CPU=86
!ENDIF

!IF "$(CPU)" == "186"
coptf= -G1
cpuid= -D_CPU=186
!ENDIF

!IF "$(CPU)" == "286"
coptf= -G2
cpuid= -D_CPU=286
!ENDIF

!IF "$(CPU)" == "386"
coptf= -G3
cpuid= -D_CPU=386
!ENDIF

ccflgs = -c -AL  
cclcl = -Zp -I $(INCL_DIR)

CFLAGS = $(ccflgs) $(coptf) $(cdebug) $(cwarn) $(cpuid) $(cclcl)

#############################################################################
#                                                                           #
# Objects needed by the mailer executable                                   #
#                                                                           #
#############################################################################
COBJS =   bt.obj aglcrc.obj   b_banner.obj b_bbs.obj    b_faxrcv.obj \
    b_flofil.obj b_frproc.obj b_fuser.obj  b_help.obj   b_initva.obj \
    b_logs.obj   b_passwo.obj b_proto.obj  b_rspfil.obj b_sbinit.obj \
    b_script.obj b_search.obj b_sessio.obj b_term.obj   b_termov.obj \
    b_today.obj  b_wazoo.obj  b_whack.obj  b_wzsend.obj brec.obj     \
    bsend.obj    btconfig.obj chat.obj     cache.obj    data.obj     \
    dosfuncs.obj download.obj emsi.obj     evtparse.obj file_dos.obj \
    fossil.obj   freepoll.obj ftsc.obj     hydra.obj    janus.obj    \
    keymap.obj   langload.obj m7rec.obj    m7send.obj   mailer.obj   \
    mailovly.obj mailroot.obj mdm_proc.obj misc.obj     nodeproc.obj \
    outbound.obj prometer.obj recvbark.obj recvsync.obj sbuf.obj     \
    sched.obj    sendbark.obj sendsync.obj squish.obj   srif.obj     \
    statetbl.obj timer.obj    upload.obj   version.obj  version7.obj \
    vfossil.obj  xmrec.obj    xmsend.obj   yoohoo.obj   zmisc.obj    \
    zreceive.obj zsend.obj    heap.obj     maxmcp.obj   css.obj
   
NOOVLOBJS = b_spwnno.obj 
OVLOBJS = b_spwnov.obj

#############################################################################
#                                                                           #
# Objects common/shared between several executables                         #
#                                                                           #
#############################################################################
COBJCOMMON = common.obj

#############################################################################
#                                                                           #
# Objects needed by the BTUTIL executable                                   #
#                                                                           #
#############################################################################
COBJ2 = btutil.obj

#############################################################################
#                                                                           #
# Objects assembled by the MASM or TASM assemblers                          #
#                                                                           #
#############################################################################
MOBJS = spawn.obj bink_asm.obj

#############################################################################
#                                                                           #
# Includes                                                                  #
#                                                                           #
#############################################################################
INCL_FILES = $(INCL_DIR)\includes.h $(INCL_DIR)\bink.h                   \
             $(INCL_DIR)\defines.h $(INCL_DIR)\com.h                     \
             $(INCL_DIR)\com_dos.h $(INCL_DIR)\xfer.h                    \
             $(INCL_DIR)\msgs.h $(INCL_DIR)\zmodem.h $(INCL_DIR)\keybd.h \
             $(INCL_DIR)\keyfncs.h $(INCL_DIR)\sbuf.h                    \
             $(INCL_DIR)\sched.h $(INCL_DIR)\timer.h                     \
             $(INCL_DIR)\vfossil.h $(INCL_DIR)\externs.h                 \
             $(INCL_DIR)\prototyp.h $(INCL_DIR)\squish.h

#############################################################################
#                                                                           #
# Dependency lists                                                          #
#                                                                           #
#############################################################################
all:    btbig.exe btutil.exe english.lng german.lng startrek.lng patchbe

btbig:  btbig.exe btutil.exe english.lng german.lng startrek.lng patchbe

bt:     bt.exe btutil.exe english.lng german.lng startrek.lng patche

#   generate language file from source

english.lng:    $(SRC_DIR)\language.txt btutil.exe
    btutil LNG key E $(SRC_DIR)\language.txt english.lng

german.lng:     $(SRC_DIR)\language.txt btutil.exe
    btutil LNG key G $(SRC_DIR)\language.txt german.lng

startrek.lng:   $(SRC_DIR)\language.txt btutil.exe
    btutil LNG key S $(SRC_DIR)\language.txt startrek.lng

#   patch language file into bt.exe

patche: bt.exe btutil.exe english.lng
    btutil LNG patch bt.exe english.lng

patchg: bt.exe btutil.exe german.lng
    btutil LNG patch bt.exe german.lng

patchs: bt.exe btutil.exe startrek.lng
    btutil LNG patch bt.exe startrek.lng

#   patch language file into btbig.exe

patchbe: btbig.exe btutil.exe english.lng
    btutil LNG patch btbig.exe english.lng

patchbg: btbig.exe btutil.exe german.lng
    btutil LNG patch btbig.exe german.lng

patchbs: btbig.exe btutil.exe startrek.lng
    btutil LNG patch btbig.exe startrek.lng

#############################################################################
#                                                                           #
# Rulez for the executables                                                 #
#                                                                           #
#############################################################################
bt.exe: $(COBJS) $(OVLOBJS) $(MOBJS) $(COBJCOMMON)
        $(LRF) @<<$(PROJ).lrf
/INFO/FAR/PACKC/PACKF/DYNAMIC:315/SEG:192 $(OVLOBJS) + $(COBJS: = +^
) +
$(MOBJS: = +^
) +
$(COBJCOMMON: = +^
)
$*/STACK:36864 /NOI
$*/MAP:FULL/LI,,BT
<<
        $(LINKER) @$(PROJ).lrf

btbig.exe: $(COBJS) $(NOOVLOBJS) $(MOBJS) $(COBJCOMMON)
        $(LRF) @<<$(PROJ).lrf
/SEG:192 $(NOOVLOBJS) + $(COBJS: = +^
) +
$(MOBJS: = +^
) +
$(COBJCOMMON: = +^
)
$*/STACK:36864 /NOI
$*/MAP/LI;
<<
        $(LINKER) @$(PROJ).lrf

bt-dbg.exe: $(COBJS) $(NOOVLOBJS) $(MOBJS)
    $(LRF) @<<$(PROJ).lrf
/CODEVIEW/NOE/INFO/FAR/DYNAMIC:300/SEG:160 $(NOOVLOBJS) + $(COBJS: = +^
) +
$(MOBJS: = +^
) +
$(COBJCOMMON: = +^
)
$*/STACK:36864 /NOI
$*/MAP/LI;
<<
    $(LINKER) @$(PROJ).lrf

btutil.exe: $(COBJ2) $(COBJCOMMON)
    link /EXEPACK $(COBJ2) $(COBJCOMMON), $*/stack:3000,$*/map/li;

#############################################################################
#                                                                           #
# Rulez for .C -> .OBJ                                                      #
#                                                                           #
#############################################################################
$(COBJS): $(SRC_DIR)\$*.c includes.obj
    $(CC) $(CFLAGS) $(SRC_DIR)\$*.c

$(COBJCOMMON): $(SRC_DIR)\$*.c includes.obj
    $(CC) $(CFLAGS) $(SRC_DIR)\$*.c

$(COBJ2): $(SRC_DIR)\$*.c includes.obj
    $(CC) $(CFLAGS) $(SRC_DIR)\$*.c

b_spwnno.obj: $(SRC_DIR)\b_spawn.c includes.obj
    $(CC) $(CFLAGS) -Fob_spwnno.obj $(SRC_DIR)\b_spawn.c

b_spwnov.obj: $(SRC_DIR)\b_spawn.c includes.obj
    $(CC) -DMOVE_OVERLAYS $(CFLAGS) -Fob_spwnov.obj $(SRC_DIR)\b_spawn.c

#############################################################################
#                                                                           #
# Rulez for .ASM -> .OBJ                                                    #
#                                                                           #
# use additional switch -Zi for debugging                                   #
#                                                                           #
#############################################################################
$(MOBJS): $(SRC_DIR)\$*.asm
    $(ASM) -mx -DMODL=LARGE $(SRC_DIR)\$*;
    
#############################################################################
#                                                                           #
# Rulez for some special cases                                              #
#                                                                           #
#############################################################################
langload.obj: $(INCL_DIR)\msgs.h

includes.obj: $(INCL_FILES)
    $(CC) $(CFLAGS) $(SRC_DIR)\includes.c

#############################################################################
#                                                                           #
# Specify this target create a distribution archive                         #
#                                                                           #
#############################################################################
distrib:
    !ren btbig.exe bt.exe
    !ren btbig.map bt.map
    !pkzip bdos_xxx bt*.exe *.lng *.map ..\doc\xe_user.doc ..\src\language.txt

#############################################################################
#                                                                           #
# Specify this target make an archive of the source files                   #
#                                                                           #
#############################################################################
pack:
    !pkzip -p -r -x*.zip xe_src \binksrc\*.*

#############################################################################
#                                                                           #
# Specify this target to scrub the project directory                        #
#                                                                           #
#############################################################################
clean :
    !del *.bak
    !del *.err
    !del *.exe
    !del *.lng
    !del *.map
    !del *.obj
    !del *.zip

