/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                                                          */
/*      ------------         Bit-Bucket Software, Co.                       */
/*      \ 10001101 /         Writers and Distributors of                    */
/*       \ 011110 /          Freely Available<tm> Software.                 */
/*        \ 1011 /                                                          */
/*         ------                                                           */
/*                                                                          */
/*              (C) Copyright 1987-96, Bit Bucket Software Co.              */
/*                                                                          */
/*                This module was written by Vince Perriello                */
/*                                                                          */
/*                   BinkleyTerm Configuration File Parser                  */
/*                                                                          */
/*                                                                          */
/*    For complete  details  of the licensing restrictions, please refer    */
/*    to the License  agreement,  which  is published in its entirety in    */
/*    the MAKEFILE and BT.C, and also contained in the file LICENSE.260.    */
/*                                                                          */
/*    USE  OF THIS FILE IS SUBJECT TO THE  RESTRICTIONS CONTAINED IN THE    */
/*    BINKLEYTERM  LICENSING  AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF    */
/*    THIS  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES,  OR IF YOU DO    */
/*    NOT HAVE THESE FILES,  YOU  SHOULD  IMMEDIATELY CONTACT BIT BUCKET    */
/*    SOFTWARE CO.  AT ONE OF THE  ADDRESSES  LISTED BELOW.  IN NO EVENT    */
/*    SHOULD YOU  PROCEED TO USE THIS FILE  WITHOUT HAVING  ACCEPTED THE    */
/*    TERMS  OF  THE  BINKLEYTERM  LICENSING  AGREEMENT,  OR  SUCH OTHER    */
/*    AGREEMENT AS YOU ARE ABLE TO REACH WITH BIT BUCKET SOFTWARE, CO.      */
/*                                                                          */
/*                                                                          */
/* You can contact Bit Bucket Software Co. at any one of the following      */
/* addresses:                                                               */
/*                                                                          */
/* Bit Bucket Software Co.        FidoNet  1:104/501, 1:343/491             */
/* P.O. Box 460398                AlterNet 7:42/1491                        */
/* Aurora, CO 80046               BBS-Net  86:2030/1                        */
/*                                Internet f491.n343.z1.fidonet.org         */
/*                                                                          */
/* Please feel free to contact us at any time to share your comments about  */
/* our software and/or licensing policies.                                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/

/* Include this file before any other includes or defines! */

#include "includes.h"

static word Zone = 1;

void cfg_macro (char *);
void cfg_dial_string (char *);
void cfg_alt_num (char *);
void cfg_mdm_trans (char *);
void cfg_domain (char *);
void cfg_Janus (char *);
void cfg_domain_kludge (char *);

#ifdef BINKLEY_SOUNDS
char *cfg_sound (char *);
#endif

static char *cfgfgets(char buf[], FILE *fp); /* MMP 960419 / TJW 960928 */

#define CFGLINEBUFSIZE 3072

static struct parse_list far config_lines[] =
{
    { 8, "SameRing"},
    {11, "NewNodeList"},
    {13, "QuickNodeList"},
    {10, "Answerback"},
    { 5, "Macro"},
    { 5, "Shell"},
    { 4, "Dial"},
    { 5, "Event"},
    { 4, "Zone"},
    { 6, "MaxReq"},
    { 8, "LogLevel"},
    { 4, "Baud"},
    { 7, "MaxPort"},
    { 4, "Port"},
    { 7, "Carrier"},
    { 9, "StatusLog"},
    { 6, "Reader"},
    { 9, "BossPhone"},
    { 7, "BossPwd"},
    { 8, "Protocol"},
    { 6, "System"},
    { 5, "Sysop"},
    { 4, "Boss"},
    { 5, "Point"},
    { 3, "Aka"},
    { 4, "Hold"},
    { 9, "DownLoads"},
    { 7, "NetFile"},
    { 4, "Init"},
    { 4, "Busy"},
    { 6, "Prefix"},
    { 8, "NodeList"},
    { 5, "Avail"},
    { 6, "OKFile"},
    { 5, "About"},
    { 8, "MailNote"},
    { 6, "Banner"},
    {10, "UnAttended"},
    { 9, "OverWrite"},
    { 4, "Rev3"},
    { 7, "ReqOnUs"},
    { 8, "LockBaud"},
    { 7, "TimeOut"},
    { 5, "NoSLO"},
    { 9, "SlowModem"},
    {11, "SmallWindow"},
    { 8, "NoPickup"},
    {10, "NoRequests"},
    { 7, "NetMail"},
    { 6, "Suffix"},
    {12, "NoFullScreen"},
    { 8, "AutoBaud"},
    { 4, "Gong"},
    { 9, "NoCollide"},
    { 8, "TBBSList"},
    { 8, "TaskView"},
    { 7, "TopView"},
    { 9, "ExtrnMail"},
    { 7, "BBSNote"},
    { 3, "BBS"},
    {10, "ScriptPath"},
    { 7, "BoxType"},
    { 9, "MultiLink"},
    { 7, "Include"},
    {11, "CaptureFile"},
    { 9, "CursorCol"},
    { 9, "CursorRow"},
    { 9, "DoingMail"},
    { 8, "EnterBBS"},
    {10, "PrivateNet"},
    { 6, "Packer"},
    { 7, "Cleanup"},
    { 9, "AfterMail"},
    { 6, "Colors"},
    { 9, "JanusBaud"},
    {11, "ReqTemplate"},
    {10, "KnownAvail"},
    {12, "KnownReqList"},
    {10, "KnownAbout"},
    {12, "KnownInbound"},
    {11, "KnownReqLim"},
    {11, "KnownReqTpl"},
    { 9, "ProtAvail"},
    {11, "ProtReqList"},
    { 9, "ProtAbout"},
    {11, "ProtInbound"},
    {10, "ProtReqLim"},
    {10, "ProtReqTpl"},
    {11, "Application"},
    { 7, "NoZones"},
    { 6, "Answer"},
    { 9, "PollTries"},
    { 7, "SwapDir"},
    { 7, "Address"},
    {10, "CurMudgeon"},
    { 7, "NoWaZOO"},
    {11, "ScreenBlank"},
    {10, "Mark_Kromm"},
    { 6, "Server"},
    {10, "ModemTrans"},
    { 7, "PreDial"},
    { 7, "PreInit"},
    { 7, "DTRHigh"},
    { 5, "Debug"},
    { 8, "NoZedZap"},
    { 8, "NoResync"},
    { 9, "NoSEAlink"},
    { 8, "FTS-0001"},
    {10, "LineUpdate"},
    { 7, "JanusOK"},
    { 8, "TermInit"},
    {12, "DomainKludge"},
    { 5, "Snoop"},
    { 5, "Flags"},
    {11, "NoHydraChat"},        /* TJW 960804 */
    { 8, "MaxBytes"},
    {13, "KnownMaxBytes"},
    {12, "ProtMaxBytes"},
    {11, "StartBlkLen"},
    { 7, "MaxTime"},
    {12, "KnownMaxTime"},
    {11, "ProtMaxTime"},
    { 9, "RingTries"},
    { 9, "NoSharing"},
    { 8, "Version7"},
    { 9, "BlankWait"},
    { 6, "Serial"},
    {10, "NoDietIfna"},
    { 8, "Version6"},
    { 8, "WinSlice"},
    { 6, "NoSize"},
    { 8, "ForcExit"},
    { 6, "Extern"},
    { 6, "NoEMSI"},
    {10, "MyLocation"},
    { 7, "MyPhone"},
    {11, "MyListFlags"},
    { 9, "MyMaxbaud"},
    { 9, "PickUpAll"},
    { 7, "FileSec"},
    { 8, "KnownSec"},
    { 7, "ProtSec"},
    { 8, "MaxAreas"},
    { 6, "Domain"},
    { 9, "MailSound"},
    { 9, "FileSound"},
    { 8, "BBSSound"},
    { 8, "EXTSound"},
    {10, "StartSound"},
    { 8, "FaxSound"},
    { 8, "FaxInDir"},
    {12, "ModemFailure"},
    {11, "ModemIgnore"},
    {12, "ModemConnect"},
    {12, "ModemRinging"},
    {13, "ModemIncoming"},
    { 8, "ModemFax"},
    { 7, "Costlog"},
    { 8, "CostUnit"},
    { 8, "EuroCost"},
    { 8, "NoFilter"},
    { 6, "BiDiOK"},
    { 8, "BiDiBaud"},
    { 7, "NoHydra"},
    { 7, "NoJanus"},
    {12, "ExtBaudRates"},
    { 9, "AfterCall"},
    { 6, "PktRsp"},
    {10, "ExtSession"},
    { 7, "FaxBaud"},
    {13, "ErrLevelShell"},
    { 9, "AltNumber"},
    {10, "ModemRetry"},
    {19, "RecentActivityLines"},
    { 8, "RingWait"},
    { 9, "WinFOSSIL"},
                           /* TJW 960414 single config file for n tasks */ 
    { 1, "["},             /* section begin with expression             */

    { 8, "TaskPath"},      /* TJW 235 178 */
    { 9, "ShowAlive"},     /* RDH *** 9604 179 */
    {12, "ReadHoldTime"},  /* mtt 96-04-16 180 */
    {10, "ReInitTime"},    /* mtt 96-04-16 181 */         
    {15, "ConditionalPoll"},     /* CE/TJW 960425 182 */
    { 6, "Reject"},              /* CE/TJW 960425 183 */
    {18, "ProtocolPreference"},  /* TJW 960719 184 */
    { 6, "PutEnv"},          /* MMP 185 */
    { 9, "SpawnInit"},       /* MMP 186 */
    { 8, "LocalLog"},        /* MMP 187 */
    {12, "ShortCostLog"},    /* MMP 188 */
    { 8, "TimeSync"},        /* MMP 189 */
    { 8, "MailFlag"},        /* MMP 190 */
    { 7, "MakeDir"},         /* TJW 191 */
    { 4, "SRIF"},            /* MTT 192 */
    {12, "AfterConnect"},    /* PE 193 */
    {10, "MaxBusyAge"},      /* MMP 194 */
    {9,  "SpawnNoOK"},       /* MMP 195 */
    {9,  "CacheHold"},       /* MMP 196 */
    { 7, "PBAreas"},         /* TE 960516 197 */
    {11, "ModemReject"},     /* CE 198 */
    {12, "AfterCallOut"},    /* CE 199 */
    {12, "ChangeMailTo"},    /* CE 200 */

    {18, "CostTimeCorrection"},   /* TJW 201 */
    { 3, "IPC"},                  /* TJW 202 */
    { 7, "ReadLog"},              /* TJW 203 */
    {10, "EMSIbanner"},           /* TJW 204 */
    { 7, "CostCPS"},              /* TJW 205 */
    { 7, "MCPpipe"},              /* TJW 206 */

    {14, "WindowTitleFmt"},       /* TJW 207 */
    {10, "XErsrvd208"},   /* XE 208 */
    {10, "XErsrvd209"},   /* XE 209 */
    
    {10, "CErsrvd210"},   /* CE 210 */
    {10, "CErsrvd211"},   /* CE 211 */
    {10, "CErsrvd212"},   /* CE 212 */
    {10, "CErsrvd213"},   /* CE 213 */
    {10, "CErsrvd214"},   /* CE 214 */
    
    {10, "MTrsrvd215"},   /* mtt 215 */
    {10, "MTrsrvd216"},   /* mtt 216 */
    {10, "MTrsrvd217"},   /* mtt 217 */
    {10, "MTrsrvd218"},   /* mtt 218 */
    {10, "MTrsrvd219"},   /* mtt 219 */
    
    {10, "RHrsrvd220"},   /* RH 220 */
    {10, "RHrsrvd221"},   /* RH 221 */
    {10, "RHrsrvd222"},   /* RH 222 */
    {10, "RHrsrvd223"},   /* RH 223 */
    {10, "RHrsrvd224"},   /* RH 224 */
    
    {11, "TJWrsrvd225"},  /* TJW 225 */
    {11, "TJWrsrvd226"},  /* TJW 226 */
    {11, "TJWrsrvd227"},  /* TJW 227 */
    {11, "TJWrsrvd228"},  /* TJW 228 */
    {11, "TJWrsrvd229"},  /* TJW 229 */

    {10, "TErsrvd230"},  /* TE 230 */
    {10, "TErsrvd231"},  /* TE 231 */
    {10, "TErsrvd232"},  /* TE 232 */
    {10, "TErsrvd233"},  /* TE 233 */
    {10, "TErsrvd234"},  /* TE 234 */

    { 0, ""}
};

void
parse_config (char *config_file)
{
    FILE *stream;
    char *temp;
    char buf[CFGLINEBUFSIZE];
    char *c=NULL;       /* TS 960620  since maybe uninitialized*/
    char *inclcfgfname; /* TJW 960414 added for "Include" verb */
    int i;
    int linecounter=0;
    long ltemp;
    char *p;
    ADDR temp_addr;    /* Used to get address */
    int usecfglines=1; /* TJW 960414 use config lines for this task ? init: yes */
    char tmpfname[PATHLEN];
    char junk[256];

    (void) sprintf (tmpfname, "%s%s", BINKpath, config_file);
    printf("Reading configuration file %s...\n",tmpfname);
    
    if ((stream = share_fopen (tmpfname, read_ascii, DENY_WRITE)) == NULL)  /* OK, let's open the file   */
        return;                 /* no file, no work to do    */

    /* MMP 960419. cfgfgets() substitutes environment variables, removes
     * comments, does conditional inclusion of lines, and truncates long lines */

    while ((temp=cfgfgets(buf, stream)) != NULL)          /* while not EOF */
                                              /* Now we parse the file ... */
    {
        linecounter++;          /* TJW 960524 */
        
        if (strlen(temp)<3)     /* If too short, ignore it */
            continue;

        if ((i = parse (temp, config_lines)) != -1)
        {
            c = &temp[(config_lines[i - 1]).p_length];
            c = skip_blanks (c);
        }

        switch (i)                  /* TJW 960414                    */
        {
        case 177:                   /* TJW 960830 [expression]       */
            p=strchr(c,']');
            if(p)
              *p='\0';              /* if there is a ] then kill it  */
            else
              goto bad_line;        /* else syntax error             */
            if((usecfglines=eval(c))<0)  /* if expr = true then use lines */ 
              goto bad_line;
            break;
        default:
        if(usecfglines) switch (i)
        {
        case 1:                 /* "SameRing"      */
            /* This modem doesn't differentiate
               between RING and RINGING */

            def_mdm_resps[2].disp = IGNORE;
            break;

        case 2:                 /* "NewNodeList"   */
        case 129:               /* "Version6"      */
            newnodelist = 1;
            break;

        case 3:                 /* "QuickNodeList" */
#ifdef QUICK_LIST
            nodefunc = QuickLookup;
            newnodelist = 0;
#endif
            break;

        case 4:                 /* "Answerback"    */
            answerback = ctl_string (c);
            break;

        case 5:                 /* "Macro"         */
            cfg_macro (c);
            break;

        case 6:                 /* "Shell"         */
            i = atoi (c);
            if ((i <= 0) || (i > 9))
            {
                (void) printf ("%s %d %s\n", MSG_TXT (M_SHELL_NUMBER), i, MSG_TXT (M_OUT_OF_RANGE));
                break;
            }
            c = skip_to_word (c);
            shells[i - 1] = calloc (1, strlen (c) + 1);
            (void) strcpy (shells[i - 1], c);
            (void) strupr (shells[i - 1]);
            break;

        case 7:                 /* "Dial"          */
            cfg_dial_string (c);
            break;

        case 8:                 /* "Event"         */
            (void) parse_event (c);
            break;

        case 9:                 /* "Zone"          */
            Zone = (word) atoi (c);
            if (!Zone)          /* if we didn't find a zone  */
                (void) printf (MSG_TXT (M_ILLEGAL_ZONE), &temp[4]);
            break;

        case 10:                /* "MaxReq"        */
            (void) sscanf (c, "%d %d", &DEFAULT.rq_Limit, &DEFAULT.rq_Cum_Limit);
            if (!DEFAULT.rq_Limit)  /* No requests??? */
                (void) printf ("0 %s\n", MSG_TXT (M_REQUESTS_ALLOWED));
            break;

        case 11:                /* "LogLevel"      */
            i = atoi (c);
            if ((abs(i) <= 6) && (i != 0))
            {
                loglevel = i;
            }
            else
            {
                (void) printf (MSG_TXT (M_BAD_LOGLEVEL), &temp[8]);
            }
            break;

        case 12:                /* "Baud"          */
            ltemp = atol (c);
            max_baud.rate_value = 0;
            for (i = 0; pbtypes[i].rate_value; i++)
            {
                if (pbtypes[i].rate_value == (unsigned long)ltemp)
                {
                    max_baud.rate_mask = pbtypes[i].rate_mask;
                    max_baud.rate_value = ltemp;
                    break;
                }
            }
            if (!max_baud.rate_value)
                goto bad_line;
            break;

        case 13:                /* "MaxPort"       */
            i = atoi (c);               /* make it binary  */
            if ((i > 0) && (i < 33))    /* if range is reasonable,   */
                MAXport = i;            /* Make it the max value     */
            else
                (void) printf (MSG_TXT (M_BAD_MAXPORT), &temp[7]);
            break;

        case 14:                /* "Port"         */
            i = atoi (c);               /* make it binary            */
            if ((i > 0) || (i < 33))    /* see if it's OK            */
            {
                port_ptr = i - 1;       /* store it if so            */
                if (MAXport < i)        /* See if MAXport is less    */
                    MAXport = i;        /* If so, increase MAXport   */
                original_port = port_ptr;
            }
            else
                (void) printf (MSG_TXT (M_ILLEGAL_PORT), &temp[4]);
            break;

        case 15:                /* "Carrier"      */
            i = 0;
            (void) sscanf (c, "%x", &i);/* convert it to binary      */
            if (i != 0)                 /* if we got anything        */
                carrier_mask = (unsigned int) i;
            else
                (void) printf (MSG_TXT (M_ILLEGAL_CARRIER), &temp[7]);
            break;

        case 16:                /* "StatusLog"    */
            if (log_name==NULL)
              log_name = ctl_string (c);
            break;

        case 17:                /* "Reader"       */
            BBSreader = ctl_string (c);
            break;

        case 18:                /* "BossPhone"    */
            BOSSphone = ctl_string (c);
            break;

        case 19:                /* "BossPwd"      */
            BOSSpwd = ctl_string (c);
            break;

        case 20:                /* "Protocol"     */
            if (extern_index == 5)
                break;
            protocols[extern_index++] = ctl_string (c);
            break;

        case 21:                /* "System"       */
            system_name = ctl_string (c);
            break;

        case 22:                /* "Sysop"        */
            sysop = ctl_string (c);
            break;

        case 23:                /* "Boss"         */
            (void) parse_address (c, &boss_addr);
            break;

        case 24:                /* "Point"        */
        case 25:                /* "Aka"          */
        case 94:                /* "Address"      */
            if (num_addrs == 0)
                alias[0].Zone = Zone;
            (void) parse_address (c, &temp_addr);
            alias[num_addrs] = temp_addr;
            Zone = alias[0].Zone;   /* First is real default */
            ++num_addrs;
            break;

        case 26:                /* "Hold"         */
            hold_area = ctl_slash_string (c);
            if (hold_area == NULL)
                break;
            domain_area = strdup (hold_area);
            domain_area[strlen (domain_area) - 1] = '\0';
            p = strrchr (domain_area, '\\');
            if (p == NULL)
            {
                p = strrchr (domain_area, '/');
                if (p == NULL)
                {
                    p = strrchr (domain_area, ':');
                    if (p == NULL)
                        p = domain_area;
                }
            }
            if (p != domain_area)
                ++p;

            domain_loc = p;
            *p = '\0';
            break;

        case 27:                /* "DownLoads"    */
            download_path = ctl_slash_string (c);
            break;

        case 28:                /* "NetFile"      */
            DEFAULT.sc_Inbound = ctl_slash_string (c);
            break;

        case 29:                /* "Init"         */
            modem_init = ctl_string (c);
            break;

        case 30:                /* "Busy"         */
            modem_busy = ctl_string (c);
            break;

        case 31:                /* "Prefix"       */
            predial = ctl_string (c);
            normprefix = predial;
            break;

        case 32:                /* "NodeList"     */
            net_info = ctl_slash_string (c);
            break;

        case 33:                /* "Avail"        */
            DEFAULT.rq_FILES = ctl_string (c);
            break;

        case 34:                /* "OKFile"       */
            DEFAULT.rq_OKFile = ctl_string (c);
            break;

        case 35:                /* "About"        */
            DEFAULT.rq_About = ctl_string (c);
            break;

        case 36:                /* "MAILnote"     */
            MAILnote = ctl_string (c);
            break;

        case 37:                /* "Banner"       */
            BBSbanner = ctl_string (c);
            break;

        case 38:                /* "UnAttended"   */
            un_attended = 1;
            command_line_un = 1;
            break;

        case 39:                /* "OverWrite"    */
            overwrite = 1;
            break;

        case 40:                /* "Rev3"         */
            rev3 = 1;
            break;

        case 41:                /* "ReqOnUs"      */
            on_our_nickel = 1;
            break;

        case 42:                /* "LockBaud"     */
            if (*c == '\0')     /* Case 1: no arg */
            {
                lock_baud = 1;  /* Always lock    */
                break;
            }

            if (isdigit (*c))   /* Case 2: baud   */
            {
                lock_baud = atol (c);   /* Lock at >= baud*/
                break;
            }

            i = 1 + strlen (c); /* Case 3: ARQ    */
            if (ARQs > 16)      /* Only 16 allowed*/
                break;
            p = calloc (1, i);  /* Allocate space */
            if (p == NULL)      /* Give up if none*/
                break;
            (void) strcpy (p, c);   /* Copy ARQ string*/
            ARQ[ARQs++] = p;    /* Update array   */
            break;

        case 43:                /* "TimeOut"      */
            BBStimeout = (unsigned) (atoi (c) * 100);
            if (BBStimeout < 2000)
                BBStimeout = 2000;
            break;

        case 44:                /* "NoSLO"        */
            no_overdrive = 1;
            break;

        case 45:                /* "SlowModem"    */
            slowmodem = 1;
            break;

        case 46:                /* "SmallWindow"  */
            small_window = 1;
            break;

        case 47:                /* "NoPickup"     */
            no_pickup = 1;
            break;

        case 48:                /* "NoRequests"   */
            no_requests = 1;
            break;

        case 49:                /* "NetMail"      */
            if (*c != '$')
                netmail = ctl_slash_string (c);
            else
                netmail = ctl_string (c);
            break;

        case 50:                /* "Suffix"       */
            postdial = ctl_string (c);
            normsuffix = postdial;
            break;

        case 51:                /* "NoFullScreen" */
            fullscreen = 0;
            break;

        case 52:                /* "AutoBaud"     */
            autobaud = 1;
            break;

        case 53:                /* "Gong"         */
            gong_allowed = 1;
            break;

        case 54:                /* "NoCollide"    */
            no_collide = 1;
            break;

        case 55:                /* "TBBSlist"     */
#ifdef OLDTBBSLIST
            nodefunc = TBBSLookup;
            newnodelist = 0;
#endif
            break;

        case 56:                /* "TaskView"     */
        case 57:                /* "TopView"      */
            have_tv = 1;
            break;

        case 58:                /* "ExtrnMail"    */ 
            if (isdigit (*c))
            {
                lev_ext_mail[num_ext_mail] = atoi (c);
                c = skip_to_blank (c);
            }
            else
            {
                lev_ext_mail[num_ext_mail] = 99;
            }
            ext_mail_string[num_ext_mail++] = ctl_string (c);
            break;

        case 59:                /* "BBSnote"      */
            BBSnote = ctl_string (c);
            break;

        case 60:                /* "BBS"          */
            BBSopt = ctl_string (c);
            break;

        case 61:                /* "ScriptPath"   */
            script_path = ctl_slash_string (c);
            break;

        case 62:                /* "BoxType"      */
            boxtype = (short) atoi (c);
            if ((boxtype > 4) || (boxtype < 0))
                boxtype = 1;
            break;

        case 63:                /* "MultiLink"    */
            have_ml = 1;
            break;

        case 64:                /* "Include"      */
            inclcfgfname = ctl_string (c); /* TJW 960414 evaluate %xyz% in */
            parse_config (inclcfgfname);   /* included files              */
            break;

        case 65:                /* "CaptureFile"  */
            logptr = ctl_string (c);
            break;

        case 66:                /* "CursorCol"    */
            cursor_col = atoi (c) - 1;
            break;

        case 67:                /* "CursorRow"    */
            cursor_row = atoi (c) - 1;
            break;

        case 68:                /* "DoingMail"    */
            (void) sprintf (junk, "\r\r%s\r\r", c);
            noBBS = calloc (1, strlen (junk) + 1);
            (void) strcpy (noBBS, junk);
            break;

        case 69:                /* "EnterBBS"     */
            (void) sprintf (junk, "\r%s\r", c);
            BBSesc = calloc (1, strlen (junk) + 1);
            (void) strcpy (BBSesc, junk);
            break;

        case 70:                /* "PrivateNet"   */
            pvtnet = (short) atoi (c);
            break;

        case 71:                /* "Packer"       */
            packer = ctl_string (c);
            break;

        case 72:                /* "Cleanup"      */
            cleanup = ctl_string (c);
            break;

        case 73:                /* "AfterMail"    */
            if ( !memicmp(c,"quick",5) && (isspace(c[5]) || c[5]=='\0') )
                quickaftermail=1, c+=5;
            aftermail = ctl_string (c);
            break;

        case 74:                /* "Colors"       */
            sscanf (c, "%hd %hd %hd %hd %hd %hd %hd %hd %hd %hd",
                &colors.background, &colors.settings, &colors.history,
                &colors.hold, &colors.call, &colors.file,
                &colors.calling, &colors.popup,
                &colors.headers, &colors.frames  /* TJW 960428 do it like EE */
                );
            break;

        case 163:               /* "BiDiBaud"     */
        case 75:                /* "JanusBaud"    */
            janus_baud = atol (c);
            break;

        case 76:                /* "ReqTemplate"  */
            DEFAULT.rq_Template = ctl_string (c);
            break;

        case 77:                /* "KnownAvail"   */
            KNOWN.rq_FILES = ctl_string (c);
            break;

        case 78:                /* "KnownReqList" */
            KNOWN.rq_OKFile = ctl_string (c);
            break;

        case 79:                /* "KnownAbout"   */
            KNOWN.rq_About = ctl_string (c);
            break;

        case 80:                /* "KnownInbound" */
            KNOWN.sc_Inbound = ctl_slash_string (c);
            break;

        case 81:                /* "KnownReqLim"  */
            (void) sscanf (c, "%d %d", &KNOWN.rq_Limit, &KNOWN.rq_Cum_Limit);
            break;

        case 82:                /* "KnownReqTpl"  */
            KNOWN.rq_Template = ctl_string (c);
            break;

        case 83:                /* "ProtAvail"   */
            PROT.rq_FILES = ctl_string (c);
            break;

        case 84:                /* "ProtReqList" */
            PROT.rq_OKFile = ctl_string (c);
            break;

        case 85:                /* "ProtAbout"   */
            PROT.rq_About = ctl_string (c);
            break;

        case 86:                /* "ProtInbound" */
            PROT.sc_Inbound = ctl_slash_string (c);
            break;

        case 87:                /* "ProtReqLim"  */
            (void) sscanf (c, "%d %d", &PROT.rq_Limit, &PROT.rq_Cum_Limit);
            break;

        case 88:                /* "ProtReqTpl"  */
            PROT.rq_Template = ctl_string (c);
            break;

        case 89:                /* "Application" */
            break;

        case 90:                /* "NoZones"     */
            no_EMSI = 1;        /* no zones means no EMSI */
            no_zones = 1;
            break;

        case 91:                /* "Answer"        */
            ans_str = ctl_string (c);
            break;

        case 92:                /* "PollTries"   */
            poll_tries = atoi (c);
            break;

        case 93:                /* "SwapDir"     */
            swapdir = ctl_slash_string(c);
            break;

        case 95:                /* "Curmudgeon"  */
            curmudgeon = 1;     /* Do it like Randy */
            break;

        case 96:                /* "NoWaZOO"     */
            no_WaZOO = 1;         /* FTSC sessions */
            break;

        case 97:                /* "ScreenBlank" */
            do_screen_blank = 1;
            if (strnicmp (c, MSG_TXT (M_KEY), 3) == 0)
                blank_on_key = 1;
            else if (strnicmp (c, MSG_TXT (M_CALL), 4) == 0)
                blank_on_key = 0;
            break;

        case 98:                /* "Mark_Kromm"  */

            /* These colors from Mark Kromm, Fidonet 1:261/1034  */

            colors.background = 112;    /* The           */
            colors.settings = 9;/* Great         */
            colors.history = 10;/* BinkleyTerm   */
            colors.hold = 11;   /* Colors        */
            colors.call = 14;   /* Contest       */
            colors.file = 12;   /* Winner!!!!    */
            colors.calling = 56;
            colors.popup = 15;
            break;

        case 99:                /* "Server"      */
            server_mode = 1;
            break;

        case 100:               /* "ModemTrans"  */
            cfg_mdm_trans (c);
            break;

        case 101:               /* "PreDial" */
            dial_setup = ctl_string (c);
            break;

        case 102:               /* "PreInit" */
            init_setup = ctl_string (c);
            break;

        case 103:               /* "DTRHigh" */
            leave_dtr_high = 1;
            break;

        case 104:               /* "Debug" */
            debugging_log = 1;
            break;

        case 105:               /* "NoZedZap" */
            no_zapzed = 1;
            break;

        case 106:               /* "NoResync" */
            no_resync = 1;
            break;

        case 107:               /* "NoSEAlink" */
            no_sealink = 1;
            break;

        case 108:               /* "FTS-0001" */
            no_sealink = 1;
            no_resync = 1;
            no_overdrive = 1;
            no_WaZOO = 1;         /* FTSC sessions  */
            no_EMSI = 1;
            break;

        case 109:               /* "LineUpdate"   */
            immed_update = 1;
            break;

        case 162:               /* "BiDiOK"       */
        case 110:               /* "JanusOK"      */
            cfg_Janus (c);
            break;

        case 111:               /* "TermInit"     */
            term_init = ctl_string (c);
            break;

        case 112:               /* "DomainKludge" */
            cfg_domain_kludge (c);
            break;

        case 113:               /* "Snoop"        */
#ifdef Snoop
            if (getenv ("SNOOPPIPE") == NULL)
            {
                strcpy (pipename, c);
                if (pipename[0] != '\0')
                    snoop_open (pipename);
            }
#endif                          /* Snoop */
            break;
            
        case 114:               /* "Flags"        */
            flag_dir = ctl_slash_string (c);
            break;

        case 115:               /* "NoHydraChat"  */ 
            nohydrachat = 1;
            break;

        case 116:               /* "MaxBytes"     */
            (void) sscanf (c, "%ld %ld", &DEFAULT.byte_Limit, &DEFAULT.byte_Cum_Limit);
            break;

        case 117:               /* "KnownMaxBytes"*/
            (void) sscanf (c, "%ld %ld", &KNOWN.byte_Limit, &KNOWN.byte_Cum_Limit);
            break;

        case 118:               /* "ProtMaxBytes" */
            (void) sscanf (c, "%ld %ld", &PROT.byte_Limit, &PROT.byte_Cum_Limit);
            break;

        case 119:               /* "StartBlkLen"  */
            fstblklen = atoi (c);
            if ((fstblklen < 64) || (fstblklen > 2048))
                fstblklen = 1024;
            break;

        case 120:               /* "MaxTime"      */
            (void) sscanf (c, "%d %ld", &DEFAULT.time_Limit, &DEFAULT.time_Cum_Limit);
            DEFAULT.time_Limit *= 60;
            DEFAULT.time_Cum_Limit *= 60;
            break;

        case 121:               /* "KnownMaxTime" */
            (void) sscanf (c, "%d %ld", &KNOWN.time_Limit, &KNOWN.time_Cum_Limit);
            KNOWN.time_Limit *= 60;
            KNOWN.time_Cum_Limit *= 60;
            break;

        case 122:               /* "ProtMaxTime"  */
            (void) sscanf (c, "%d %ld", &PROT.time_Limit, &PROT.time_Cum_Limit);
            PROT.time_Limit *= 60;
            PROT.time_Cum_Limit *= 60;
            break;

        case 123:               /* "RingTries"    */
            ring_tries = atoi (c);
            break;

        case 124:               /* "NoSharing"    */
            no_sharing = 1;
            break;

        case 125:               /* "Version7"     */
            version7 = 1;
            nodefunc = ver7find;
            userfunc = opususer;
            newnodelist = 0;
            break;

        case 126:               /* "BlankWait"    */
            blank_time = (unsigned int) 100 *(unsigned int) atoi (c);
            break;

        case 127:               /* "Serial"       */
            if (isdigit (*c))
                serial = atoi (c);
            break;

        case 128:               /* "NoDietIfna"   */
            no_dietifna = 1;
            break;

        case 130:
            winslice = 1;       /* "WinSlice"     */
            break;

        case 131:               /* "NoSize"       */
            no_size = 1;
            break;

        case 132:               /* "ForcExit" */
            forcexit = atoi (c);
            break;

        case 133:               /* "Extern" */
            EXTopt = ctl_string (c);
            break;

        case 134:               /* "NoEMSI"       */
            no_EMSI = 1;
            break;

        case 135:               /* "MyLocation"   */
            location = ctl_string (c);
            break;

        case 136:               /* "MyPhone"      */
            myphone = ctl_string (c);
            break;

        case 137:               /* "MyListFlags"  */
            myflags = ctl_string (c);
            break;

        case 138:               /* "MyMaxbaud"    */
            mymaxbaud = ctl_string (c);
            break;

        case 139:               /* "PickupALL"    */
            pickup_all = 1;
            break;

        case 140:               /* "FileSec"      */
            (void) sscanf (c, "%d", &DEFAULT.security);
            break;

        case 141:               /* "KnownSec"     */
            (void) sscanf (c, "%d", &KNOWN.security);
            break;

        case 142:               /* "ProtSec"      */
            (void) sscanf (c, "%d", &PROT.security);
            break;

        case 143:               /* "MaxAreas"     */
            fri_areadat = ctl_string (c);
            fri_areatype = FRI_MAXIMUS;
            break;

        case 144:               /* "Domain"       */
            cfg_domain (c);
            break;

        case 145:               /* "MailSound"    */
#ifdef BINKLEY_SOUNDS
            fnMailSound = cfg_sound (c);
#endif
            break;

        case 146:               /* "FileSound"    */
#ifdef BINKLEY_SOUNDS
            fnFileSound = cfg_sound (c);
#endif
            break;

        case 147:               /* "BBSSound"     */
#ifdef BINKLEY_SOUNDS
            fnBBSSound = cfg_sound (c);
#endif
            break;

        case 148:               /* "EXTSound"     */
#ifdef BINKLEY_SOUNDS
            fnEXTSound = cfg_sound (c);
#endif
            break;

        case 149:               /* "StartSound"   */
#ifdef BINKLEY_SOUNDS
            fnStartSound = cfg_sound (c);
#endif
            break;

        case 150:               /* "FaxSound"     */
#ifdef BINKLEY_SOUNDS
            fnFAXSound = cfg_sound (c);
#endif
            break;

        case 151:               /* "FaxInDir"     */
            fax_in = ctl_slash_string (c);
            break;

        case 152:               /* "ModemFailure" */
            i = FAILURE;
            goto add_mdm_response;

        case 153:               /* "ModemIgnore"  */
            i = IGNORE;
            goto add_mdm_response;

        case 154:               /* "ModemConnect" */
            i = CONNECTED;
            goto add_mdm_response;

        case 155:               /* "ModemRinging" */
            i = RINGING;
            goto add_mdm_response;

        case 156:               /* "ModemIncoming"*/
            i = INCOMING;
            goto add_mdm_response;

        case 157:               /* "ModemFax"     */
            i = FAX;
            /* goto add_mdm_response; */

add_mdm_response:

            /* If we have no space, bump the array */
            if (resp_count == max_resp_count)
            {
                int new_max = max_resp_count + 32;
                int new_size = new_max * sizeof (struct resp_str);
                struct resp_str *new_resps;

                new_resps = (struct resp_str *) realloc (mdm_resps, new_size);
                if (new_resps == NULL)
                    break;
                max_resp_count = new_max;
                mdm_resps = new_resps;
            }
            mdm_resps [resp_count].resp = ctl_string (c);
            mdm_resps [resp_count++].disp = i;
            break;

        case 158:               /* "CostLog"    */
            if (cost_log_name==NULL)
              cost_log_name = ctl_string (c);
            break;

        case 159:               /* "CostUnit"   */
            if (isdigit (*c))
                cost_unit = atoi (c);
            break;

        case 160:               /* "EuroCost"   */
            eurocost = 1;
            break;

        case 161:               /* "NoFilter"   */
            if (!*c)
                goto bad_line;

            if (cMNP > 16)      /* Only 16 allowed */
                break;

            /*  Allocate space to copy the result code */

            i = 1 + strlen (c);
            p = calloc (1, i);
            if (p == NULL)
                break;

            /*  Copy the result code and store the pointer */

            (void) strcpy (p, c);
            pchMNP[cMNP++] = p;
            break;

        case 164:               /* "NoHydra"    */
            no_hydra = 1;
            break;

        case 165:               /* "NoJanus"    */
            no_janus = 1;
            break;

        case 166:               /* "ExtBaudRates"    */
#ifdef EXTENDED_BAUDRATES
            extend_rates = 1;
            pbtypes = exttypes;
#endif
            break;

        case 167:               /* "AfterCall" *//* MB 93-12-12  AFTERCALL */
            if (isdigit (*c))
            {
                lines_aftercall = atoi (c);
                c = skip_to_blank (c);
            }
            modem_aftercall = ctl_string (c);
            break;

        case 168:               /* "PktRsp"      *//* MB 93-12-12  PKTRSP */
            pktrsp = 1;
            break;

        case 169:               /* "ExtSession" */
            if (!isxdigit (*c))
                goto bad_line;

            sscanf (c, "%x", &i);
            ExtMailMask = (word) i;
            if (!ExtMailMask)
                goto bad_line;

            c = skip_to_blank (c);
            ExtMailAgent = ctl_string (c);
            break;
            
        case 170:               /* "FaxBaud"    */
            ltemp = atol (c);
            fax_baud = -1;
            for (i = 0; pbtypes[i].rate_value; i++)
            {
                if (pbtypes[i].rate_value == (unsigned long)ltemp)
                {
                    fax_baud = i;
                    break;
                }
            }
            if (fax_baud == -1)
                goto bad_line;
            break;

        case 171:               /* "ErrLevelShell" */
            i = atoi (c);
            if ((i <= 0) || (i > 255))
            {
                (void) printf (MSG_TXT (M_BAD_ERRLEVELSHELLNUM), i);
                break;
            }
            c = skip_to_word (c);
            if (c)
            {
                strupr (c);
                errlvlshell[i-1] = ctl_string (c);
            }
            break;

        case 172:               /* AltNumber */
            cfg_alt_num (c);
            break;

        case 173:               /* ModemRetry */
            i = DIALRETRY;
            goto add_mdm_response;


        case 174:               /* RecentActivityLines */
            i = atoi (c) ;

            if (i > 0) 
                scrllines = i;
            break;

        case 175:               /* "RingWait"    */
            ring_wait = atoi (c);
            break;

        case 176:               /* "WinFOSSIL"   */
#ifdef _WIN32
            use_winfossil = 1;
#endif
            break;

        /* case 177: TJW 960830 [expression] - see switch before */

        case 178:                          /* TJW 960830 TaskPath */ 
            task_dir = ctl_slash_string (c);
            break;
            
        case 179:                          /* RDH 9604 "ShowAlive" */
            should_show_alive = 1;
            break;

        case 180:                          /* "ReadHoldTime" */
             readholdtime = atoi(c) * 60L;
             break;
 
        case 181:                          /* "ReInitTime" */
             reinittime = atoi(c);                        /* TJW 960421 */
             if(reinittime<1)  reinittime=1;              /* TJW 960421 */
             if(reinittime>10) reinittime=10;             /* TJW 960421 */ 
             reinittime=(unsigned int)(reinittime*6000L); /* TJW 960421 */ 
             break;

        case 182:                          /* CE/TJW 960425 "ConditionalPoll" */
             if(!strnicmp(c,"AND",3))
               fpaddr[num_fpaddrs].and_op = 1;       /* TJW 960505 boolean op */
             else
             if(!strnicmp(c,"OR",2))
               fpaddr[num_fpaddrs].and_op = 0;       /* TJW 960505 boolean op */
               
             c = skip_to_word (c);
             (void) parse_address (c, &temp_addr);         /* AKA */ 
             fpaddr[num_fpaddrs].addr = temp_addr; /* TJW 960505 assign struct */
             
             c = skip_to_word (c);
             fpaddr[num_fpaddrs].Size = atoi (c);         /* Size KB */
             
             c = skip_to_word (c);
             fpaddr[num_fpaddrs].Delta_T = atoi (c);      /* delta time */
             
             c = skip_to_word (c);
             fpaddr[num_fpaddrs].Number = ctl_string (c); /* phonenumber */

             num_fpaddrs++;                             /* one more now! */
             break;
             
        case 183:       /* CE/TJW 960425 "Reject" */
            call_reject = ctl_string (c);
            break;

        case 184:                          /* TJW 960719 "ProtocolPreference" */ 
            (void) strncpy (protpreference, c, 20);
            protpreference[19]='\0';
            (void) strupr (protpreference);
            break;

        case 185:               /* MMP 960419 "PutEnv" */
            i = 1 + strlen (c);
            p = calloc (1, i);
            if (p == NULL)
                break;
            strcpy(p,c);
            for (c=p; *c && *c!='='; c++)
                *c=(char)toupper(*c);
            putenv(p);
            break;

        case 186:               /* MMP 960419 "SpawnInit" */
#ifndef DOS16   /*  TS 960902   */
            system(c);
#else   /*  #ifndef DOS16   */
            b_spawn(c);
#endif  /*  #ifndef DOS16   */
            break;

        case 187:               /* MMP 960419 "LocalLog" */
            locallogs = 1;
            break;

        case 188:               /* MMP 960422 "ShortCostLog" */
            if (*c) {
                i = 1 + strlen (c);
                p = calloc (1, i);
                if (p == NULL)
                    break;
                strcpy(p,c);
                shortcostlog = p;
            } else {
                shortcostlog = "$02y$02m$02d $02H$02M$02S $# $14A $6B $4s $4c $4r $8i $8o $8C $4Y";
                               /* MMP 960513 Spaces removed */
                               /* TJW 960813 added $4r for remote's cost */
            }
            break;

        case 189:               /* MMP/TJW 960419 "TimeSync" */
            (void) parse_address (c, &TimeSync);
            c = skip_to_word (c);
            max_diff_trx = atol(c);
            break;

        case 190:               /* MMP 960419 "MailFlag" */
            MailFlag = *c ? ctl_string(c) : "";
            break;

        case 191:               /* TJW 960429 "MakeDir" */
            makedir = 1;
            break;
        
        case 192:               /* mtt 960502 "SRIF" */           
             for(p=SRIF_commandline;*c;c++,p++){ /* TJW 960503 */
                switch(*c){
                   case '$':
                     *p='%';      /* % is comment, so use $ instead in cfg */
                     break;
                   default:
                     *p=*c;
                }
             }
             *p='\0';
             break;

        case 193:               /* "AfterConnect" *//* PE 1996-04-25  AFTERCONNECT */
            if (isdigit (*c))
            {
                lines_afterconnect = atoi (c);
                c = skip_to_blank (c);
            }
            modem_afterconnect = ctl_string (c);
            break;

        case 194:                          /* MaxBusyAge */
            maxbsyage = atol(c);
            break;

        case 195:                          /* SpawnNoOK */
            spawnnook = ctl_string(c);
            break;

        case 196:                          /* CacheHold */
            cachehold = *c ? atoi(c) : 2;
            c = skip_to_word (c);
            if (!memicmp(c,"stat",4))
                rescanstat = 1;
            break;

        case 197:                          /* "PBAreas"     */
            fri_areadat = ctl_string (c);
            fri_areatype = FRI_PROBOARD;
            break;

        case 198:                          /* ModemReject */
            ModemReject = ctl_string (c);
            break;

        case 199:                          /* AfterCallOut */
            if (isdigit (*c))
            {
                lines_aftercallout = atoi (c);
                c = skip_to_blank (c);
            }
            modem_aftercallout = ctl_string (c);
            break;

        case 200:                          /* ChangeMailTo */
            if (!stricmp (c, "CRASH"))
                chg_mail = (int) 'C';
            else
                if (!stricmp (c, "DIRECT"))
                    chg_mail = (int) 'D';
                else
                    if (!stricmp (c, "HOLD"))
                        chg_mail = (int) 'H';
                    else
                        if (!stricmp (c, "NORMAL"))
                            chg_mail = (int) 'F';
                        else
                            goto bad_line;
            break;

        case 201:                        /* TJW 960721 CostTimeCorrection */ 
            sscanf(c, "%d %d", &costtimecorr_conn, &costtimecorr_sess);
            break;
            
        case 202:               /* "IPC"        */
            ipc_dir = ctl_slash_string (c);
            break;

        case 203:               /* "ReadLog"    */
            if (readlog_name==NULL)
              readlog_name = ctl_string (c);
            break;
        
        case 204:                          /* "EMSIbanner" */ 
            EMSIbanner = 1;
            break;
            
        case 205:                          /* TJW 960802 "CostCPS" */ 
            costcps = atol(c);
            break;
            
        case 206:                          /* TJW 961010 "MCPpipe" */
#ifdef MAX_MCP
            if (getenv ("MCPPIPE") == NULL)
              strcpy (MCPpipename, c);
            else
              strcpy(MCPpipename, getenv("MCPPIPE"));
#endif
            break;

        case 207:                          /* TJW 961013 "WindowTitleFmt" */
#ifdef OS2SETTITLE
            title_format = *c ? ctl_string(c) : NULL;
#endif            
            break;
            
        case 208:                          /* XE reserved */
        case 209:
             break;

        case 210:                         /* CE reserved */ 
        case 211:
        case 212:
        case 213:
        case 214:
             break;
             
        case 215:                         /* MTT reserverd */
        case 216:
        case 217:
        case 218:
        case 219:
             break;

        case 220:                          /* RH reserved */
        case 221:
        case 222:
        case 223:
        case 224:
             break;
             
        case 225:                          /* TJW reserved */
        case 226:
        case 227:
        case 228:
        case 229:
             break;

        case 230:                          /* TE reserved */
        case 231:
        case 232:
        case 233:
        case 234:
             break;

        default:
bad_line:
            (void) printf (MSG_TXT (M_UNKNOWN_LINE), tmpfname, linecounter, temp);
            ltemp = timerset (500U);
            while (!timeup (ltemp))
               ;                   /* TJW 960522 5s delay to show error msgs */
        }
        }
    }
    (void) fclose (stream);     /* close input file */

    if (debugging_log)
        loglevel = 6;
}


/* MMP 960419
 * cfgfgets() also has some advantages: it supports indented config verbs,
 * another method to specifying task specific lines, and it does not
 * dynamically allocate memory. Using it instead of fgets() we are also sure
 * to NUL-terminate the input line (the original code in parse_config() could
 * trap on very long lines).
 *
 * TJW 960422 changed comment definition to support stuff like:
 *            PutEnv PATH=x:\bt;x:\bt\bin and similar cases.
 *
 * TE 960516 made some changes around the "%"-char to support % in Init strings.
 *
 * TJW 960523 well, just another change, because this was really TOO
 * complicated and yet makes too much complications, therefore the solution
 * (hopefully) - see comments below.
 */

static char *cfgfgets(char buf[], FILE *fp)
{
    int bufpos = 0;                  /* index into buf[] */
    char tmp[CFGLINEBUFSIZE];        /* temporary buffer */
    int tmppos;                      /* index into tmp[] */
    int max=CFGLINEBUFSIZE-1;        /* make sure there is room for end '\0' */
    int ch;
    int comment = 0;        /* 1 if reading a comment */
    int whitespacebefore=1; /* TJW 960422 begin of line or char(s)
                             * before were white space
                             */
    char *var;

    /*
     * Check for EOF - we might have encountered it while reading previous line
     */
    if (feof(fp))
        return(NULL);

    /*
     * Read line. Check for the special characters '%' and ';' as we read.
     */
    while ( ch=fgetc(fp), ch!='\n' && ch!=EOF )
        if ( !comment )
            switch(ch) {
                case '%':
                    /*
                     * '%' has a VERY ambiguous semantics:
                     * - it can prefix a "rest of line"-comment (if it is at
                     *   1st non-white-space position in the config line)
                     * - it can designate the start/end of an environment
                     *   variable reference.
                     * - it can be used simply as "%" character
                     *
                     * We resolve this the following way:
                     * 1. if first non-white-space char, it is a comment.
                     * 2. if there are some characters between 2 percent signs
                     *    try to evaluate an environment variable of this name.
                     *    if not, meaning it is followed immediately by another
                     *    %, use it as a single "%" character.
                     * 4. if there is only a single "%" in this line, also use
                     *    it as a single "%" character.
                     */
                    if(whitespacebefore && !bufpos){ /* "%" first non-ws char */
                       comment=1;
                       break;
                    }
                    whitespacebefore = 0;
                    tmppos=0;
                    while(ch=fgetc(fp), ch!=EOF && ch!='\n' && ch!='%')
                       if (tmppos<max)
                          tmp[tmppos++]=(char)ch;
                    tmp[tmppos]='\0';

                    /* how did we get out of this loop ? */
                    if(ch=='%'){            /* there was a 2nd "%" */
                       if(tmppos==0){    /* it was "%%" -> just use one % */
                          if(bufpos<max)
                             buf[bufpos++]='%';
                       }else{            /* env var name is in tmp[] */
                          strupr(tmp);   /* TJW 960422 make it uppercase! */
                          var=getenv(tmp);
                          if(var){       /* valid env var -> copy value */
                            while(*var)
                              if(bufpos<max)
                                 buf[bufpos++]=*var++;
                          }else{         /* there is no such env var */
                                         /* -> copy verbatim chars in tmp */
                             for(tmppos=0;
                                 tmp[tmppos] && (bufpos<max);
                                 buf[bufpos++]=tmp[tmppos++]
                             );
                             if(bufpos<max)
                                buf[bufpos++]='%'; /* copy terminating %, too */
                          }
                       }
                    }else{              /* EOF or \n -  there was no 2nd "%" */
                       if(bufpos<max)
                          buf[bufpos++]='%'; /* use it as verbatim "%" char */
                       for(tmppos=0;
                           tmp[tmppos] && (bufpos<max);
                           buf[bufpos++]=tmp[tmppos++]
                       );
                       goto done;       /* there's nothing more in this line */
                    }   
                    break;

                    /*
                     * ';' is a bit WEIRD, too:
                     * - it can be a "rest-of-line"-comment
                     * - it can be a literal character
                     *
                     * We resolve this the following way:
                     * 1. if first non-white-space char, it is a comment.
                     * 2. if there is a sequence "<white space>;;", then don't
                          begin a comment, but just read it as "<white space>;"
                     * 3. if there is a sequence "<white space>;" (and no ";"
                     *    following), it is a comment
                     * 4. in any other case, just read it as ";" - no comment.
                     */
                case ';':
                    if(whitespacebefore){
                      if(!bufpos)          /* ";" first non-ws char */
                         comment=1;
                      else{
                        whitespacebefore = 0;
                        ch=fgetc(fp);        /* look at next char */
                        if(ch==';'){         /* ";" follows again - no comment */
                           if(bufpos<max)
                              buf[bufpos++]=';';
                        }else if(ch=='\n' || ch==EOF){ /* Null length comment */
                           goto done;
                        }else                /* this IS a comment */
                           comment=1;
                      }
                      break;
                    }
                    /* else: no break! fall through into default ! */
                    
                default:
                     whitespacebefore = isspace(ch);   /* don't store blanks  */
                     if (!whitespacebefore || bufpos){ /* at begining of line */
                        /* Store character literally */
                        if (bufpos<max)
                            buf[bufpos++]=(char)ch;
                     }
                     break;
            }

    /*
     * Done. Strip off trailing spaces, terminate buffer.
     */
done:
    while (bufpos && isspace(buf[bufpos-1]))
        bufpos--;
    buf[bufpos]='\0';

    /*
     * If the line begins with a digit, we parse this as a task number. If
     * it is different from TaskNumber, we clear the line buffer. If equal,
     * we remove the number. In this way, we can mix configurations of
     * different lines in the same configuration file.
     */
    if (isdigit(*buf))
        if (atoi(buf)!=TaskNumber) {
            *buf='\0';
        } else {
            char *src, *dest;
            dest=src=buf;
            src=skip_to_word (src);
            while (*src)
                *dest++ = *src++;
            *dest='\0';
        }

    /*
     * Return result.
     */
    if (*buf=='\0' && ch==EOF)
      return(NULL);
    else
      return(buf);
}


#ifdef BINKLEY_SOUNDS
char *
cfg_sound (char *c)        /* malloc & copy to ctl      */
{
    char *dest;
    int i;

    i = (int) strlen (c);       /* get length of remainder   */
    if (i < 1)                  /* must have at least 1      */
        return (NULL);          /* if not, return NULL       */
    if (!dexists (c))           /* Is the file there?        */
        return (NULL);          /* No, return NULL           */
    dest = calloc (1, (unsigned int) (i + 1));  /* allocate space for string */
    if (dest == NULL)           /* If we failed to get it,   */
        return (NULL);          /* Just return NULL          */
    (void) strcpy (dest, c);    /* copy the stuff over       */
    return (dest);              /* return the directory name */
}
#endif

void
cfg_macro (char *c)
{
    int i;
    char *p;

    i = atoi (c);
    if ((i <= 0) || (i > 9))
    {
        (void) printf ("%s %d %s\n", MSG_TXT (M_MACRO_NUMBER), i, MSG_TXT (M_OUT_OF_RANGE));
        return;
    }
    c = skip_to_word (c);
    keys[i - 1] = calloc (1, strlen (c) + 1);
    p = keys[i - 1];
    while (*c && (*c != '\n'))
    {
        if (*c == '|')
            *p++ = '\r';
        else
            *p++ = *c;
        ++c;
    }
    *p = '\0';
}

void
cfg_dial_string (char *p)
{
    char *p1;
    PN_TRNS *tpn;

    if ((*p == ';') || (*p == '\0'))
    {
        return;
    }

    tpn = (PN_TRNS *) calloc (1, sizeof (PN_TRNS));

    p1 = tpn->num;
    while ((*p) && (!isspace (*p)))
    {
        *p1++ = *p++;
    }
    *p1 = '\0';

    p = skip_blanks (p);

    p1 = tpn->pre;
    while ((*p) && (*p != '/') && (!isspace (*p)))
    {
        *p1++ = *p++;
    }
    *p1 = '\0';

    p1 = tpn->suf;
    if ((*p == '\0') || (isspace (*p)))
    {
        *p1 = '\0';
    }
    else
    {
        ++p;

        while ((*p) && (!isspace (*p)))
        {
            *p1++ = *p++;
        }
        *p1 = '\0';
    }

    tpn->len = (short) strlen (tpn->num);
    tpn->next = NULL;
    if (pn_head == NULL)
    {
        pn_head = tpn;
    }
    else
    {
        pn->next = tpn;
    }
    pn = tpn;
}

void
cfg_mdm_trans (char *p)
{
    char *p1;
    MDM_TRNS *tmm;

    if ((*p == ';') || (*p == '\0'))
    {
        return;
    }

    tmm = (MDM_TRNS *) calloc (1, sizeof (MDM_TRNS));

    tmm->mdm = (byte) atoi (p);
    p = skip_to_word (p);

    p1 = tmm->pre;
    while ((*p) && (*p != '/') && (!isspace (*p)))
    {
        *p1++ = *p++;
    }
    *p1 = '\0';

    p1 = tmm->suf;
    if ((*p == '\0') || (isspace (*p)))
    {
        *p1 = '\0';
    }
    else
    {
        ++p;

        while ((*p) && (!isspace (*p)))
        {
            *p1++ = *p++;
        }
        *p1 = '\0';
    }

    tmm->next = NULL;
    if (mm_head == NULL)
    {
        mm_head = tmm;
    }
    else
    {
        mm->next = tmm;
    }
    mm = tmm;
}

void
cfg_domain (char *p)
{
    char *p1, *p2;
    int i;

    p = strtok (p, "; \t\r\n");
    if (p != NULL)
    {
        /* Have the domain name, now get the abbreviated name */
        p2 = strtok (NULL, "; \t\r\n");
        if (strlen (p2) > 8)
            *(p2 + 8) = '\0';
        if (p2 != NULL)
        {
            /* Have the domain name and abbreviated, now get the list name */
            p1 = strtok (NULL, "; \t\r\n");
            if (p1 == NULL)
                p1 = p2;

            /* Have the list name and the domain name, now use them */
            for (i = 0; domain_name[i] != NULL; i++)
            {
                if (strnicmp (domain_name[i], p, strlen (domain_name[i])) == 0)
                {
                    domain_nodelist[i] = strdup (p1);
                    domain_abbrev[i] = strdup (p2);
                    break;
                }
            }
            if ((domain_name[i] == NULL) && (i < 49))
            {
                domain_name[i] = strdup (p);
                domain_abbrev[i] = strdup (p2);
                domain_nodelist[i] = strdup (p1);
            }
        }
    }
}

void
cfg_domain_kludge (char *p)
{
    word i;

    if (isdigit (*p))
    {
        if (num_domain_kludge == size_domain_kludge)
        {
            size_domain_kludge += 16;
            if (!num_domain_kludge)
                domainkludge = (DOMAINKLUDGE *) malloc (size_domain_kludge * sizeof (DOMAINKLUDGE));
            else
                domainkludge = (DOMAINKLUDGE *) realloc ((void *) domainkludge, size_domain_kludge * sizeof (DOMAINKLUDGE));
        }
        domainkludge[num_domain_kludge].zone = (word) atoi (p);
        p = skip_to_blank (p);
    }
    else
        return;

    p = skip_blanks (p);
    p = strtok (p, "; \t\r\n");
    if (p != NULL)
    {
        for (i = 0; domain_name[i] != NULL; i++)
        {
            if (strnicmp (domain_name[i], p, strlen (domain_name[i])) == 0)
            {
                domainkludge[num_domain_kludge++].domain = i;
                break;
            }
        }
    }
    return;
}

void
cfg_Janus (char *p)
{
    J_TYPES *tjt;
    char *p1;

    if ((*p == ';') || (*p == '\0'))
    {
        return;
    }

    tjt = (J_TYPES *) calloc (1, sizeof (J_TYPES));

    p1 = tjt->j_match;
    while ((*p) && (!isspace (*p)))
    {
        *p1++ = *p++;
    }
    *p1 = '\0';

    tjt->next = NULL;

    if (j_top == NULL)
    {
        j_top = tjt;
    }
    else
    {
        j_next->next = tjt;
    }
    j_next = tjt;
}

void
cfg_alt_num (char *c)
{
    char *n1, *n2, *p;
    P_ANUMS a1, a2;

    /* Primary starts at first non-blank, ends at blank */

    p = skip_to_blank (c);
    *p++ = '\0';

    n1 = ctl_string (c);
    if (n1 == NULL)
    {
        printf (MSG_TXT (M_ALTNUM_NO_PRIMARY));
        return;
    }

    c = skip_blanks (p);
    n2 = ctl_string (c);
    if (n2 == NULL)
    {
        printf (MSG_TXT (M_ALTNUM_NO_ALTERNATE));
        free (n1);
        return;
    }

    a1 = calloc (1, sizeof(ANUMS));

    a1->num = n1;
    a1->alt = n2;

    if (anum_head == NULL)
    {
        anum_head = a1;
        return;
    }
    else
    { 
        /* we need to maintain the order they were entered */

        a2 = anum_head;
        while (a2->next != NULL)
            a2 = a2->next;
        a2->next = a1;
    }
}

