/* Include this file before any other includes or defines! */
#include "includes.h"

/****************************************************************************/
/*  local static data                                                       */
/****************************************************************************/
static REGIONP callwinsave;
static REGIONP chatwinl, chatwinr;

static char *chatstart = "\007\007 * Chat mode start\r\n";
static char *chatend   = "\007\007\r\n * Chat mode end\r\n";
static char *chattime  = "\007\007\r\n * Chat mode timeout\r\n";

static word     chatfill;
static long     lasttimer;


/****************************************************************************/
/*  local prototypes                                                        */
/****************************************************************************/

/****************************************************************************/
/*  functions                                                               */
/****************************************************************************/

#if 0

#ifdef DOS
#define CHATBUFSIZE 256
#else
#define CHATBUFSIZE 4096
#endif

static char buffer1[CHATBUFSIZE], buffer2[CHATBUFSIZE]; /* chat buffers */
static int bufpos=0;

void chat_prepare(void){
   bufpos=0;
}

#define CHATDELAY    30  /* how many 10ms units min. delay between packets */

int chat_putc(int c){
   static long flushtimer;
   int ret=0;
   
   if(!nohydrachat){
      if(c>=0){
        if(bufpos>0 && (c=='\n' || c==' ')) /* if this is not the 1st char and we */
          flushtimer=timerset(1);         /* have a blank/newline, don't wait long*/
        else                              /* do not immediately send char, but    */ 
          flushtimer=timerset(CHATDELAY); /* wait for some others maybe following */
        if(bufpos<CHATBUFSIZE){
          buffer1[bufpos++] = (char) c;
          ret=0;
        }else
          ret=-1;
      }
      
      if(bufpos && hydra_devfree() && (timeup(flushtimer) || (bufpos>CHATBUFSIZE/2))){
         strncpy(buffer2, buffer1, bufpos);
         hydra_devsend("CON",buffer2,bufpos);
         bufpos=0;
      }
   }
   return ret;
}

void chat_puts(char *s){
   if(!nohydrachat)
      while(*s)
         chat_putc(*s++);
}

#endif /* 0 */


void chat_prepare(void){
    chatfill  = 0;
    chattimer = -1L;
    lasttimer = 0L;
}

void chat_init(void)
{
    short rows, rows_call, rows_chl, rows_chr;
    char junk[256];
    
    if (!nohydrachat && un_attended && fullscreen && !do_chat)
    {
       hydra_devsend("CON",(byte *) chatstart,(word) strlen(chatstart));
       
       callwinsave = callwin;

       rows = SB_ROWS-TOP_PARTS_ROWS-BOTTOM_PARTS_ROWS;
       if(rows<21)
          rows_call = 4;
       else
          rows_call = 7;
       rows -= rows_call;
       rows_chl = rows/2;
       rows_chr = rows-rows_chl;

       sprintf(junk, MSG_TXT (M_CHAT_WINLOCAL), sysop);

       chatwinl = sb_new_win (TOP_PARTS_ROWS, 0, rows_chl, (short) SB_COLS,
                              ' ', colors.call,
                              boxtype, colors.frames,
                              junk);
       chatwinl->wflags |= SB_SCROLL;
       
       sprintf(junk, MSG_TXT (M_CHAT_WINREMOTE), remote_sysop);
       
       chatwinr = sb_new_win (TOP_PARTS_ROWS+rows_chl, 0, rows_chr, (short) SB_COLS,
                              ' ', colors.call,
                              boxtype, colors.frames,
                              junk);
       chatwinr->wflags |= SB_SCROLL;
       
       callwin = sb_new_win ((short) (TOP_PARTS_ROWS+rows_chl+rows_chr), 0, rows_call, (short) SB_COLS,
                             ' ', colors.call,
                             boxtype, colors.frames,
                             MSG_TXT (M_RECENT_ACTIVITY));
       callwin->linesize  = callwinsave->linesize;
       callwin->buffer    = callwinsave->buffer;
       callwin->endbuff   = callwinsave->endbuff;
       callwin->lastline  = callwinsave->lastline;
       callwin->lastshown = callwinsave->lastshown;
       redraw_callwin();
       sb_dirty();
       sb_show();
       do_chat = 1;
    }
}

void chat_put_remote(char *s){
   char buf[150];
   char *d;
   int i;
   if(!nohydrachat){
      if(chatwinr){
         sb_puts(chatwinr, s);
         elapse_time(1);
      }else{
         while(*s){
            strcpy(buf,"*HCON: ");
            i=0; d=buf+7;
            while(*s && (*s!='\n') && (i<132) ){  /* break up into lines */
               *d++=*s++;
               i++;
            }
            if(*s=='\n') s++;
            *d='\0';
            status_line(buf);
         }
      }
   }
}

void chat_exit(void)
{
    if (!nohydrachat && un_attended && fullscreen && do_chat)
    {
       hydra_devsend("CON",(byte *) chatend,(word) strlen(chatend));
       sb_puts(chatwinl, &chatend[2]);
       do_chat=0;
       callwinsave->linesize  = callwin->linesize;
       callwinsave->buffer    = callwin->buffer;
       callwinsave->endbuff   = callwin->endbuff;
       callwinsave->lastline  = callwin->lastline;
       callwinsave->lastshown = callwin->lastshown;
       callwin = callwinsave;
       sb_filla(callwin, colors.call);
       sb_box(callwin, boxtype, colors.frames);
       redraw_callwin();
       sb_dirty();
       elapse_time(1);
       chatwinr=chatwinl=NULL;
    }
}

#if 0
int chat_get_key(void)
{
   int ret=0;
   char c;

   while (KEYPRESS ())
   {
      c = READKB ();
      ret=0;
      if(!do_chat){
         switch(c){
            case 27:
               while (KEYPRESS ())
                 (void) READKB ();
               ret=27;
               break;
            case '!':
               chat_init();
               break;
         }
      }else{ /* do_chat */
         switch(c){
            case 27:
               chat_exit();
               break;
            case '\r':
               c='\n';
               /* fall through into default */
            default:
               sb_putc(chatwinl, c);
               chat_putc(c);
               break;
         }
      }
   }
   chat_putc(-1); /* if there's something old to send, send it! */
   return ret;
}
#endif

BOOL keyabort (void)
{
#define CHATLEN 256
   static byte chatbuf1[CHATLEN + 5],
               chatbuf2[CHATLEN + 5],
               *curbuf = chatbuf1;
   static BOOL warned = FALSE;
   BOOL     esc = FALSE;
   char     *p;
   word     c;

   if (chattimer > 0L) {
      if (unix_time (NULL) > chattimer) {
         chattimer = lasttimer = 0L;
         hydra_devsend("CON",(byte *) chattime,(word) strlen(chattime));
         sb_puts(chatwinl, &chattime[2]);
         sb_show();
         chat_exit();
      }
      else if ((unix_time (NULL) + 10L) > chattimer && !warned) {
         sb_puts(chatwinl, "\007\r\n * Warning: chat mode timeout in 10 seconds\r\n");
         elapse_time(1);
         warned = TRUE;
      }
   }
   else if (chattimer != lasttimer) {
      if (chattimer ==  0L) {
         if (gong_allowed) p = " * Remote has chat facility (bell enabled)\n";
         else              p = " * Remote has chat facility (bell disabled)\n"; 
         hydra_devsend("CON",(byte *) p,(word) strlen(p));
         /* commented out ...
         status_line(" * Hydra session in progress, chat facility now available");
         */
      }
      /* commented out ...
      else if (chattimer == -1L)
         status_line(" * Hydra session in init state, can't chat yet\r\n");
      else if (chattimer == -2L)
         status_line(" * Remote has no chat facility available\r\n");
      else if (chattimer == -3L) 
         status_line(" * Hydra session in exit state, can't chat anymore\r\n");
      */
      lasttimer = chattimer;
   }

   while (KEYPRESS()) {
      switch (c = (word) get_key()) {
         case ESC:
           esc = TRUE;
           break;

         case 0x2e00:
           if (chattimer == 0L) {
              chat_init();
              sb_puts(chatwinl, &chatstart[2]);
              elapse_time(1);
              chattimer = lasttimer = unix_time (NULL) + CHAT_TIMEOUT;
           }
           else if (chattimer > 0L) {
              chattimer = lasttimer = 0L;
              chat_exit();
           }
           else
              sb_putc(chatwinl, '\007');
           break;

         default:
           if (c < ' ' || c == 127 || c == 255) /*AGL:23apr93*/
              break;

         case '\r':
         case '\a':
         case '\b':
           if (chattimer <= 0L)
              break;

           chattimer = unix_time (NULL) + CHAT_TIMEOUT;
           warned = FALSE;

           if (chatfill >= CHATLEN)
              sb_putc(chatwinl, '\007');
           else {
              switch (c) {
                case '\r':
                  curbuf[chatfill++] = '\n';
                  sb_putc(chatwinl, '\n');
                  break;

                case '\b':
                  if (chatfill > 0 && curbuf[chatfill - 1] != '\n')
                     chatfill--;
                  else {
                     curbuf[chatfill++] = '\b';
                     curbuf[chatfill++] = ' ';
                     curbuf[chatfill++] = '\b';
                  }
                  sb_puts(chatwinl, "\b \b");
                  break;

                default:
                  curbuf[chatfill++] = (byte) c;
                  if (c != 7)
                     sb_putc(chatwinl, c);
                  break;
              }
           }
           break;
      }
   }

   if (chatfill > 0){
     if(hydra_devsend("CON",curbuf,chatfill)) {
       curbuf = (curbuf == chatbuf1) ? chatbuf2 : chatbuf1;
       chatfill = 0;
     }
     elapse_time(1);
   }

   return (esc);
}/*keyabort()*/

