/* These are some routines and data used by BT and BTCTL in common */


/* Include this file before any other includes or defines! */

#include "includes.h"

#ifdef __BORLANDC__
extern unsigned _stklen=32768U; /* jl 960726:  set borland c stack space */
#endif

short pvtnet = -1;

char *
fancy_str (char *string)
{
    register int flag = 0;
    char *s;

    s = string;

    while (*string)
    {
        if (isalpha (*string))  /* If alphabetic,     */
        {
            if (flag)           /* already saw one?   */
                *string = (char) tolower (*string); /* Yes, lowercase it  */
            else
            {
                flag = 1;       /* first one, flag it */
                *string = (char) toupper (*string); /* Uppercase it       */
            }
        }
        else                    /* if not alphabetic  */
            flag = 0;           /* reset alpha flag   */
        string++;
    }

    return (s);
}

char *
skip_blanks (char *string)
{
    while (*string && isspace (*string))
        ++string;
    return (string);
}

char *
skip_to_blank (char *string)
{
    while (*string && (!isspace (*string)))
        ++string;
    return (string);
}

char *
skip_to_word (char *string)
{
    string=skip_to_blank(string);
    return skip_blanks(string);
}

int 
parse (char *input, struct parse_list far list[])
{
    int i, j;
    char *p;
    char far *q;

    /* MMP 960420 Before we only required a listed string to match the
     * prefix of input. Therefore "Task" could also be specified as
     * "TaskNumber", and "Log" as "LogFile" etc. This may sound nice,
     * but it also meant that we could not have two keywords where the
     * first was a prefix of the other. And we could not set environment
     * variables on the command line where the name of the variable also
     * was the name of a keyword (Task)
     */

    for (i = 0; list[i].p_length; i++)
    {
        p = input;
        q = list[i].p_string;
        for (j = list[i].p_length; j > 0; j--, p++, q++)
        {
            if (tolower (*p) != tolower (*q))
                break;
        }
        if (j == 0 &&
            (*p=='\0' ||             /* MMP 960420 / TJW 960504 */ 
             strchr(" \n\r\t!\"$%&/()=?'+*~#-_.:,;<>|",*p)!=NULL ||
             *--p=='['   /* TJW 960830 special treatment for [expression] */             
            )
           )
          return (++i);
    }
    return (-1);
}


char *
ctl_string (char *source)       /* malloc & copy to ctl      */
{
    char *dest, *c;
    int i;

    c = skip_blanks (source);   /* get over the blanks       */
    i = (int) strlen (c);       /* get length of remainder   */
    if (i < 1)                  /* must have at least 1      */
        return (NULL);          /* if not, return NULL       */
    dest = calloc (1, (unsigned int) (i+2));    /* allocate space for string */
                                               /* allocate 1 byte more for \ */
    if (dest == NULL)           /* If we failed to get it,   */
        return (NULL);          /* Just return NULL          */
    (void) strcpy (dest, c);    /* copy the stuff over       */
    return (dest);              /* and return the address    */
}

char *
ctl_slash_string (char *source) /* malloc & copy to ctl      */
{
    char *dest;
    int i;
    struct stat buffer;

    dest=ctl_string(source); /* TJW 960414 replaced redundant code by */
    if(dest==NULL)           /*            function call to above fn  */
      return NULL;

    (void) delete_backslash (dest); /* get rid of trailing stuff */
    /* Check to see if the directory exists */
    i = stat (dest, &buffer);
    if (i || (!(buffer.st_mode & S_IFDIR)))
    {
        (void) printf ("Directory '%s' does not exist!\n", dest);
        (void) printf ("This program may fail to execute properly because of this!\n");
        free (dest);
        return (NULL);
    }
    (void) add_backslash (dest);/* add the backslash         */
    return (dest);              /* return the directory name */
}

char *
add_backslash (char *str)
{
    char *p;

    p = str + strlen (str) - 1;

    if (p >= str)
    {
        /* Strip off the trailing blanks */
        while ((p >= str) && (isspace (*p)))
        {
            *p = '\0';
            --p;
        }

        /* Put a backslash if there isn't one */
        if ((p >= str) && (*p != '\\') && (*p != '/'))
        {
            *(++p) = '\\';
            *(++p) = '\0';
        }
    }

    return (fancy_str (str));
}

char *
delete_backslash (char *str)
{
    char *p;

    p = str + strlen (str) - 1;

    if (p >= str)
    {
        /* Strip off the trailing blanks */
        while ((p >= str) && (isspace (*p)))
        {
            *p = '\0';
            --p;
        }

        /* Get rid of backslash if there is one */
        if ((p >= str) && ((*p == '\\') || (*p == '/')))
        {
            if ((p > str) && (*(p - 1) != ':')) /* Don't delete on root */
                *p = '\0';
        }
    }

    return (fancy_str (str));
}
