#ifdef DEBUG

#include <stdio.h>
#include <stdlib.h>

FILE *share_fopen (char *fname, char *mode, int shflag);

int allocs=0, filesopen=0;

static FILE *debug_files_log, *debug_mem_log;

void debug_files_log_open(int TaskNumber){
  static char fname[20]="";
  sprintf(fname,"dbgfiles.l%02x", TaskNumber);
  debug_files_log=fopen(fname,"a");
}

void debug_files_log_close(void){
  fclose(debug_files_log);
}

void debug_mem_log_open(int TaskNumber){
  static char fname[20]="";
  sprintf(fname,"dbgalloc.l%02x", TaskNumber);
  debug_mem_log=fopen(fname,"a");
}

void debug_mem_log_close(void){
  fclose(debug_mem_log);
}

FILE *debug_share_fopen (char *fname, char *mode, int shflag, int TaskNumber){
  FILE *fp;
  fp=share_fopen(fname,mode,shflag);
  if(fp)
    filesopen++;
  debug_files_log_open(TaskNumber);
  fprintf(debug_files_log,"%08p %8d %-8s %-40s %-4s %4x\n", fp, filesopen, "sh_fopen", fname, mode, shflag);
  debug_files_log_close();  
  return fp;
}

FILE *debug_fopen(char *fname, char *mode, int TaskNumber){
  FILE *fp;
  fp=fopen(fname,mode);
  if(fp)
    filesopen++;
  debug_files_log_open(TaskNumber);
  fprintf(debug_files_log,"%08p %8d %-8s %-40s %-4s %4s\n", fp, filesopen, "fopen", fname, mode, "");
  debug_files_log_close();  
  return fp;
}

int debug_fclose(FILE *fp, int TaskNumber){
  int rc;
  rc=fclose(fp);
  if(!rc && fp)
    filesopen--;
  debug_files_log_open(TaskNumber);
  fprintf(debug_files_log,"%08p %8d %-8s %-40s %-4s %4s %d\n", fp, filesopen, "fclose", "", "", "", rc);
  debug_files_log_close();  
  return rc;
}

void *debug_calloc(size_t x, size_t y, int TaskNumber, char *cfname, int cline){
  void *p;
  p=calloc(x,y);
  if(p)
    allocs++;
  debug_mem_log_open(TaskNumber);
  fprintf(debug_mem_log,"%08p %8d %-8s %8d %8d %d %s\n", p, allocs, "calloc", x, y, cline, cfname);
  debug_mem_log_close();  
  return p;
}

void *debug_malloc(size_t amount, int TaskNumber, char *cfname, int cline){
  void *p;
  p=malloc(amount);
  if(p)
    allocs++;
  debug_mem_log_open(TaskNumber);
  fprintf(debug_mem_log,"%08p %8d %-8s %8d %8s %d %s\n", p, allocs, "malloc", amount, "", cline, cfname);
  debug_mem_log_close();  
  return p;
}

void debug_free(void *p, int TaskNumber, char *cfname, int cline){
  free(p);
  if(p)
    allocs--;
  debug_mem_log_open(TaskNumber);
  fprintf(debug_mem_log,"%08p %8d %-8s %8s %8s %d %s\n", p, allocs, "free", "", "", cline, cfname);
  debug_mem_log_close();  
}

#endif
