/*                                                                          */
/*              This module was written by Carsten Ellermann                */
/*              Some modifications done by Thomas Waldmann                  */
/*                                                                          */
/*                      BinkleyTerm FreePoll Function                       */
/*                                                                          */

/* Include this file before any other includes or defines! */

#include "includes.h"

static long netsize (ADDRP p, char *ext);
static long filesize (ADDRP p, char *ext);
static int callacceptaka(int i, time_t t);

int freepoll (char * number)
{
    int i;                                  /* counter */
    int found = 0;
    int accept = -1;
    time_t t;
    
    t = unix_time (NULL);

    for(i=0; (i < num_fpaddrs) && (fpaddr[i].Number != NULL); i++)
    {
        if (strstr(number, fpaddr[i].Number))   /* compare the numbers */
        {
            if(accept<0)                        /* no boolean op on first */
              accept=callacceptaka(i,t);
            else{
              if(fpaddr[i].and_op)
                accept=callacceptaka(i,t) && accept;
              else
                accept=callacceptaka(i,t) || accept;
            }
            found = 1;
        }
    }

    if (!found)
        return (0);                             /* user is not listed */

    status_line(accept ? MSG_TXT(M_ACCEPT) : MSG_TXT(M_REJECT));
    return (!accept);
}

void change_call_to_normal (ADDRP addr)
{
    char iname[PATHLEN], oname[PATHLEN], tname[PATHLEN];
    char *HoldName, *result, *temp_buffer;
    struct FILEINFO fileinfo;
    int ihandle, ohandle;
    int check;
    int written = 1;
    int buff_size;

    status_line (MSG_TXT (M_CALL_REJECT));
    /* first append/rename ?LO */

    HoldName = HoldAreaNameMunge (addr);
    (void) sprintf (tname, "%s%s.?LO", HoldName,
        Hex_Addr_Str (addr));
    (void) sprintf (oname, "%s%s.%cLO", HoldName,
        Hex_Addr_Str (addr), (char) chg_mail);

    if (!dfind (&fileinfo, tname, 0))
    {
        do
        {
            buff_size = (int) fileinfo.size;
            strcpy (iname, fileinfo.name);
            result = strchr (iname, '.');
            result++;
            if (*result != 'F')
            {
                if (addr->Point != 0)
                {
                    (void) sprintf (iname, "%s%04x%04x.PNT\\%s",
                        HoldName, addr->Net, addr->Node, fileinfo.name);
                }
                else
                {
                    (void) sprintf (iname, "%s%s", HoldName, fileinfo.name);
                }
                if (rename (iname, oname) != 0)
                {
                    temp_buffer = (char *) malloc (buff_size + 1);
                    ihandle = open (iname, O_RDONLY | O_BINARY);
                    ohandle = open (oname, O_APPEND | O_BINARY | O_WRONLY);
                    do
                    {
                        check = read (ihandle, temp_buffer, buff_size);
                        if (check)
                            written = write (ohandle, temp_buffer, check);
                    }
                    while (check > 0);
                    close (ihandle);
                    close (ohandle);
                    if (written > 0)
                        unlink (iname);
                    free (temp_buffer);
                }
            }
        }
        while (!dfind (&fileinfo, NULL, 1));
        (void) (dfind (&fileinfo, NULL, 2));
    }

    /* now rename ?UT */

    if ((char) chg_mail == 'F')
        chg_mail = (int) 'O';

    (void) sprintf (tname, "%s%s.?UT", HoldName,
        Hex_Addr_Str (addr));
    (void) sprintf (oname, "%s%s.%cUT", HoldName,
        Hex_Addr_Str (addr), (char) chg_mail);

    if ((char) chg_mail == 'O')
        chg_mail = (int) 'F';

    if (!dfind (&fileinfo, tname, 0))
    {
        do
        {
            buff_size = (int) fileinfo.size;
            strcpy (iname, fileinfo.name);
            result = strchr (iname, '.');
            result++;
            if (*result != 'O')
            {
                if (addr->Point != 0)
                {
                    (void) sprintf (iname, "%s%04x%04x.PNT\\%s",
                        HoldName, addr->Net, addr->Node, fileinfo.name);
                }
                else
                {
                    (void) sprintf (iname, "%s%s", HoldName, fileinfo.name);
                }
                if ((rename (iname, oname)) != 0)
                {
                    if (buff_size > 60)
                        append_pkt (NULL, buff_size, iname, oname);
                }               /* do append if rename failed */
            }
        }
        while (!dfind (&fileinfo, NULL, 1));
        (void) (dfind (&fileinfo, NULL, 2));
    }                           /* end of ?UT handling */
    xmit_reset (0, 1);
    return;
}

static int callacceptaka(int i, time_t t)
{
    char trx_file[PATHLEN];
    FILE *trx;
    struct stat stat_buffer;
    long size;

    (void) sprintf (trx_file, "%s%s.TRX",
        HoldAreaNameMunge (&fpaddr[i].addr),
        Hex_Addr_Str (&fpaddr[i].addr));

    if (unix_stat (trx_file, &stat_buffer))              /* get info from file */
    {
        status_line(MSG_TXT(M_ACCEPT1STCALLEVER),  /* TJW 960530 */
                  Full_Addr_Str(&fpaddr[i].addr),
               fpaddr[i].and_op ? "AND" : "OR" );
        if ((trx = fopen(trx_file, write_binary)) == NULL) /* create first TRX-file */

            status_line(MSG_TXT(M_WRITEPROTECTED),
                                        trx_file);
        else
            fclose (trx);                          /* close new file */
        return 1;                                  /* accept call */
    }

    if ((trx = fopen(trx_file, write_binary)) == NULL)     /* touch file */
    {
        status_line(MSG_TXT(M_WRITEPROTECTED),
                                       trx_file);
        return (1);
    }
    fclose (trx);

    if ((long)(t - stat_buffer.st_mtime) < fpaddr[i].Delta_T)
    {                                 /* caller rings twice in delta_t time */ 
        status_line(MSG_TXT(M_ACCEPT2NDCALL),   /* TJW 960505 added addr */ 
                    Full_Addr_Str(&fpaddr[i].addr),
                    (long)fpaddr[i].Delta_T,
                    fpaddr[i].and_op ? "AND" : "OR" );
        return 1;                     /* accept call */
    }

    size = totalsize(&fpaddr[i].addr);      /* get size of all for a user */

    if (size <= fpaddr[i].Size*1024L)               /* netsize <= Limit? */
    {
        status_line(MSG_TXT(M_REJECTLOWVOLUMECALL), /* TJW 960505 added addr */ 
                    Full_Addr_Str(&fpaddr[i].addr),
                    size/1024, fpaddr[i].Size,
                    fpaddr[i].and_op ? "AND" : "OR" );
        return 0;                               /* reject call */
    }

    status_line(MSG_TXT(M_ACCEPTHIGHVOLUMECALL), /* TJW 960505 added addr */
                Full_Addr_Str(&fpaddr[i].addr),
                size/1024, fpaddr[i].Size,
                fpaddr[i].and_op ? "AND" : "OR" );
    return 1;                                  /* accept call */
}

long totalsize(ADDRP addr){               /* get size of all for a user */ 
   long size=0L;
   
   size += filesize(addr, "CUT");         
   size += filesize(addr, "HUT");
   size += filesize(addr, "DUT");
   size += filesize(addr, "OUT");

   size += netsize(addr, "CLO");
   size += netsize(addr, "HLO");
   size += netsize(addr, "DLO");
   size += netsize(addr, "FLO");
   return size;
}
   
static long filesize (ADDRP p, char *ext)
{
    struct stat stbuf;
    char net_path[PATHLEN];

    (void) sprintf (net_path, "%s%s.%s",
        HoldAreaNameMunge (p),
        Hex_Addr_Str (p),
        ext);

    if (unix_stat (net_path, &stbuf))    /* file exist? */
        return (0L);
    else
        return (stbuf.st_size);
}

static long netsize (ADDRP p, char *ext)
{
    struct stat stbuf;
    char net_path[PATHLEN];
    char *ptr;
    FILE *temp;
    long accum = 0L;
    char *q;

    /* Append the ARCmail file name to the path line  */

    (void) sprintf (net_path, "%s%s.%s",
        HoldAreaNameMunge (p),
        Hex_Addr_Str (p),
        ext);

    temp = fopen (net_path, read_binary);

    if (temp == (FILE *) NULL)
        return (accum);

    while (!feof (temp))
    {
/*
 *      Make sure of a nice zero there if we're at an undetected EOF.
 *      Then try to read a line from the file.
 */
        net_path[0] = '\0';
        (void) fgets (net_path, 79, temp);
/*
 *      Clean up anything we don't want to see (blanks, tabs, CR, etc)
 */
        for (q = ptr = net_path; *q; q++)
            if (*q <= ' ')
                *q = '\0';
/*
 *      File disposition commands should be skipped over to get to
 *      actual filenames.
 */
        if ((*ptr == TRUNC_AFTER) ||
            (*ptr == DELETE_AFTER) ||
            (*ptr == SHOW_DELETE_AFTER) ||
            (*ptr == NOTHING_AFTER))
            ptr++;
/*
 *      Now -- if what's left starts with a semicolon, it's a comment.
 *      If it starts with a tilde, the file has already been sent. If
 *      what we see is a zero, there's nothing on the line. In any one
 *      of these cases, we should skip this line.
 */
        if ((*ptr == '\0') ||
            (*ptr == ';') ||
            (*ptr == '~'))
            continue;
/*
 *      Get the file size by doing a stat call. If it's not there
 *      then we obviously need not add any size in. If the file is
 *      arcmail, figure it into the "oldest" calculation.
 */
        if (unix_stat (ptr, &stbuf)) /* file exist? */
            continue;
        else
        {
            accum += stbuf.st_size;
        }
    }
    (void) fclose (temp);
    return (accum);
}

