/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                                                          */
/*      ------------         Bit-Bucket Software, Co.                       */
/*      \ 10001101 /         Writers and Distributors of                    */
/*       \ 011110 /          Freely Available<tm> Software.                 */
/*        \ 1011 /                                                          */
/*         ------                                                           */
/*                                                                          */
/*              (C) Copyright 1987-96, Bit Bucket Software Co.              */
/*                                                                          */
/*                 This module was written by Bob Hartman                   */
/*                                                                          */
/*                   BinkleyTerm Mail Control Routines                      */
/*                                                                          */
/*                                                                          */
/*    For complete  details  of the licensing restrictions, please refer    */
/*    to the License  agreement,  which  is published in its entirety in    */
/*    the MAKEFILE and BT.C, and also contained in the file LICENSE.260.    */
/*                                                                          */
/*    USE  OF THIS FILE IS SUBJECT TO THE  RESTRICTIONS CONTAINED IN THE    */
/*    BINKLEYTERM  LICENSING  AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF    */
/*    THIS  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES,  OR IF YOU DO    */
/*    NOT HAVE THESE FILES,  YOU  SHOULD  IMMEDIATELY CONTACT BIT BUCKET    */
/*    SOFTWARE CO.  AT ONE OF THE  ADDRESSES  LISTED BELOW.  IN NO EVENT    */
/*    SHOULD YOU  PROCEED TO USE THIS FILE  WITHOUT HAVING  ACCEPTED THE    */
/*    TERMS  OF  THE  BINKLEYTERM  LICENSING  AGREEMENT,  OR  SUCH OTHER    */
/*    AGREEMENT AS YOU ARE ABLE TO REACH WITH BIT BUCKET SOFTWARE, CO.      */
/*                                                                          */
/*                                                                          */
/* You can contact Bit Bucket Software Co. at any one of the following      */
/* addresses:                                                               */
/*                                                                          */
/* Bit Bucket Software Co.        FidoNet  1:104/501, 1:343/491             */
/* P.O. Box 460398                AlterNet 7:42/1491                        */
/* Aurora, CO 80046               BBS-Net  86:2030/1                        */
/*                                Internet f491.n343.z1.fidonet.org         */
/*                                                                          */
/* Please feel free to contact us at any time to share your comments about  */
/* our software and/or licensing policies.                                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/

/* Include this file before any other includes or defines! */

#include "includes.h"

static int LOCALFUNC xmit_install (MAILP, ADDRP);
static MAILP LOCALFUNC xmit_find (MAILP, ADDRP);
static int LOCALFUNC any_mail (MAILP);
static void LOCALFUNC do_xmit_line (char *, MAILP);
static void LOCALFUNC xmit_sort (void);
static void LOCALFUNC addr_sort (void);
static void LOCALFUNC kill_one (ADDRP address);
static void LOCALFUNC xmit_delete_one (void);
char *mail_status_chars (unsigned int);

static char *HoldFmtStr = "%-17.17s %3d %4.4s %3d %5s%c";
static char *HoldNoSize = "%-30.30s %5s%c";

void 
xmit_sameplace ()
{
    MAILP p, p1;

    /* Find the guy we just gave mail to */

    p = find_mail (&remote_addr);
    remote_addr.Zone = remote_addr.Net = remote_addr.Node = remote_addr.Point = 0;
    remote_addr.Domain = NULL;
    if (p == NULL)
    {
        /* He is not there */
        return;
    }

    /* Save our current pointer */

    p1 = next_mail;
    if (p != next_mail)
    {
        /* If it is not the one we just gave mail to, save ptr and delete */
        next_mail = p;
        xmit_delete ();
        next_mail = p1;
    }
    else
    {
        /* It was the guy at the head of the list, so just delete him */
        xmit_delete ();
    }

    /* If we came in with a null, leave with a null */

    if (p1 == NULL)
        next_mail = NULL;

    return;
}

MAILP 
find_mail (ADDRP address)
{
    MAILP p;

    p = mail_top;
    while (p != NULL) /* TJW 961001 Peter Prenting reported SYS3175 here */
    {                 /* vvvvvvvvvvvvvvvvvv------ how does it come ?     */
        if ((no_zones || (p->mail_addr.Zone == address->Zone)) &&
            (p->mail_addr.Net == address->Net) &&
            (p->mail_addr.Node == address->Node) &&
            (p->mail_addr.Point == address->Point) &&
            ((p->mail_addr.Domain == address->Domain) ||
                ((p->mail_addr.Domain == my_addr.Domain) &&
                    (address->Domain == NULL))))
            break;
        p = p->next;
    }

    return (p);
}

static void LOCALFUNC
xmit_setflags (MAILP p1)
{
    MDM_TRNS *m;                    /* MB 93-11-27 */
    long totcost=0, dummy;
    word cost=0;
    
    if (!nodefind (&(p1->mail_addr), 0))
    {
        p1->mailtypes |= MAIL_UNKNOWN;
        status_line(">set_flags: %s : %s",
                    Full_Addr_Str(&p1->mail_addr),
                    "system unlisted, no send."
        );
        return;
    }
    
    /* if modem type is undialable, don't call */   /* MB 93-11-27 */

    m = mm_head;
    while(m)
    {
        if (m->mdm == newnodedes.ModemType)
        {
            if (!(*(m->pre) || *(m->suf))){
                status_line(">set_flags: %s : %s",
                            Full_Addr_Str(&p1->mail_addr),
                            "undialable modem type, no send."
                );
                return;
            }
            else
                break;
        }
        m = m->next;
    }

    /* If there's no event, set mail to 'go' */

    if (cur_event<0)
    {
        p1->mailtypes &= ~MAIL_COST;
        p1->mailtypes &= ~MAIL_QSMALL;
        p1->mailtypes |= MAIL_WILLGO;
        status_line(">set_flags: %s : %s",
                    Full_Addr_Str(&p1->mail_addr),
                    "no event setup, send!"
        );
        return;
    }

    /* If it is a crash only event and we have crashmail, set it to go */

    if ( e_ptrs[cur_event].behavior&MAT_HIPRICM &&
         p1->mailtypes&MAIL_CRASH &&
         newnodedes.NodeFlags&B_CM )
    {
         p1->mailtypes &= ~MAIL_COST;
         p1->mailtypes &= ~MAIL_QSMALL;
         p1->mailtypes |= MAIL_WILLGO;
         status_line(">set_flags: %s : %s",
                     Full_Addr_Str(&p1->mail_addr),
                     "crash only event, crash mail, send!"
         );
         return;
    }

    /* If it is a crash only event and this wasn't crash, return */

    if (!(p1->mailtypes&MAIL_CRASH) && e_ptrs[cur_event].behavior&MAT_CM){
        status_line(">set_flags: %s : %s",
                    Full_Addr_Str(&p1->mail_addr),
                    "crash only event, no crash mail, no send."
        );
        return;
    }

    /* Is this a local only event? */

    if (e_ptrs[cur_event].behavior&MAT_LOCAL)
    {
        /*
         * calculate minimum call costs (assuming "costcps" xfer rate)
         * and get out if they are too high/low
         */

        calc_cost(1, 0, p1->callsize/costcps, &dummy, &totcost, &dummy, &cost);

        if (e_ptrs[cur_event].node_cost >= 0)
        {
            if (totcost > e_ptrs[cur_event].node_cost){
                p1->mailtypes |= MAIL_COST;
                status_line(">set_flags: %s : %s (%ld > %u)",
                            Full_Addr_Str(&p1->mail_addr),
                            "cost too high, no send.",
                            totcost, e_ptrs[cur_event].node_cost
                );
                return;
            }
        }
        else
        {
            if ( totcost < -(e_ptrs[cur_event].node_cost)){   /* <= ??? */
                p1->mailtypes |= MAIL_COST;
                status_line(">set_flags: %s : %s (%ld < %u)",
                            Full_Addr_Str(&p1->mail_addr),
                            "cost too low, no send.",
                            totcost, e_ptrs[cur_event].node_cost
                );
                return;
            }
        }
    }

    /* Is this a non-mail window event? */

    if ( (newnodelist || version7) &&
         !(e_ptrs[cur_event].behavior&MAT_NOMAIL24) )
    {
        /* If this guy can't handle crash, get out and try again */

        if (!(newnodedes.NodeFlags&B_CM)){
            status_line(">set_flags: %s : %s",
                        Full_Addr_Str(&p1->mail_addr),
                        "non-mail window, node not CM, no send."
            );
            return;
        }
    }

    /* Is this a receive only event? */

    if (e_ptrs[cur_event].behavior&MAT_NOOUT){
        status_line(">set_flags: %s : %s",
                    Full_Addr_Str(&p1->mail_addr),
                    "receive only event, no send."
        );
        return;
    }
     
    /* Is this a non-CM event? */

    if ((newnodelist || version7) &&
        e_ptrs[cur_event].behavior&MAT_NOCM &&
        newnodedes.NodeFlags&B_CM){
        status_line(">set_flags: %s : %s",
                    Full_Addr_Str(&p1->mail_addr),
                    "non-CM event, node is CM, no send."
        );
        return;
    }

    /* See if we spent too much calling him already */

    if (bad_call (&(p1->mail_addr), 0))
    {
        p1->mailtypes |= MAIL_TOOBAD;
        status_line(">set_flags: %s : %s",
                    Full_Addr_Str(&p1->mail_addr),
                    "calls to this node failed too often -> bad, no send."
        );
        return;
    }

    /* See if we have enough mail to send */

    if (!no_size && e_ptrs[cur_event].mailqsize>p1->callsize)
    {
        p1->mailtypes |= MAIL_QSMALL;
        status_line(">set_flags: %s : %s",
                    Full_Addr_Str(&p1->mail_addr),
                    "mail size is too low, no send."
        );
        return;
    }

    if (p1->mailtypes&MAIL_CRASH ||p1->mailtypes&MAIL_DIRECT || p1->mailtypes&MAIL_NORMAL ){
        p1->mailtypes &= ~MAIL_COST;
        p1->mailtypes &= ~MAIL_QSMALL;
        p1->mailtypes |= MAIL_WILLGO;
        status_line(">set_flags: %s : %s",
                    Full_Addr_Str(&p1->mail_addr),
                    "nothing special, send!"
        );
        return;
    }

    /* Don't call for "HOLD" or "REQ" stuff. */

    if ( p1->mailtypes&MAIL_HOLD || p1->mailtypes&MAIL_REQUEST ){
        status_line(">set_flags: %s : %s",
                    Full_Addr_Str(&p1->mail_addr),
                    "only hold or req mail, no send."
        );
        return;
    }

    status_line(">set_flags: %s : mailtypes=%04x, event.behaviour=%04x",
                Full_Addr_Str(&p1->mail_addr),
                p1->mailtypes,
                e_ptrs[cur_event].behavior
    );

}

static int LOCALFUNC 
xmit_install (MAILP p, ADDRP addr)
{
    MAILP p1, p2;
    int rettype;
    long sztemp;

    p2 = find_mail (addr);

    if (p2 == NULL)
    {
        /* We didn't find it in what we have already */
        p1 = p;
        p1->mail_addr = *addr;
        p1->oldest = (unsigned long) unix_time (NULL);
        rettype = 0;
    }
    else
    {
        /* We found it, so we have to make sure the higher level routine knows */
        p1 = p2;
        rettype = 1;
    }

/*
 * Get the size of the entry. If it's a FLO-type file,
 * call netsize to find out how big the stuff contained in it
 * actually is. If it's a packet, just take its size.
 *
 * Hold packets don't count.
 */

    if (!no_size)
    {
        sztemp = 0L;

        if (!strncmp (&(dta_str.name[10]), "LO", 2) ||
            !strncmp (&(dta_str.name[9]), "REQ", 3))
            p1->mailsize += (sztemp = cachenetsize (p1));

        if ((!strncmp (&(dta_str.name[10]), "UT", 2)) ||
                (!strncmp (&(dta_str.name[9]), "REQ", 3)))
        {
            struct tm tmstruc;
            time_t curr_time;

            p1->numfiles++;
            p1->mailsize += (sztemp = dta_str.size);

            curr_time = unix_time (NULL);
            tmstruc = *unix_localtime (&curr_time);  /* Structure assignment */

            tmstruc.tm_year = (int) (dta_str.time >> 25) + 80; 
            tmstruc.tm_mon  = (int) ((dta_str.time >> 21) & 0x0f) - 1; 
            tmstruc.tm_mday = (int) (dta_str.time >> 16) & 0x1f; 

            tmstruc.tm_hour = (int) (dta_str.time >> 11) & 0x1f; 
            tmstruc.tm_min =  (int) (dta_str.time >> 5) & 0x3f; 
            tmstruc.tm_sec =  (int) (dta_str.time & 0x1f) * 2;  
            /* TJW 960427 * 2 because of double-seconds ! */
                               
            curr_time = unix_mktime (&tmstruc);
            p1->oldest = min ((unsigned long)curr_time, p1->oldest);
        }

        if (dta_str.name[9] != 'H')
            p1->callsize += sztemp;
    }

    switch (dta_str.name[9])
    {
    case 'C':                   /* Crash */
        p1->mailtypes |= MAIL_CRASH;
        break;

    case 'H':                   /* Hold */
        p1->mailtypes |= MAIL_HOLD;
        break;

    case 'F':                   /* Normal */
    case 'O':
        p1->mailtypes |= MAIL_NORMAL;
        break;

    case 'D':                   /* Direct */
        p1->mailtypes |= MAIL_DIRECT;
        break;

    case 'R':                   /* Request */
        p1->mailtypes |= MAIL_REQUEST;
        break;
    }

    xmit_setflags(p1);
    
    return (rettype);
}

char
mail_stat (MAILP p)
{
    if (p->mailtypes & MAIL_UNKNOWN)
        return ('!');
    if (p->mailtypes & MAIL_TOOBAD)
        return ('x');
    if (p->mailtypes & MAIL_TRIED)
        return ('#');
    if (p->mailtypes & MAIL_WILLGO)
        return ('*');
    if (p->mailtypes & MAIL_QSMALL)
        return ('<');
    if (p->mailtypes & MAIL_COST)       /* TJW 960731 */
        return ('$');
    return ('-');
}

char *
mail_status_chars (unsigned int p)
{
    static char msc[8];  /* TJW 960616 made it local/static */
    char *q = msc;

    *q++ = (p & MAIL_HOLD)    ? 'H' : ' '; /* TJW 960422 */
    *q++ = (p & MAIL_NORMAL)  ? 'N' : ' '; /* TJW 960422 */
    *q++ = (p & MAIL_DIRECT)  ? 'D' : ' '; /* TJW 960422 */
    *q++ = (p & MAIL_REQUEST) ? 'R' : ' '; /* TJW 960422 */
    *q++ = (p & MAIL_CRASH)   ? 'C' : ' '; /* TJW 960422 */
    *q++ = '\0';

    return (msc);
}

void
xmit_window (MAILP p)
{
    short i;
    char j1[40];

    if (!fullscreen)
        return;

    sb_fillc (holdwin, ' ');

    if (p == NULL)
    {
        sb_move_puts (holdwin, 2, 4, MSG_TXT (M_NOTHING_IN_OUTBOUND));
        return;
    }

    sb_move_puts (holdwin, 0, 0, MSG_TXT (M_OUTBOUND_HEADER));
    for (i = 1; p!=NULL && i<SB_ROW_HOLD-1; p = p->next, i++)
    {
        do_xmit_line (j1, p);
        sb_move_puts (holdwin, i, 0, j1);
    }
    sb_show ();
}

static void LOCALFUNC 
do_xmit_line (char *line, MAILP p)
{
    ADDR addrT;

    if (no_size)
    {
        (void) sprintf (line, HoldNoSize,
            Full_Addr_Str (&(p->mail_addr)),
            mail_status_chars (p->mailtypes),
            mail_stat (p));
    }
    else
    {
        addrT = p->mail_addr;
        addrT.Domain = NULL;
        (void) sprintf (line, HoldFmtStr,
            Full_Addr_Str (&addrT),
            (p->numfiles > 999 ? 999 : p->numfiles),
            numdisp (p->mailsize),
            (int)((unix_time (NULL) - p->oldest) / 86400L), /* TJW 960506 "Age" */
            mail_status_chars (p->mailtypes),
            mail_stat (p));
    }
}

static void LOCALFUNC
xmit_sort ()
{
    MAILP p, p1, p2;

    addr_sort();                           /* jl 960710: sorted outbound */

    p = mail_top;

    /* Find the first that is sendable */

    while (p != NULL)
    {
        if ( (p->mailtypes & MAIL_WILLGO) &&
            !(p->mailtypes & MAIL_TOOBAD) &&
            !(p->mailtypes & MAIL_UNKNOWN))
            break;
        p = p->next;
    }

    if (p == NULL)
        return;

    /* Put the first sendable one on top */

    if (p != mail_top)
    {
        p->prev->next = p->next;
        if (p->next != NULL)
            p->next->prev = p->prev;
        p->prev = NULL;
        p->next = mail_top;
        mail_top->prev = p;
        mail_top = p;
    }

    p1 = p;
    p = p1->next;
    while (p != NULL)
    {
        if ( (p->mailtypes & MAIL_WILLGO) &&
            !(p->mailtypes & MAIL_TOOBAD) &&
            !(p->mailtypes & MAIL_UNKNOWN))
        {
            if (p->prev == p1){
                p1 = p;
                p = p->next;
            }else{
                p2 = p->next;
                p->prev->next = p->next;
                if (p->next != NULL)
                    p->next->prev = p->prev;
                p->next = p1->next;
                if (p1->next != NULL)
                    p1->next->prev = p;
                p->prev = p1;
                p1->next = p;
                p1 = p;
                p = p2;
            }
        }else
            p = p->next;
    }
}

/* jl 960710: sorted outbound, start ------------------------------------- */

static int
cmp_mail(const void *p1, const void *p2)
/* qsort function to compare two 5d addresses.  used by addr_sort() */
{
    ADDRP a1, a2;
    int i;

    a1 = &(* (MAILP *) p1)->mail_addr;
    a2 = &(* (MAILP *) p2)->mail_addr;

    if (a1->Zone == a2->Zone){  /* TJW960727 only sort Domains in same zone */
       if (a1->Domain != NULL)
       {
          if (a2->Domain != NULL)
          {
             if ((i = stricmp(a1->Domain, a2->Domain)) != 0)
                return(i);
          }
          else
             return(1);
       }
       else if (a2->Domain != NULL)
          return(-1);
    }
    
    if (a1->Zone != a2->Zone)
       return(a1->Zone > a2->Zone ? 1 : -1);

    if (a1->Net != a2->Net)
       return(a1->Net > a2->Net ? 1 : -1);

    if (a1->Node != a2->Node)
       return(a1->Node > a2->Node ? 1 : -1);

    if (a1->Point != a2->Point)
       return(a1->Point > a2->Point ? 1 : -1);

    return(0);
}

/* ----------------------------------------------------------------------- */

static void LOCALFUNC
addr_sort (void)
{
    size_t nelt;
    MAILP *qbuf, *q;
    MAILP p;

    /* no entries in list?  don't sort them */

    if (mail_top == NULL)
        return;

    /* count entries in mail_top */

    nelt = 0;
    p = mail_top;

    while (p != NULL)
    {
        nelt++;
        p = p->next;
    }

    /* allocate memory for qsort buffer */

    qbuf = malloc(nelt * sizeof(*qbuf));

    /* no mem?  don't do anything */

    if (qbuf == NULL)
        return;

    /* fill qsort buffer */

    p = mail_top;
    q = qbuf;

    while (p != NULL)
    {
        *q = p;

        p = p->next;
        q++;
    }

    /* sort the stuff */

    qsort(qbuf, nelt, sizeof(*qbuf), cmp_mail);

    /* build sorted mail_top list */

    mail_top = *qbuf;
    q = qbuf;

    while (nelt-- != 0)
    {
       p = *q;

       if (p == mail_top)
          p->prev = NULL;
       else
          p->prev = *(q - 1);

       if (nelt == 0)
          p->next = NULL;
       else
          p->next = *(q + 1);
          
       q++;
    }

    /* free qsort buffer */

    free(qbuf);
}

/* jl 960710: sorted outbound, end --------------------------------------- */

void
global_rescan(void)          /* TJW 960722 global_rescan for all lines */ 
{
   char flagfname[PATHLEN];
   FILE *fp;

   sprintf (flagfname, "%sBTRESCAN.FLG", flag_dir);
   unlink(flagfname);
   if((fp=share_fopen(flagfname, write_binary, DENY_WRITE))!=NULL)
      fclose(fp);
}

/* TJW 960616 misc BTRESCAN.BSY / .## changes */
void 
xmit_reset (int display, int forcerescan)
{
    MAILP p;
    int i, k, zone;
    char *q, *s;
    char *domain;
    char *HoldName;
    char pointspec[PATHLEN];
    ADDR tmp;
    DIRBUF zone_dir;
    DIRBUF pnt_dir;
    long longzone;

    char flgfname[PATHLEN], flg2fname[PATHLEN];
    FILE *fp;
    struct stat buf, buf2;

    DUMPMAIL dumpmail;
    time_t tnow;
  
    if(forcerescan){        /* force rescan = recreate/touch BTRESCAN.FLG */
       global_rescan();
    }
      
    sprintf (flgfname, "%sBTRESCAN.BSY", flag_dir);
    if (!unix_stat (flgfname, &buf)){                  /* BTRESCAN.BSY exists ?  */
      sprintf (flgfname, "%sBTRESCAN.%02x", flag_dir, TaskNumber); /* yes:  */
      if(unix_stat (flgfname, &buf)){       /* if BTRESCAN.xx not already exists */
        if((fp=fopen(flgfname, write_binary))!=NULL) /* create it to do re- */
          (void) fclose(fp);                         /* scan later ...      */
      }
      return;                                                 /* return now */
    }
    if((fp=fopen(flgfname, write_binary))!=NULL)   /* no: create BSY flag   */
      (void) fclose(fp);
    sprintf (flgfname, "%sBTRESCAN.%02x", flag_dir, TaskNumber);
    unlink (flgfname);                            /* delete task rescan flg */

    /* First get rid of all the old junk */

    cacheflush();
    p = mail_top;
    if (p != NULL)
    {
        while (p->next != NULL)
            p = p->next;
        while (p->prev != NULL)
        {
            p = p->prev;
            free (p->next);
        }
        if (p != NULL)
            free (p);
    }

    p = mail_top = NULL;

    tnow = unix_time (NULL);

    sprintf (flgfname, "%sBTRESCAN.DMP", flag_dir);
    if (!unix_stat (flgfname, &buf)){                  /* BTRESCAN.DMP exists ?  */ 

      sprintf (flg2fname, "%sBTRESCAN.FLG", flag_dir);
      if(!unix_stat (flg2fname, &buf2) && buf2.st_mtime > buf.st_mtime)
         goto rescan;            /* BTRESCAN.FLG is newer than BTRESCAN.DMP */

      if(buf.st_mtime + readholdtime < tnow)     /* BTRESCAN.DMP is older  */
         goto rescan;                            /* than "readholdtime"    */
              
      if((fp=share_fopen(flgfname, read_binary, DENY_NONE))!=NULL){
                                                  /* yes: read it in ...    */
         if(display){
           if (un_attended && fullscreen)
              do_ready (M_READY_READING);
           else
              status_line (MSG_TXT (M_OUTBOUND_READING)); /* Tell what we are doing */
         }
         status_line (">Reading Outbound");    /* TJW 960721 */ 

         while(!feof(fp)){
            if(fread(&dumpmail, sizeof(DUMPMAIL), 1, fp)==1){
               dumpmail.mail.mail_addr.Domain=NULL;
               dumpmail.mail.prev=dumpmail.mail.next=NULL;
               if(dumpmail.Domain[0]){
                  for(i=0;domain_name[i];i++){
                     if(!strnicmp(dumpmail.Domain, domain_abbrev[i], 8)){
                        dumpmail.mail.mail_addr.Domain=domain_name[i];
                        break;
                     }
                  }
               }
               if (p == NULL){
                  p = mail_top = (MAILP) calloc (sizeof (MAIL), 1);
                  *p = dumpmail.mail;
               }else{
                  p->next = (MAILP) calloc (sizeof (MAIL), 1);
                  *(p->next) = dumpmail.mail;
                  p->next->prev = p;
                  p = p->next;
               }

               p->mailtypes &= ~(MAIL_UNKNOWN|MAIL_QSMALL|MAIL_WILLGO|MAIL_TOOBAD);
               xmit_setflags (p);
            }
         }
         (void) fclose(fp);
      }
    }
    else
    {

rescan:

       if(display){  /* TJW 960526 */ 
         if (un_attended && fullscreen)
            do_ready (M_READY_RESCAN);
         else
            status_line (MSG_TXT (M_OUTBOUND_RESCAN)); /* Tell what we are doing */
       }
       status_line (">Rescanning Outbound");    /* TJW 960526 */ 


/*
 * Initialize domain to scan. Choose (of course) the
 * domain of our primary address.
 *
 * This domain is special in that its outbound is hold_area.
 * All the other domains use their abbreviation as the name
 * of their outbounds.
 */

       k = 0;
       domain = domain_name[0];
       (void) strcpy (pointspec, hold_area);
       q = &(pointspec[strlen (pointspec)]) - 1;
       do
       {

/*
 * Initialize scan of zones in this domain. Using findfirst/findnext,
 * get all the matching directories.
 */
           (void) strcpy (q, ".*");
           dirinit(&zone_dir, pointspec);
           while (!dirread(&zone_dir))
           {

/*
 * We have a match on the outbound spec. Make sure it's a directory.
 *
 * Then:
 *
 *       1) If no extension, we may only use it if this is alias 0
 *       2) If an extension, it must be a 3-digit hex number
 *
 * If the extension passes one of these tests, get to work!
 */
               if (!(zone_dir.attr & FA_SUBDIR))
                   continue;
   
               q = strchr (zone_dir.name, '.');
               if (q == NULL)
               {
                   if (k != 0 && !no_zones)
                       continue;
                   zone = (int) alias[0].Zone;
               }
               else
               {
                   if (no_zones)
                       continue;
                   s = ++q;
                   longzone = strtol (q, &q, 16);
                   if ((s + 3) != q)
                       continue;
                   zone = (int) longzone;
/*
 * To avoid duplicating the primary scan,
 * make sure that this isn't "domainname[0].zone[0]".
 */
                   if (!k && zone == (int) alias[0].Zone)
                       continue;
               }
/*
 * OK. We have a domain, an outbound directory, and a zone.
 * That means there's an outbound to scan.
 *
 * Start by scanning the nodes.
 */
               tmp.Zone = zone;
               tmp.Domain = domain;
               tmp.Net = tmp.Node = tmp.Point = 0;
               p = xmit_find (p, &tmp);
/*
 * Now we do the points contained in this outbound.
 *
 */

               if (pvtnet <= 0)
               {
                   HoldName = HoldAreaNameMunge (&tmp);
                   (void) sprintf (pointspec, "%s*.PNT", HoldName);
   
                   dirinit(&pnt_dir,pointspec);
                   while (!dirread(&pnt_dir))
                   {
                       if (sscanf (pnt_dir.name, "%04hx%04hx.", &(tmp.Net), &(tmp.Node)) != 2)
                           continue;
                       tmp.Point = 1;
                       p = xmit_find (p, &tmp);
                   }               /* got one */
                   dirquit(&pnt_dir,0);
               }
           }                       /* while !dfind (...) to get outbounds */
           dirquit(&zone_dir,0);
           cacheflush();
           
/*
 * See if there are any more domains. If so, set up the right name
 * for the outbound, so we can find 'em.
 */
           if ((domain = domain_name[++k]) != NULL)
           {
               *domain_loc = '\0';
               (void) strcpy (pointspec, domain_area);
               q = &(pointspec[strlen (pointspec)]);
               s = domain_abbrev[k];
               if (s != NULL)
                   while (*s)
                       *q++ = *s++;
           }
       }while (domain != NULL);

       /* dump mail info here to btrescan.dmp */
       
       sprintf (flgfname, "%sBTRESCAN.DMP", flag_dir);
       unlink (flgfname);
       if((fp=share_fopen(flgfname, write_binary, DENY_WRITE))!=NULL){
          for(p = mail_top;p;p=p->next){
             dumpmail.mail = *p;
             if(p->mail_addr.Domain){
                for(i=0;domain_name[i];i++){
                   if(p->mail_addr.Domain==domain_name[i]){
                      strncpy(dumpmail.Domain,domain_abbrev[i],8);
                      break;
                   }
                }
             }else
                dumpmail.Domain[0]='\0';
             fwrite(&dumpmail, sizeof(DUMPMAIL), 1,fp);
          }
          fclose(fp);
       }

    }
    
    next_mail = NULL;
    xmit_sort ();
    
    if(display){ /* TJW 960526 */
        xmit_window (mail_top);
        if (un_attended && fullscreen)
           do_ready (M_READY_WAITING);
    }

    sprintf (flgfname, "%sBTRESCAN.BSY", flag_dir);
    unlink(flgfname);                         /* TJW 960616 delete BSY flag */
    
    /* mtt 96-04-16 */
    next_rescan = (long)unix_time (NULL) + readholdtime; /* At least "readholdtime" min to next scan */
    list_next_event();
    sb_show();
}

static MAILP LOCALFUNC 
xmit_find (MAILP p, ADDRP address)
{
    char next_one[PATHLEN];
    char *HoldName;
    ADDR tmp;

    tmp = *address;

    HoldName = HoldAreaNameMunge (address);

    if (address->Point != 0)
    {
        (void) sprintf (next_one, "%s%04hx%04hx.PNT\\*.*",
            HoldName, address->Net, address->Node);
        tmp.Point = 0;
    }
    else
    {
        (void) sprintf (next_one, "%s*.*", HoldName);
        tmp.Net = 0;
        tmp.Node = 0;
        tmp.Point = 0;
    }

    dirinit(&dta_str,next_one);
    while (!dirread(&dta_str))
    {
        /* We have a match. Was it a .FLO file or a .OUT file? */

        if (strncmp (&(dta_str.name[10]), "LO", 2) == 0)
        {
            /* FLO, DLO, CLO and HLO are the only ones! */

            if (strchr ("FDCH", dta_str.name[9]) == NULL)
                continue;
        }
        else if (strncmp (&(dta_str.name[10]), "UT", 2) == 0)
        {
            /* OUT, DUT, CUT and HUT are the only ones! */

            if (strchr ("ODCH", dta_str.name[9]) == NULL)
                continue;
        }
        else if (strncmp (&(dta_str.name[9]), "REQ", 3) != 0)
            continue;

        /* We found a name, remember it */

        if (address->Point != 0)
        {
            if (sscanf (dta_str.name, "%08hx.", &(tmp.Point)) != 1)
                continue;
        }
        else if (sscanf (dta_str.name, "%04hx%04hx.", &(tmp.Net), &(tmp.Node)) != 2)
            continue;

        if (p == NULL)
        {
            p = mail_top = (MAILP) calloc (sizeof (MAIL), 1);
        }
        else
        {
            p->next = (MAILP) calloc (sizeof (MAIL), 1);
            p->next->prev = p;
            p = p->next;
        }

        if (xmit_install (p, &tmp))
        {
            /* No good */

            if (p->prev != NULL)
            {
                p = p->prev;
                free (p->next);
                p->next = NULL;
            }
            else
            {
                free (p);
                p = mail_top = NULL;
            }
        }

    }                           /* while (!done) */
    dirquit(&dta_str,0);

    return (p);
}

int 
xmit_next (ADDRP xaddr)
{
    int i, j;

    cacheflush();

    for (i = 0; i < 2; i++)
    {
        /* Set up the proper pointer */

        if ((next_mail == NULL) || (next_mail->next == NULL))
        {
            if (next_rescan < (long) unix_time (NULL))
                xmit_reset (1, 0);
            next_mail = mail_top;
        }
        else
        {
            next_mail = next_mail->next;
        }

        /* Loop through till we find something we can send */

        while (next_mail != NULL)
        {
            if ( (next_mail->mailtypes & MAIL_WILLGO) &&
                !(next_mail->mailtypes & MAIL_UNKNOWN) &&
                !(next_mail->mailtypes & MAIL_TOOBAD))
            {
                if (bad_call (&(next_mail->mail_addr), 0))
                {
                    next_mail->mailtypes |= MAIL_TOOBAD;
                }
                else
                {
                    /* If multitasking, check for mail before calling */

                    if ((TaskNumber != 0) && ((j = any_mail (next_mail)) <= 0))
                    {
                        if (j == 0)
                            xmit_delete ();
                        else
                            next_mail = next_mail->next;
                        continue;
                    }
                    *xaddr = next_mail->mail_addr;
                    xmit_window (next_mail);
                    return (1);
                }
            }
            next_mail = next_mail->next;
        }
    }

    /* Oh well, we tried */

    xmit_window (mail_top);
    return (0);
}

static void LOCALFUNC
xmit_delete_one ()
{
    MAILP p;

    if (any_mail (next_mail) != 0)
    {
        status_line (MSG_TXT (M_STILL_HAVE_MAIL), Full_Addr_Str (&(next_mail->mail_addr)));

        /* We still have something for him */

        next_mail->mailtypes &= ~MAIL_WILLGO;
        next_mail->mailtypes |= MAIL_TRIED;
        return;
    }

    if (next_mail != mail_top)
    {
        p = next_mail->next;
        next_mail = next_mail->prev;
        free (next_mail->next);
        next_mail->next = p;
        if (p != NULL)
            p->prev = next_mail;
    }
    else
    {
        mail_top = mail_top->next;
        free (next_mail);
        if (mail_top != NULL)
            mail_top->prev = NULL;
        next_mail = NULL;
    }
}

void
xmit_delete ()
{
    int i;

    if (next_mail == NULL && no_EMSI_Session)
        return;

    cacheflush();

    if (no_EMSI_Session)
    {
        xmit_delete_one ();
    }
    else
    {
        for (i = (num_rakas - 1); i >= 0; i--)
        {
            next_mail = find_mail (&remote_akas[i]);
            if (next_mail != NULL)
                xmit_delete_one ();
        }
    }

    xmit_window (mail_top);
    no_EMSI_Session = TRUE;
    num_rakas = 0;
}

int 
bad_call (ADDRP baddr, int rwd)
{
    int res;
    int i, j;
    DIRBUF bad_dta;
    FILE *bad_wazoo;
    char *p;
    char *HoldName;

    char fname[PATHLEN];
    char fname1[PATHLEN];

    cacheflush();

    HoldName = HoldAreaNameMunge (baddr);

    (void) sprintf (fname, "%s%s.$$?", HoldName, Hex_Addr_Str (baddr));
    j = (int) strlen (fname) - 1;           /* Point at ?          */
    res = -1;                               /* Initialize to fail  */

    dirinit(&bad_dta,fname);                /* This says findfirst */
    while (!dirread(&bad_dta))              /* as long as we match */
    {
        if (isdigit (bad_dta.name[11]))     /* is there a digit?   */
        {
            fname[j] = bad_dta.name[11];    /* Yes, copy to fname  */
            res = fname[j] - '0';           /* Save it for testing */
            break;                          /* Get out of while    */
        }
        else
            i = 1;                          /* Else use findnext   */
    }
    dirquit(&bad_dta,0);

    if (res == -1)                          /* Successful search?  */
    {
        fname[j] = '0';                     /* No, base digit = 0  */
    }

    if (rwd > 0)
    {
        /* Writing a bad call  */

        /* First create a filename that is one higher than what we've got */

        (void) strcpy (fname1, fname);
        fname1[j]++;
        if (fname1[j] > '9')
            fname1[j] = '9';

        if (res == -1)                      /* Did we have a file? */
        {                                   /* No, make one.       */
            if (rwd == 2)                   /* No carrier */
                res = open (fname, O_CREAT | O_WRONLY | O_BINARY, S_IREAD | S_IWRITE);
            else                            /* With carrier */
                res = open (fname1, O_CREAT | O_WRONLY | O_BINARY, S_IREAD | S_IWRITE);
            i = rwd - 1;                    /* zero-based count    */
            (void) write (res, (char *) &i, sizeof (int));
            (void) close (res);             /* close the file      */
        }
        else
        {                                   /* There was a file    */

            /*
             * 2 = Unsuccessful, No carrier. Update contents of the file.
             */

            if (rwd == 2)
            {
                i = open (fname, O_RDONLY | O_BINARY);
                (void) read (i, (char *) &res, sizeof (int));
                (void) close (i);
                ++res;
                i = open (fname, O_CREAT | O_WRONLY | O_BINARY, S_IREAD | S_IWRITE);
                (void) write (i, (char *) &res, sizeof (int));
                (void) close (i);
            }

            /*
             * 1 = Unsuccessful, Carrier. Update file name to reflect the
             * failure.
             */

            else
                (void) rename (fname, fname1);
        }
    }
    else if (rwd == 0)
    {
        /*
         * 0 = We are reading a bad call status
         */

        /* Is it automatically ok (no .$$ file there) ? */

        if (res == -1)
            return (0);

        /* Were there too many connects with carrier? */

        if (res >= max_connects)
            return (1);

        /* Ok, check for connects without carrier */

        res = 0;
        i = open (fname, O_RDONLY | O_BINARY);
        (void) read (i, (char *) &res, sizeof (int));
        (void) close (i);
        return (res >= max_noconnects);
    }
    else
    {
        /*
         * -1 = Cleanup of bad call status. This happens in two steps:
         * a) delete 'netnode.$$?' in hold area;
         * b) if a 'netnode.Z' file exists in hold area,
         *    1) delete all BADWAZOO.xxx files listed in the .Z file;
         *    2) delete the 'netnode.z' file.
         */

        if (res != -1) {
            (void) unlink (fname);
            cacheflush();
        }

        if (!mail_finished)
            return (0);

        (void) sprintf (fname, "%s%s.Z", HoldName, Hex_Addr_Str (baddr));
        if (dexists (fname))
        {
            if ((bad_wazoo = fopen (fname, read_ascii)) == NULL)
                (void) got_error (MSG_TXT (M_OPEN_MSG), fname);
            else
            {
                while (!feof (bad_wazoo))
                {
                    e_input[0] = '\0';
                    if (!fgets (e_input, 64, bad_wazoo))
                        break;

                    /* Point to BADWAZOO.xxx */

                    p = strchr (e_input, ' ') + 1;

                    /* Then just past it and terminate */

                    p = strchr (p, ' ');
                    *p = '\0';

                    /* Back to where we were */

                    p = strchr (e_input, ' ') + 1;

                    /* Build file name and delete file */

                    (void) strcpy (fname1, CURRENT.sc_Inbound);
                    (void) strcat (fname1, p);
                    (void) unlink (fname1);
                    cacheflush();

                }
                (void) fclose (bad_wazoo);
            }
            (void) unlink (fname);
            cacheflush();
        }
    }
    return (0);
}

void 
set_up_outbound (int forcerescan)
{
    MAILP mp;
    clock_t start;
    clock_t ticks;
  
    start=clock();

    set_xy ("");

    xmit_reset (1, forcerescan);

    /* and remember where we left off */

    if (hist.next_addr.Net != 0)
    {
        next_addr = hist.next_addr;
        next_addr.Domain = NULL;
        mp = find_mail (&next_addr);
        if ((mp == NULL) || (mp->prev == NULL))
        {
            next_mail = NULL;
            xmit_window (mail_top);
        }
        else
        {
            next_mail = mp->prev;
            xmit_window (next_mail);
        }
    }
    else
    {
        next_addr.Zone = next_addr.Net = next_addr.Node = next_addr.Point = 0;
        next_addr.Domain = NULL;
        xmit_window (mail_top);
    }

    ticks = clock()-start;
    rescanticks += ticks;
    rescancnt++;
    if (rescanstat) {
        clock_t average=rescanticks/rescancnt;
        status_line(MSG_TXT (M_CACHEHOLD_STAT),

            /* jl 960729:  added CPS10 to avoid floating point arithmetik
                           with BC++/DOS (CLOCKS_PER_SEC is 18.2)

                           added casts to match format string */

#define CPS10 ((int) (CLOCKS_PER_SEC * 10))

            (int)  (ticks*10/CPS10), (int) (((1000*ticks)/CPS10)%100),
            (long) (cachehold ? (cachemem+512)/1024 : 0),
            (int)  (average*10/CPS10), (int) (((1000*average)/CPS10)%100));
    }
    heapchk();
}

void 
kill_bad (void)
{
    int k, zone;
    char *q, *s;
    char *domain;
    char *HoldName;
    char pointspec[PATHLEN];
    ADDR tmp;
    DIRBUF zone_dir;
    DIRBUF pnt_dir;
    long longzone;

    cacheflush();
    
/*
 * Initialize domain to scan. Choose (of course) the
 * domain of our primary address.
 *
 * This domain is special in that its outbound is hold_area.
 * All the other domains use their abbreviation as the name
 * of their outbounds.
 */
    k = 0;
    domain = domain_name[0];
    (void) strcpy (pointspec, hold_area);
    q = &(pointspec[strlen (pointspec)]) - 1;
    do
    {
/*
 * Initialize scan of zones in this domain. Using findfirst/findnext,
 * get all the matching directories.
 */
        (void) strcpy (q, ".*");
        dirinit(&zone_dir,pointspec);
        while (!dirread(&zone_dir))
        {
/*
 * We have a match on the outbound spec. Make sure it's a directory.
 *
 * Then:
 *
 *       1) If no extension, we may only use it if this is alias 0
 *       2) If an extension, it must be a 3-digit hex number
 *
 * If the extension passes one of these tests, get to work!
 */
            if (!(zone_dir.attr & FA_SUBDIR))
                continue;

            q = strchr (zone_dir.name, '.');
            if (q == NULL)
            {
                if (k != 0 && !no_zones)
                    continue;
                zone = (int) alias[0].Zone;
            }
            else
            {
                if (no_zones)
                    continue;
                s = ++q;
                longzone = strtol (q, &q, 16);
                if ((s + 3) != q)
                    continue;
                zone = (int) longzone;
/*
 * To avoid duplicating the primary scan,
 * make sure that this isn't "domainname[0].zone[0]".
 */
                if (!k && zone == (int) alias[0].Zone)
                    continue;
            }
/*
 * OK. We have a domain, an outbound directory, and a zone.
 * That means there's an outbound to scan.
 *
 * Start by scanning the nodes.
 */
            tmp.Zone = zone;
            tmp.Domain = domain;
            tmp.Net = tmp.Node = tmp.Point = 0;
            kill_one (&tmp);
/*
 * Now we do the points contained in this outbound.
 *
 */
            if (pvtnet <= 0)
            {
                HoldName = HoldAreaNameMunge (&tmp);
                (void) sprintf (pointspec, "%s*.PNT", HoldName);

                dirinit(&pnt_dir,pointspec);
                while (!dirread(&pnt_dir))
                {
                    if (sscanf (pnt_dir.name, "%04hx%04hx.", &(tmp.Net), &(tmp.Node)) != 2)
                        continue;
                    tmp.Point = 1;
                    kill_one (&tmp);
                }               /* got one */
                dirquit(&pnt_dir,0);
            }
        }                       /* while !dfind (...) to get outbounds */
        dirquit(&zone_dir,0);
/*
 * See if there are any more domains. If so, set up the right name
 * for the outbound, so we can find 'em.
 */
        if ((domain = domain_name[++k]) != NULL)
        {
            *domain_loc = '\0';
            (void) strcpy (pointspec, domain_area);
            q = &(pointspec[strlen (pointspec)]);
            s = domain_abbrev[k];
            if (s != NULL)
                while (*s)
                    *q++ = *s++;
        }
    }
    while (domain != NULL);
}

static void LOCALFUNC 
kill_one (ADDRP address)
{
    char next_one[PATHLEN];
    char thisfile[PATHLEN];
    char *HoldName;
    char *p;

    HoldName = HoldAreaNameMunge (address);

    if (address->Point != 0)
    {
        (void) sprintf (next_one, "%s%04hx%04hx.PNT\\*.$$?",
            HoldName, address->Net, address->Node);
    }
    else
    {
        (void) sprintf (next_one, "%s*.$$?", HoldName);
    }

    strcpy (thisfile, next_one);
    p = strrchr (thisfile, '*');

    strcpy (thisfile, next_one);
    p = strrchr (thisfile, '*');
    dirinit(&dta_str,next_one);
    while (!dirread(&dta_str))
    {
        strcpy (p, dta_str.name);
        status_line (">Deleting %s", thisfile);
        unlink (thisfile);

    }                           /* while (!done) */
    dirquit(&dta_str,0);
}

/*
 * Calculate size of mail queued for outbound.
 * Used to determine whether or not we have enough
 * mail to merit an outbound call.
 *
 * The original version of this code was donated by Henry Clark.
 */

#if 0 /* OBSOLETE */
/* MMP 960513 This code is now obsolete, replaced by cachenetsize()
 * in cache.c
 */

static long LOCALFUNC netsize (MAILP);

static long LOCALFUNC 
netsize (MAILP p)
{
    struct stat stbuf;
    char net_path[PATHLEN];
    char *ptr;
    FILE *temp;
    long accum = 0L;
    char *q;

    /* Append the ARCmail file name to the path line  */

    (void) sprintf (net_path, "%s%s.%s",
        HoldAreaNameMunge (&(p->mail_addr)),
        Hex_Addr_Str (&(p->mail_addr)),
        &dta_str.name[9]);

    temp = share_fopen (net_path, read_binary, DENY_NONE);

    if (temp == (FILE *) NULL)
        return (accum);

    while (!feof (temp))
    {
/*
 *      Make sure of a nice zero there if we're at an undetected EOF.
 *      Then try to read a line from the file.
 */
        net_path[0] = '\0';
        (void) fgets (net_path, 79, temp);
/*
 *      Clean up anything we don't want to see (blanks, tabs, CR, etc)
 */
        for (q = ptr = net_path; *q; q++)
            if (*q <= ' ')
                *q = '\0';
/*
 *      File disposition commands should be skipped over to get to
 *      actual filenames.
 */
        if ((*ptr == TRUNC_AFTER) ||
            (*ptr == DELETE_AFTER) ||
            (*ptr == SHOW_DELETE_AFTER) ||
            (*ptr == NOTHING_AFTER))
            ptr++;
/*
 *      Now -- if what's left starts with a semicolon, it's a comment.
 *      If it starts with a tilde, the file has already been sent. If
 *      what we see is a zero, there's nothing on the line. In any one
 *      of these cases, we should skip this line.
 */
        if ((*ptr == '\0') ||
            (*ptr == ';') ||
            (*ptr == '~'))
            continue;
/*
 *      Get the file size by doing a stat call. If it's not there
 *      then we obviously need not add any size in. If the file is
 *      arcmail, figure it into the "oldest" calculation.
 */
        if (unix_stat (ptr, &stbuf)) /* file exist? */
            continue;
        else
        {
            p->numfiles++;
            accum += stbuf.st_size;
            if (is_arcmail (ptr, 1 - strlen (ptr)))
                p->oldest = min ((unsigned long)stbuf.st_ctime, p->oldest);
        }
    }
    (void) fclose (temp);
    return (accum);
}
#endif /* OBSOLETE */

/*
 * Figure out if there's any mail for the specified node.
 * Used to determine if we need to actually make a call.
 *
 * Returns 0 for no mail at all
 *         1 for non-hold mail
 *        -1 for hold mail only
 *
 * I can't remember who was the first to demand this. I've been
 * pummelled for a while and there's almost certainly been a loss
 * of grey matter.
 */

static int LOCALFUNC 
any_mail (MAILP node)
{
    DIRBUF dta;
    char next_one[PATHLEN];
    int ret = 0;

    (void) sprintf (next_one, "%s%s.*",
        HoldAreaNameMunge (&(node->mail_addr)),
        Hex_Addr_Str (&(node->mail_addr)));

    dirinit(&dta,next_one);
    while (!dirread(&dta))
    {

#ifndef JACK_DECKER
        if (dta.name[9] == 'H')
        {
            ret = -1;
            continue;
        }
#endif
        if ((!strncmp (&(dta.name[10]), "LO", 2)) ||
            (!strncmp (&(dta.name[10]), "UT", 2)))
        {
            ret = 1;
            break;
        }
    }
    dirquit(&dta,0);
    return (ret);
}

