/*  Maximus 3.x MCP IPC and Maximus 2.x IPC routines */


#ifdef  OS_2    /*  TS 960829   */
#define INCL_DOS
#define INCL_DOSFILEMGR
#include <os2.h>

#ifdef OS2SETTITLE                    /* TJW 961010 */
APIRET16 APIENTRY16 DOSSMSETTITLE(char *); /* set OS/2 window title */
#endif

#endif          /*  #ifdef  OS_2    */  /*  TS 960829   */

#include <stdio.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <share.h>
#include <string.h>
#include <time.h>

#ifndef __FLAT__
#define __FLAT__
#endif

#include "includes.h"
#include "msgs.h"

#ifdef  MAX_MCP    /*  TJW 961010   */

#include "maxmcp.h"

/* stuff snatched from the max include files */

/* from max.h */

struct _cstat
{
  unsigned short avail;
  unsigned char username[36];
  unsigned char status[80];
};

/* Data element in IPCxx.BBS file (see MAX_CHAT.C) */

struct _cdat
{
  unsigned short tid;
  unsigned short type;
  unsigned short len;
  unsigned short dest_tid;
  unsigned int rsvd1;
};

static struct _cstat cstat;

char MCPpipename[_MAX_PATH];

static HPIPE hpMCP;

#endif  /*  #ifdef  MAX_MCP    */

/* IPC struct (also used for DOS IPC) */

struct Ipc {
  short          avail;
  char           username[36];
  char           status[80];
  unsigned short msgs_waiting;
  unsigned long  next_msgofs;
  unsigned long  new_msgofs;
};

/* Send a ping to the server every two minutes to show that we're awake */

void IPC_Ping(void)
{
#ifdef MAX_MCP
  static long time_ping=0L;

  if (hpMCP && timeup(time_ping))
  {
    McpSendMsg(hpMCP, PMSG_PING, NULL, 0);
    time_ping=timerset(2 * 60 * 100); /* send ping to MCP ~ once every two minutes */
  }
#endif
}

/* Send IPC message                                                       *
 * Also OS/2 Window Title is changed if OS2SETTITLE is defined - this     *
 * makes use of an undocumented OS/2 function.                            */

void IPC_SetStatus(char *status)
{

#ifdef OS2SETTITLE
  if(title_format){
    char windowtitle[256];
    sprintf(windowtitle, title_format, TaskNumber, status);
    DOSSMSETTITLE(windowtitle);
  }
#endif  

#ifdef MAX_MCP
  if(hpMCP){
    cstat.avail=0;
    strcpy(cstat.username, MSG_TXT(M_MCP_USERNAME));
    strcpy(cstat.status, status);
    McpSendMsg(hpMCP, PMSG_SET_STATUS, (unsigned char *)&cstat, sizeof cstat);
  }
#endif  

  if(ipc_dir){
    struct Ipc ipc;
    char fname[PATHLEN];
    FILE *fp;
    memset (&ipc, 0, sizeof(ipc));
    strncpy (ipc.username, PRDCT_PRTY PRDCT_VRSN, 35);
    strncpy (ipc.status, status, 79);
    ipc.avail = 0;
    sprintf (fname, "%sIPC%02X.BBS", ipc_dir, TaskNumber);
    if((fp = share_fopen (fname, write_binary, DENY_WRITE))!=NULL){
      fwrite (&ipc, sizeof(ipc), 1, fp);
      fclose (fp);
    }
  }
}

/* Open / Initialize IPC */

int IPC_Open(void)
{
#ifdef MAX_MCP
  char MCPszPipe[_MAX_PATH];
  APIRET rc, usAction;
  unsigned char tid=TaskNumber;
  long retrytimer;

  /* IPC already open */

  if (hpMCP)
    return -1;


  /* The place for us to connect is the "\maximus" side off the root
   * MCP "path".
   */

  if (!*MCPpipename)     /* there was no MCP pipe definition in binkley.cfg */
  {
    hpMCP = 0;
    return 0;
  }
  
  strcpy(MCPszPipe, MCPpipename);
  strcat(MCPszPipe, "\\maximus");

  retrytimer = timerset(500);        /* retry for 5s to open pipe */
  
  do{
    rc = DosOpen(MCPszPipe, &hpMCP, &usAction, 0L, FILE_NORMAL,
                 OPEN_ACTION_FAIL_IF_NEW | OPEN_ACTION_OPEN_IF_EXISTS,
                 OPEN_SHARE_DENYREADWRITE | OPEN_ACCESS_READWRITE |
                 OPEN_FLAGS_NOINHERIT, NULL);
    DosSleep(100L);
  }while(rc && !timeup(retrytimer));
    
  if (!rc)
    McpSendMsg(hpMCP, PMSG_HELLO, &tid, 1);
  else
  {
    status_line(MSG_TXT (M_MCP_COULDNTOPENPIPE), MCPpipename);
    hpMCP = 0;
  }
#endif

  return 0;
}


/* Close IPC */

void IPC_Close(void)
{
#ifdef MAX_MCP
  if (hpMCP)
  {
    McpSendMsg(hpMCP, PMSG_EOT, NULL, 0);
    DosClose(hpMCP);
    hpMCP=0;
  }
#endif
}

