/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                                                          */
/*      ------------         Bit-Bucket Software, Co.                       */
/*      \ 10001101 /         Writers and Distributors of                    */
/*       \ 011110 /          Freely Available<tm> Software.                 */
/*        \ 1011 /                                                          */
/*         ------                                                           */
/*                                                                          */
/*              (C) Copyright 1987-96, Bit Bucket Software Co.              */
/*                                                                          */
/*                This module was written by Vince Perriello                */
/*                     with code from several authors                       */
/*                Miscellaneous routines used by BinkleyTerm                */
/*                                                                          */
/*                                                                          */
/*    For complete  details  of the licensing restrictions, please refer    */
/*    to the License  agreement,  which  is published in its entirety in    */
/*    the MAKEFILE and BT.C, and also contained in the file LICENSE.260.    */
/*                                                                          */
/*    USE  OF THIS FILE IS SUBJECT TO THE  RESTRICTIONS CONTAINED IN THE    */
/*    BINKLEYTERM  LICENSING  AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF    */
/*    THIS  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES,  OR IF YOU DO    */
/*    NOT HAVE THESE FILES,  YOU  SHOULD  IMMEDIATELY CONTACT BIT BUCKET    */
/*    SOFTWARE CO.  AT ONE OF THE  ADDRESSES  LISTED BELOW.  IN NO EVENT    */
/*    SHOULD YOU  PROCEED TO USE THIS FILE  WITHOUT HAVING  ACCEPTED THE    */
/*    TERMS  OF  THE  BINKLEYTERM  LICENSING  AGREEMENT,  OR  SUCH OTHER    */
/*    AGREEMENT AS YOU ARE ABLE TO REACH WITH BIT BUCKET SOFTWARE, CO.      */
/*                                                                          */
/*                                                                          */
/* You can contact Bit Bucket Software Co. at any one of the following      */
/* addresses:                                                               */
/*                                                                          */
/* Bit Bucket Software Co.        FidoNet  1:104/501, 1:343/491             */
/* P.O. Box 460398                AlterNet 7:42/1491                        */
/* Aurora, CO 80046               BBS-Net  86:2030/1                        */
/*                                Internet f491.n343.z1.fidonet.org         */
/*                                                                          */
/* Please feel free to contact us at any time to share your comments about  */
/* our software and/or licensing policies.                                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/

/* Include this file before any other includes or defines! */

#include "includes.h"

#define EVENT_ENABLE

#ifdef Snoop
#include "snserver.h"
extern HSNOOP hsnoop;
#endif                          /* Snoop */

static void LOCALFUNC fill_in_status (char *);
static int LOCALFUNC find_addr (char *, ADDR *, char *);

int 
set_baud (unsigned long baudrate, int log)
{
    register int i;
    unsigned long search_rate;
    char junk[256];

    search_rate = (baudrate > max_baud.rate_value) ? max_baud.rate_value : baudrate;

    for (i = 0; pbtypes[i].rate_value; i++)
    {
        if (pbtypes[i].rate_value >= search_rate)
        {
            if ((pbtypes[i].rate_value != search_rate) && (search_rate <= 2400))
                break;

            if (baud != i)      /* same as what we have?     */
            {
                if (log && !un_attended)
                    status_line (MSG_TXT (M_SETTING_BAUD), baudrate);
                baud = i;       /* need this for ALT-B       */
                program_baud ();
                cur_baud = pbtypes[i];
            }

            cur_baud.rate_value = baudrate;

            if (un_attended && fullscreen)
            {
                (void) sprintf (junk, "%-2d  %6lu", port_ptr+1, baudrate);
                sb_move_puts (settingswin, SET_PORT_ROW, SET_COL, junk);
                sb_show ();
            }
            return (1);
        }
    }
    return (0);
}

void 
program_baud ()
{
    unsigned long rate_mask;

    set_prior (PRIO_MODEM);

    if ((lock_baud && (pbtypes[baud].rate_value >= lock_baud))
        || (ARQ_lock != 0))
        rate_mask = max_baud.rate_mask;
    else
        rate_mask = pbtypes[baud].rate_mask;

    MDM_ENABLE (rate_mask);

    set_prior (PRIO_REGULAR);

}

static struct tm *tp;
static time_t ltime;

void
log_screen(char *timestr_scr, char *e_input)
{
   char tmp[300];
   if (!fullscreen || !un_attended){
        (void) sprintf (tmp, "\r\n%c %8s " PRDCT_SHRT " %s",
                        e_input[0], timestr_scr, &e_input[1]);
        scr_printf (tmp);
   }else{
        (void) sprintf (tmp, stat_str,
                             e_input[0], timestr_scr, &e_input[1]);
        fill_in_status (tmp);
   }
#ifdef Snoop
   if (hsnoop != (HSNOOP) 0L)
     (void) SnoopWrite (hsnoop, (PSZ) tmp);
#endif /* Snoop */
}

void _cdecl 
status_line (char *fmt,...)
{
    va_list arg_ptr;
    
    static char *specifiers = "!*+:# >";
    char timestr_scr[40];
    char *p;

    va_start (arg_ptr, fmt);
    if (fmt == NULL || !fmt[0] || !fmt[1])
        return;                               /* TJW 960708:                   */
    (void) vsprintf (e_input, fmt, arg_ptr);  /* this is DANGEROUS if too long */
    e_input[E_INPUT_LEN-1]='\0';              /* avoid the worst of "open end" */
    
    /* Modified by Paul Edwards to protect against control characters
       in the string */
    for (p = e_input; *p != '\0'; p++)
    {
        if (iscntrl(*p))
        {
            if(*p=='\007' && gong_allowed)
                WRITE_ANSI ('\007');        /* TJW 960706 chat request gong */
            *p = '.';
        }
    }

    (void) unix_time (&ltime);
    tp = unix_localtime (&ltime);
    
    if ((e_input[0] != '>') || debugging_log){
    
       (void) sprintf (timestr_scr, "%02i:%02i:%02i",
                                    tp->tm_hour, tp->tm_min, tp->tm_sec);
       log_screen(timestr_scr, e_input);
    }

    /* TJW 960522 negative loglevels log only with CARRIER_DETECT=1 */ 
    if ((status_log != NULL) &&  
        ((strchr (specifiers, e_input[0]) - strchr (specifiers, '!')) <= abs(loglevel)))
    {
      if(loglevel>0 || CARRIER){
#ifdef BTPE
        int hours;
        int mins;
        int secs;
        int hths;
        
        dostime(&hours, &mins, &secs, &hths);
        (void) fprintf(status_log, 
                       "%c %02i %s %02i:%02i:%02i.%02i BINK %s\n", 
                       e_input[0],
                       tp->tm_mday, mtext[tp->tm_mon], 
                       hours, mins, secs, hths,
                       &e_input[1]);
#else                               

        (void) fprintf (status_log,
                        "%c %02i %s %02i:%02i:%02i BINK %s\n",
                         e_input[0],
                         tp->tm_mday, mtext[tp->tm_mon],
                         tp->tm_hour, tp->tm_min, tp->tm_sec,
                         &e_input[1]);
#endif

        if (immed_update)
        {
            (void) fflush (status_log);
            (void) real_flush (fileno (status_log));
            need_update = 0;
        }
        else
        {
            need_update = 1;
        }
      }
    }
    va_end (arg_ptr);
}

/*--------------------------------------------------------------------------*/
/* THROUGHPUT                                                               */
/* Print throughput message at end of transfer                              */
/*--------------------------------------------------------------------------*/
long 
throughput (int opt, unsigned long bytes)
{
    static long started = 0L;
    static long stopped = 0L;  /* TJW 960513 */
    static long elapsed = 0L;  /* TJW 960513 */  
    long cps, lowcps, highcps; /* TJW 960513 */ 
    long efficiency;           /* TJW 960513 */ 
    
    if (!opt)
        started = unix_time (NULL);
    else if (started)
    {   
        stopped = unix_time (NULL);
        /* The next line tests for day wrap without the date rolling over */
        if (stopped < started)
            stopped += 86400L;
        elapsed = stopped-started;
        if (elapsed == 0L)
            elapsed = 1L;
        cps = (long) (bytes / (unsigned long) elapsed);
        efficiency = (cps * 1000L) / ((long) cur_baud.rate_value);
        lowcps = (long) (bytes / ((unsigned long) (stopped-started+1)));
        if(stopped-started>1)
          highcps = (long) (bytes / ((unsigned long) (stopped-started-1)));
        else
          highcps = cps;
        status_line ((char *) MSG_TXT (M_CPS_MESSAGE),
                     cps, lowcps, highcps, bytes, efficiency);
    }
    return elapsed;
}                               /* throughput */

static void LOCALFUNC 
fill_in_status (char *stat_line)
{
    int     i;

    (void) unix_time (&ltime);
    tp = unix_localtime (&ltime);
    if (fullscreen && callwin)           /*PLF Sun  12-01-1991  07:25:27 */
    {
        if (strlen (callwin->endbuff)) 
            memmove (callwin->buffer, 
                callwin->buffer + callwin->linesize, 
                (size_t)(callwin->endbuff - callwin->buffer));
        else if (strlen (callwin->buffer))
        {
            callwin->lastshown += callwin->linesize;
            callwin->lastline += callwin->linesize;
        }

        i = min (strlen (stat_line), (unsigned int)callwin->linesize-2);
        /* i is now string length (without \0 char at end) or max. line    */
        /* length (also without the two \0 chars already there from calloc */
        memcpy (callwin->lastline, stat_line, i);
        /* we needn't copy nor set last char to \0 as it is already there! */
        
        sb_scrl (callwin, 1);
        sb_move_puts (callwin, callwin->lines-1, 0, callwin->lastshown);

        timedisplay();
        sb_show ();
    }
    IPC_Ping();
}

void
timedisplay(void)  /* TJW 960624 display date / time in settings window */
{
   char junk[256];
   
   (void) unix_time (&ltime);
   tp = unix_localtime (&ltime);
   (void) sprintf (junk, "%02d/%02d/%02d %s %02d:%02d",
       tp->tm_year, tp->tm_mon+1, tp->tm_mday,
       wkday[tp->tm_wday],
       tp->tm_hour, tp->tm_min);

   sb_move_puts (settingswin, SET_TIME_ROW, SET_TIME_COL, junk);
}

void 
set_xy (char *string)
{
    WRITE_ANSI ('\r');
    WRITE_ANSI ('\n');
    scr_printf (string);
    locate_x = wherex ();
    locate_y = wherey ();
}

void _cdecl 
time_release ()
{
    EventSub ();
    if (need_update)
    {
        if (status_log != NULL) /*WRA*/
        {
            (void) fflush (status_log);
            (void) real_flush (fileno (status_log));
        }
        need_update = 0;
    }

    dos_break_off ();           /* Turn off ^C trapping */

    (*mtask_idle) ();

}

void 
timer (int interval)
{
    long timeout;

    timeout = timerset ((unsigned int) (interval * 10));
    while (!timeup (timeout))
        time_release ();
}

void 
big_pause (int secs)
{
    long timeout;

    timeout = timerset ((unsigned int) (secs * 100));
    while (!timeup (timeout))
    {
        if (CHAR_AVAIL ())
            break;
        time_release ();
    }
}

void 
scr_printf (char *string)
{
    if (string != NULL)
        if (vfossil_installed)
            (void) VioWrtTTY ((PCH) string, (USHORT) strlen (string), (HVIO) 0L);
        else
            while (*string != 0)
                WRITE_ANSI (*string++);
}

void 
send_can ()
{
    int i;

    CLEAR_OUTBOUND ();
    CLEAR_INBOUND ();

    for (i = 0; i < 10; i++)
        SENDBYTE (CAN);
    for (i = 0; i < 10; i++)
        SENDBYTE (BS);
}

void 
invent_pkt_name (char string[])
{
    struct tm *ptm;
    time_t lTimeT;

    (void) unix_time (&lTimeT);
    ptm = unix_localtime (&lTimeT);
    (void) sprintf (string, "%02i%02i%02i%02i.pkt",
        ptm->tm_mday, ptm->tm_hour, ptm->tm_min, ptm->tm_sec);
}

int 
is_user (char *p)
{
    char *q;
    char far *r;
    int i, j;

    if (cur_event < 0)
        return (0);

    q = strchr (p, '.');
    if (q != NULL)
    {
        ++q;
        for (i = 0; i < 6; i++)
        {
            r = &(e_ptrs[cur_event].err_extent[i][0]);
            for (j = 0; j < 3; j++)
            {
                if ((tolower (q[j]) != tolower (r[j])) && (q[j] != '?'))
                    break;
            }
            if (j == 3)
                user_exits[i] = 1;
        }
    }

    return (0);
}

int 
is_arcmail (char *p, int n)
{
    int i, l;
    char c[PATHLEN];
    static char *suffixes[8]={"SU", "MO", "TU", "WE", "TH", "FR", "SA", NULL};

    l = n;
    if (n < 0)
        n = -n;

    if (!isdigit (p[n]))
    {
        return (l >= 0 ? is_user (p) : 0);
    }

    (void) strcpy (c, p);
    (void) strupr (c);

    for (i = n - 11; i < n - 3; i++)
    {
        if ((!isdigit (c[i])) && ((c[i] > 'F') || (c[i] < 'A')))
            return (l >= 0 ? is_user (p) : 0);
    }

    for (i = 0; i < 7; i++)
    {
        if (strnicmp (&c[n - 2], suffixes[i], 2) == 0)
            break;
    }

    if (i >= 7)
    {
        return (l > 0 ? is_user (p) : 0);
    }

    if (l >= 0)
        got_arcmail = 1;
    return (1);
}

int 
get_number (char *target)
{
    ADDR gaddr;
    int k;

    (void) fgets (target, 100, stdin);
    k = (int) strlen (target);

    if (k == 1)
        return (0);
    target[--k] = '\0';         /* no '\n' */

    if (!isdigit (target[0]) && target[0] != '\"')
    {
        (*userfunc) (target, &gaddr);
        if ((gaddr.Net != 0xffff) && (gaddr.Node != 0xffff) && (gaddr.Zone != 0xffff))
        {
            (void) sprintf (target, "%s", Full_Addr_Str (&gaddr));
        }
        else
            return (0);         /* Gotta have addr */
    }
    return (1);
}

void 
gong ()
{
   long t;
   int i;

   if (gong_allowed){
      for (i = 0; i < 15; i++)
      {
          WRITE_ANSI ('\007');        /* Bell code       */
          t = timerset (100);         /* 1 second        */
          while (!timeup (t))
          {
              if (KEYPRESS ())        /* If key pressed, */
              {
                  (void) READKB ();   /* Throw it away   */
                  return;             /* And get out     */
              }
          }
      }
   }
}

#ifdef DEBUG
void 
show_debug_name (char *string)
{
    int x, y;
    static char *filler = "                           ";

    x = wherex ();
    y = wherey ();
    gotoxy (40, 0);
    scr_printf (string);
    scr_printf (&filler[strlen (string)]);
    gotoxy (x, y);
}
#endif


void 
change_prompt ()
{
    char *s;
    static char newstring[200];

    if (newstring[0])
    {
        (void) putenv (newstring);
        return;
    }

    (void) strcpy (newstring, "PROMPT=[");
    (void) strcat (newstring, xfer_id);
    (void) strcat (newstring, " Shell]$_");
    s = getenv ("PROMPT");
    if (s)
    {
        (void) strcat (newstring, s);
    }
    else
    {
        (void) strcat (newstring, "$P$G");
    }

    (void) putenv (newstring);
}

static char *last_str[] =
{
    "None",      /* 0 */
    "WaZOO",     /* 1 */
    "FTS-0001",  /* 2 */
    "BBS",       /* 3 */
    "Ext Mail",  /* 4 */
    "FAX",       /* 5 */
    "Inc-Fail"   /* 6 TJW 960625 incoming call failed, CONNECT - nothing more */
};

void 
last_type (int n, ADDRP taddr)
{
    char *p;
    char s1[90], s2[90];

    if ((n == 1) || (n == 2))
    {
        if (((short) taddr->Zone != -1000) && ((short) taddr->Net > 0))
        {
            hist.last_addr = *taddr;
            (void) sprintf (s1, "%s", Full_Addr_Str (taddr));
            if ((p = strchr (s1, '@')) != NULL)
                *p = '\0';
            (void) sprintf(s2, "%19s", s1); /* TJW 960703                */
            /* %19s will overwrite "Last:" prefix with long addrs, but   */
            /* better than truncating any part. "Last:" will be redrawn. */
        }
        else
            (void) sprintf(s2, "%11s%8s", "", "FTS-0001");

    }
    else
    {
        if ((n < 0) || (n > 6))
            n = 0;

        (void) sprintf(s2, "%11s%8s", "", last_str[n]);
    }

    if (fullscreen){
       sb_move_puts (historywin, HIST_LAST_ROW, 0, s2); /* TJW 960703 */
       sb_move_puts (historywin, HIST_LAST_ROW, 0, MSG_TXT (M_LAST));
    }
    
    hist.last_caller = (short) n;
}

/*--------------------------------------------------------------------------*/
/* CHECK_NETFILE -- find out if the file we've got is a netfile.            */
/*--------------------------------------------------------------------------*/

char *
check_netfile (char *fname)
{
    register char *p;
    register int n;

    p = fname;
    n = (int) strlen (p) - 1;

    if ((p[n] == 't') && (p[n - 1] == 'k') && (p[n - 2] == 'p') && (p[n - 3] == '.'))
    {
        got_packet = 1;
        got_mail = 1;
        p = MSG_TXT (M_MAIL_PACKET);
    }
    else if (is_arcmail (p, n))
    {
        got_mail = 1;
        p = MSG_TXT (M_COMPRESSED_MAIL);
    }
    else
    {
        short mytask = -1;

        if (p[n - 2] != 'R' || p[n - 3] != '.')
            got_mail = 1;
        else
        {
            sscanf (&p[n - 1], "%02hx", &mytask);
            if (mytask != TaskNumber)
                got_mail = 1;
        }
        p = MSG_TXT (M_NET_FILE);
    }

    return (p);
}

/*--------------------------------------------------------------------------*/
/* UNIQUE_NAME                                                              */
/* Increments the suffix of a filename as necessary to make the name unique */
/*--------------------------------------------------------------------------*/
void 
unique_name (char *fname)
{
    static char suffix[] = ".001";
    register char *p;
    register int n;

    if (dexists (fname))
    {                           /* If file already exists...      */
        p = fname;
        while (*p && *p != '.')
            p++;                /* ...find the extension, if
                                                  * any  */
        for (n = 0; n < 4; n++) /* ...fill it out if
                                                  * neccessary   */
            if (!*p)
            {
                *p = suffix[n];
                *(++p) = '\0';
            }
            else
                p++;

        while (dexists (fname)) /* ...If 'file.ext' exists
                                                  * suffix++ */
        {
            p = fname + strlen (fname) - 1;
            for (n = 3; n--;)
            {
                if (!isdigit (*p))
                    *p = '0';
                if (++(*p) <= '9')
                    break;
                else
                    *p-- = '0';
            }                   /* for */
        }                       /* while */
    }                           /* if exist */
}                               /* unique_name */

int 
got_ESC ()
{
    while (KEYPRESS ())
    {
        screen_blank = 0;
        if (fullscreen && un_attended)
            sb_show ();
        if (READKB () == 27)    /* ESC pressed?        */
        {
            while (KEYPRESS ())
                (void) READKB ();
            return (1);
        }
    }
    return (0);
}

static long maxin_size;
static long maxin_time;
static long maxout_size;
static long maxout_time;

void
calc_cost(int caller, long s, long e,
          long *_dt, long *_totcost, long *_totrcost, word *_cost)
{
   BINK_EVENTP costp;
   long t;
   long tx10,totcost;
   int ci;                  /* costvector index */
   
   if(!e)              /* if e is not supplied, calc min call cost for 1s */
      e = 1;

   if(!s){             /* if s is not supplied, take time NOW and e as dt */
      s = unix_time (NULL);
      e += (s + costtimecorr_sess); /* also add session overhead (def:5s) */
   }
   
   s -= costtimecorr_conn;  /* it takes some worthy time until "CONNECT" */
                            /* default is "5" for costtimecorr_conn      */
   *_dt = e - s;
   *_cost = 0;
   *_totcost = 0;
   *_totrcost = 0;
   
   ci = newnodedes.RealCost;  /* cost field of nodelist index is cv index */
   if(ci==0 || ci>16) 
     return;                  /* ==0: all done, >16: emergency exit ! */ 
   *_cost = ci;               /* TJW 960825 take cv index as cost value   */
   ci--;                      /* make cost index 0-based              */

   costp = find_costevent(s);
   if(eurocost){
      t=s; tx10=0;                     /* 1 s and 1/10 s counter             */
      totcost=0;                       /* cumulated costs of call            */
      while(t<e){                      /* simulate costs until we reach end  */
         totcost += cost_unit;         /* we have spent 1 more cost unit now */
         if(costp)
           tx10 += costp->costvector[ci]; /* and can do now for some time... */
         else
           return;                     /* emergency exit !                   */
         t = s + tx10/10;              /* calculate real time t              */
         costp = find_costevent(t);    /* which tariff do we have now?       */
      }
   }else{
      t=s;                             /* 1 s counter                        */
      totcost=0;                       /* cumulated costs of call            */
      while(t<e){                      /* simulate costs until we reach end  */
         if(costp)
           totcost += costp->costvector[ci]; /* we spent some cents more now */
         t += 60;                      /* and can do now for 1 minute more.. */
         costp = find_costevent(t);    /* which tariff do we have now?       */
      }
   
   }
   if(caller)
     *_totcost = totcost;              /* my costs */
   else
     *_totrcost = totcost;             /* remote's costs */
}

long 
cost_of_call (int caller, long s, long e)
{
    long mycost, remotecost, dt;
    char *p;
    char logline[34];
    int session = 0;
    word costidx;
    
    int infiles, outfiles, totfiles;
    long insize, outsize, totsize;
    long intime, outtime, tottime;
    long incps, outcps, totcps;
    long ineff, outeff, toteff;
    long inerr, outerr, toterr;

    char costfmt[] = "%s Stat %4s %3hd Files (%s)%5ldcps %3hd%%  Err: %ld\n";

    (void) unix_time (&ltime);
    tp = unix_localtime (&ltime);
    
    sprintf (logline, "%02i %03s %02i:%02i",
        tp->tm_mday, mtext[tp->tm_mon], tp->tm_hour, tp->tm_min);

    calc_cost(caller, s, e, &dt, &mycost, &remotecost, &costidx);   

    if (caller || hist.last_caller == 1 || hist.last_caller == 2)
    {
        session = 1;
        p = Full_Addr_Str (&remote_addr);
    }
    else if (hist.last_caller>=3 && hist.last_caller<=6)
        p = last_str[hist.last_caller];
    else
        p = "Unknown";

    infiles = hist.files_in - start_hist.files_in;
    outfiles = hist.files_out - start_hist.files_out;
    totfiles = infiles + outfiles;

    insize = hist.size_in - start_hist.size_in;
    outsize = hist.size_out - start_hist.size_out;
    totsize = insize + outsize;

    intime = hist.time_in - start_hist.time_in;
    outtime = hist.time_out - start_hist.time_out;
    tottime = dt;

    inerr = hist.err_in - start_hist.err_in;
    outerr = hist.err_out - start_hist.err_out;
    toterr = inerr + outerr;

    start_hist = hist;

    incps = (intime) ? (insize / intime) : 0;
    outcps = (outtime) ? (outsize / outtime) : 0;
    totcps = (tottime) ? (totsize / tottime) : 0;

    ineff = incps * 1000L / cur_baud.rate_value;
    outeff = outcps * 1000L / cur_baud.rate_value;
    toteff = totcps * 1000L / cur_baud.rate_value;

    if (session && cost_log != NULL)
    {
        if (shortcostlog) {

            char *c;
            char pad='\0';  /*  TS 960620 since maybe used unititialized  */
            char buf[40];
            int len;

            for (c=shortcostlog; *c; c++)
                if (*c!='$') {
                    fputc(*c, cost_log);
                } else {
                    len=0, c++;
                    if (isdigit(*c)) {
                        pad = (*c=='0') ? '0' : ' ';
                        len=atoi(c);
                        while (isdigit(*c))
                            c++;
                    }
                    switch(*c) {
                        case '#': sprintf(buf, "%d", TaskNumber); break;
                        case '$': sprintf(buf, "$"); break;
                        case '<': sprintf(buf, "\n"); break;
                        case 'A': sprintf(buf, "%s", p); break;
                        case 'B': sprintf(buf, "%lu", connectrate); break;
                        case 'C': sprintf(buf, "%ld", max(maxin_size,maxout_size)); break;
                        case 'D': sprintf(buf, "%ld", maxin_size>maxout_size ? maxin_time : maxout_time); break;
                        case 'E': sprintf(buf, "%lu", toterr); break;
                        case 'H': sprintf(buf, "%d", tp->tm_hour); break;
                        case 'I': sprintf(buf, "%d", infiles); break; /*  TS 960620 */
                        case 'J': sprintf(buf, "%ld", incps); break;
                        case 'M': sprintf(buf, "%d", tp->tm_min); break;
                        case 'O': sprintf(buf, "%d", outfiles); break;/*  TS 960620 */
                        case 'P': sprintf(buf, "%ld", outcps); break;
                        case 'S': sprintf(buf, "%d", tp->tm_sec); break;
                        case 'T': sprintf(buf, "%d", totfiles); break;/*  TS 960620 */
                        case 'U': sprintf(buf, "%ld", totcps); break;
                        case 'V': sprintf(buf, "%ld", maxin_size); break;
                        case 'W': sprintf(buf, "%ld", maxin_time); break;
                        case 'X': sprintf(buf, "%ld", maxin_time ? (maxin_size/maxin_time) : 0); break;
                        case 'Y': sprintf(buf, "%ld", maxin_size>maxout_size
                            ? (maxin_time ? (maxin_size/maxin_time) : 0)
                            : (maxout_time ? (maxout_size/maxout_time) : 0));
                            break;
                        case 'b': sprintf(buf, "%s", mtext[tp->tm_mon]); break;
                        case 'c': sprintf(buf, "%ld", mycost); break;
                        case 'd': sprintf(buf, "%d", tp->tm_mday); break;
                        case 'f': sprintf(buf, "%ld", inerr); break;
                        case 'g': sprintf(buf, "%ld", outerr); break;
                        case 'i': sprintf(buf, "%ld", insize); break;
                        case 'j': sprintf(buf, "%ld", ineff); break;
                        case 'm': sprintf(buf, "%d", tp->tm_mon+1); break;
                        case 'o': sprintf(buf, "%ld", outsize); break;
                        case 'p': sprintf(buf, "%ld", outeff); break;
                        case 'r': sprintf(buf, "%ld", remotecost); break; /* TJW 960812 */
                        case 's': sprintf(buf, "%ld", dt); break;
                        case 't': sprintf(buf, "%ld", totsize); break;
                        case 'u': sprintf(buf, "%ld", toteff); break;
                        case 'v': sprintf(buf, "%ld", maxout_size); break;
                        case 'w': sprintf(buf, "%ld", maxout_time); break;
                        case 'x': sprintf(buf, "%ld", maxout_time ? (maxout_size/maxout_time) : 0); break;
                        case 'y': sprintf(buf, "%d", tp->tm_year%100); break;
                        case 'z': sprintf(buf, "%hu", costidx); break; 
                        default : sprintf(buf, "$%c",*c); break;
                    }
                    len-=strlen(buf);
                    while (len-- > 0)
                        fputc(pad, cost_log);
                    fputs(buf, cost_log);
                }
            fputc('\n', cost_log);

        } else {

        fprintf (cost_log, costfmt, logline, "Sent ",
            (outfiles > 999) ? 999 : outfiles,
            numdisp (outsize), outcps, outeff,
            outerr);
        fprintf (cost_log, costfmt, logline, "Rcvd ",
            (infiles > 999) ? 999 : infiles,
            numdisp (insize), incps, ineff,
            inerr);
        fprintf (cost_log, costfmt, logline, "Total",
            (totfiles > 999) ? 999 : totfiles,
            numdisp (totsize), totcps, toteff,
            toterr);

        fprintf (cost_log, "%s Seconds: %ld  Cost: %ld  System: %s\n\n",
            logline, dt, mycost, p);

        }

        fflush (cost_log);
        real_flush (fileno (cost_log));
    }

    if (session)
    {                             /* TJW 960527 changed to KB display */
        status_line (MSG_TXT (M_BYTE_SUMMARY),
            infiles, (insize+512)/1024,
            outfiles, (outsize+512)/1024,
            totfiles, (totsize+512)/1024,
            totcps);
    }

    status_line (MSG_TXT (M_COST_SUMMARY), dt, costidx, mycost, remotecost, p);

    if (modem_aftercall && remote_pickup != -2)
    {
        int i = 0;
        long lwait;

        mdm_cmd_string (modem_aftercall, 0);

        lwait = timerset (300);
        for (i = 0; i < lines_aftercall && !timeup (lwait); i++)
        {
            if (!stricmp (get_response (lwait, 1), "NO CARRIER"))
            {
                CLEAR_INBOUND ();
                break;
            }
            lwait = timerset (100);
        }
    }

    maxin_size = 0;
    maxin_time = 0;
    maxout_size = 0;
    maxout_time = 0;

    return (mycost);
}

/*
 * Create a small character string for mailsize and cost_of_call.
 */

char *
numdisp (long number)
{
    int i = 0;
    char tempstr[6];
    long quotient, oldq;
    int intq;
    static char e_stuff[6];         /* buffer for result */
    static char *rgchSize = " KMG"; /* TJW 960528 was "bKMG" - if we don't */
                                    /* say "Kb" then also don't say "b" !  */
    oldq = 0L;
    quotient = number;

    while (quotient >= 1024L)
    {
        oldq = quotient;
        quotient = oldq / 1024L;
        i++;
    }

    intq = (int) quotient;

  /*
   * If more than 999 but less than 1024, it's a big fraction of
   * the next power of 1024. Get top two significant digits
   * (so 1023 would come out .99K, for example)
   */

    if (intq > 999)
    {
        intq = (intq * 25) / 256;   /* 100/1024 */
        (void) sprintf (e_stuff, ".%2d%c", intq, rgchSize[++i]);
    }

  /*
   * If less than 10 and not small units, then get some decimal
   * places (e.g. 1.2M)
   */

    else if ((intq < 10) && (i != 0))
    {
        intq = (int) ((oldq * 5L) / 512L);  /* 10/1024 */
        (void) sprintf (tempstr, "%02d", intq);
        (void) sprintf (e_stuff, "%c.%c%c", tempstr[0], tempstr[1], rgchSize[i]);
    }

  /*
   * Simple case. Just do it.
   */

    else
        (void) sprintf (e_stuff, "%3d%c", intq, rgchSize[i]);

    return (e_stuff);
}

void
update_files (int sender, char *filename, long filesize, long filetime, long err)
{
    if (un_attended)
    {
        if (sender)
        {
            ++hist.files_out;
            hist.size_out += filesize;
            hist.time_out += filetime;
            hist.err_out += err;
        }
        else
        {
            ++hist.files_in;
            hist.size_in += filesize;
            hist.time_in += filetime;
            hist.err_in += err;
        }
    }

#ifndef CSS
    do_today(1);
#endif    

    if ( (cost_log != NULL) && isOriginator && shortcostlog==NULL )
    {
        char *pszMode = (sender ? "Sent" : "Received");

        ltime = unix_time (NULL);
        tp = unix_localtime (&ltime);

        fprintf (cost_log, "%02i %03s %02i:%02i %s: %s (%ld)\n",
            tp->tm_mday, mtext[tp->tm_mon], tp->tm_hour, tp->tm_min,
            pszMode, filename, filesize);
        fflush (cost_log);
        real_flush (fileno (cost_log));
    }

    if (sender) {
        if (filesize>maxout_size) {
            maxout_size = filesize;
            maxout_time = filetime;
        }
    } else {
        if (filesize>maxin_size) {
            maxin_size = filesize;
            maxin_time = filetime;
        }
    }
}

void 
screen_clear ()
{
    unsigned int r;
    PCH q;

    if (!vfossil_installed)
        scr_printf ("\033[H\033[2J");
    else
    {
        for (r = 0; (int) r <= (int) SB_ROWS; r++)
        {
            q = (PCH) blanks;
            (void) VioWrtCellStr ((PCH) q, (USHORT) (SB_COLS * 2), (USHORT) r, (USHORT) 0, (HVIO) 0L);
        }

        gotoxy (0, 0);
    }
}

void 
clear_eol ()
{
    unsigned int x, y;
    PCH q;

    if (!vfossil_installed)
        (void) printf ("\033[K");
    else
    {
        x = (unsigned int) wherex ();
        y = (unsigned int) wherey ();

        q = (PCH) blanks;
        (void) VioWrtCellStr ((PCH) q, (USHORT) ((SB_COLS - x) * 2), (USHORT) y, (USHORT) x, (HVIO) 0L);
    }
}

void 
log_product (int product, int version, int subversion) /* TJW 960528 modified */
{
    char buf[20], *prodstr, buf2[100];

    switch (product)
    {

    /*  Special case Opus. */
    case isOPUS:
        prodstr="Opus";
        if(subversion == 48)
          subversion=0;
        break;  

    default:
        if (product>=0 && product<PrdctHdr.ElemCnt)
            prodstr = PRDCT_NM(product);
        else
            sprintf(prodstr=buf,"Product#%d",product);
        break;
    }
    sprintf(buf2,"%s %s %s %d.%02d",
                 MSG_TXT (M_REMOTE_USES), prodstr,
                 MSG_TXT (M_VERSION), version, subversion);
    status_line (buf2);
    strcpy(remote_mailer,buf2); /* TJW 960528 SRIF */
    return;
}

int 
next_minute ()
{
    time_t long_time;
    struct tm *tm;

    /* Get the current time into a structure */

    (void) unix_time (&long_time);
    tm = unix_localtime (&long_time);

    return ((60 - (tm->tm_sec % 60)) * 100 + 1);
}

void 
can_Janus (char *p)
{
    J_TYPESP j;

    janus_OK = 0;
    for (j = j_top; j != NULL; j = j->next)
    {
        if (strnicmp (p, j->j_match, strlen (j->j_match)) == 0)
        {
            janus_OK = 1;
            break;
        }
    }
}

int 
check_failed (char *fname, char *theirname, char *info, char *ourname)
{
    FILE *abortlog;
    char linebuf[64];
    char *p, *badname;
    int ret;

    ret = 0;
    if ((abortlog = fopen (fname, read_ascii)) == NULL)
    {
        (void) got_error (MSG_TXT (M_OPEN_MSG), fname);
    }
    else
    {
        while (!feof (abortlog))
        {
            linebuf[0] = '\0';
            if (!fgets ((p = linebuf), 64, abortlog))
                break;
            while (*p >= ' ')
                ++p;
            *p = '\0';
            p = strchr (linebuf, ' ');
            *p = '\0';
            if (!stricmp (linebuf, theirname))
            {
                p = strchr ((badname = ++p), ' ');
                *p = '\0';
                if (!stricmp (++p, info))
                {
                    (void) strcpy (ourname, badname);
                    ret = 1;
                    break;
                }
            }
        }
        (void) fclose (abortlog);
    }

    return (ret);
}

void 
add_abort (char *fname, char *rname, char *cname,
    char *cpath, char *info)
{
    FILE *abortlog;
    char namebuf[100];

    (void) strcpy (namebuf, cpath);
    (void) strcat (namebuf, "BadWaZOO.001");
    unique_name (namebuf);
    (void) rename (cname, namebuf);
    if ((abortlog = fopen (fname, append_ascii)) == NULL)
    {
        (void) got_error (MSG_TXT (M_OPEN_MSG), fname);
        (void) unlink (namebuf);
    }
    else
    {
        (void) fprintf (abortlog, "%s %s %s\n", rname, namebuf + strlen (cpath), info);
        (void) fclose (abortlog);
    }
}

void 
remove_abort (char *fname, char *rname)
{
    FILE *abortlog, *newlog;
    char namebuf[100];
    char linebuf[100];
    char *p;
    int c;

    if (!dexists (fname))
        return;

    if ((abortlog = fopen (fname, read_ascii)) == NULL)
    {
        (void) got_error (MSG_TXT (M_OPEN_MSG), fname);
    }
    else
    {
        (void) strcpy (namebuf, fname);
        (void) strcpy (namebuf + strlen (namebuf) - 1, "TMP");
        c = 0;
        if ((newlog = fopen (namebuf, write_ascii)) == NULL)
        {
            (void) got_error (MSG_TXT (M_OPEN_MSG), namebuf);
            (void) fclose (abortlog);
        }
        else
        {
            while (!feof (abortlog))
            {
                linebuf[0] = '\0';
                if (!fgets (linebuf, 64, abortlog))
                    break;
                p = linebuf;
                while (*p > ' ')
                    ++p;
                *p = '\0';
                if (stricmp (linebuf, rname))
                {
                    *p = ' ';
                    (void) fputs (linebuf, newlog);
                    ++c;
                }
            }
            (void) fclose (abortlog);
            (void) fclose (newlog);
            (void) unlink (fname);
            if (c)
                (void) rename (namebuf, fname);
            else
                (void) unlink (namebuf);
        }
    }
}

/*
 * The next routine is used in the count down timer during file
 * transfers, and were provided by Jon Sabol, along with the other code that
 * calls these routines.
 */

void 
elapse_time (int show)
{
    time_t lTimeT;
    time_t etime;
    long eh;
    long em;
    long es;
    char junk[256];

    show_alive();        /* RDH *** 05.14.96 12:15:12 */
    if (fullscreen && (un_attended || doing_poll))
    {
        (void) unix_time (&lTimeT);

        if (lTimeT < etm)
            lTimeT += 86400L;
        etime = lTimeT - etm;
        eh = etime / 3600L;
        em = (etime / 60L) - (eh * 60L);
        es = etime - (eh * 3600L) - (em * 60L);

        (void) sprintf (junk,
            "%s:  %02ld:%02ld:%02ld",
            MSG_TXT (M_ELAPSED),
            eh, em, es);
        sb_move_puts (settingswin, SET_TASK_ROW, SET_TIME_COL, junk);
        if(show)
            sb_show ();
    }
    IPC_Ping();
}

static int LOCALFUNC 
find_addr (char *node, ADDRP addr, char *d)
{
    int ret;

    ret = 1;
    if (alias[0].Node == 0xffff)
        addr->Zone = 0;
    else
        addr->Zone = alias[0].Zone;
    addr->Net = 0;
    addr->Node = 0;
    addr->Point = 0;
    addr->Domain = NULL;
    d[0] = '\0';
    if ((ret = sscanf (node, "%hd:%hd/%hd.%hd@%s",
                &(addr->Zone), &(addr->Net), &(addr->Node), &(addr->Point), d)) < 3)
    {
        if (alias[0].Node == 0xffff)
            addr->Zone = 0;
        else
            addr->Zone = alias[0].Zone;
        d[0] = '\0';
        if ((ret = sscanf (node, "%hd/%hd.%hd@%s",
                    &(addr->Net), &(addr->Node), &(addr->Point), d)) < 2)
        {
            addr->Net = alias[0].Net;
            d[0] = '\0';
            if (sscanf (node, "%hd.%hd@%s",
                    &(addr->Node), &(addr->Point), d) < 1)
            {
                if ((ret = sscanf (node, ".%hd@%s", &(addr->Point), d)) > 0)
                {
                    addr->Node = alias[0].Node;
                }
            }
            else if (ret == 1)
            {
                d[0] = '\0';
                (void) sscanf (node, "%hd@%s",
                    &(addr->Node), d);
            }
        }
        else if (ret == 2)
        {
            d[0] = '\0';
            (void) sscanf (node, "%hd/%hd@%s",
                &(addr->Net), &(addr->Node), d);
        }
    }
    else if (ret == 3)
    {
        d[0] = '\0';
        (void) sscanf (node, "%hd:%hd/%hd@%s",
            &(addr->Zone), &(addr->Net), &(addr->Node), d);
    }

    return (ret);
}

int 
parse_address (char *node, ADDRP addr)
{
    int ret;
    char d[100];

    ret = find_addr (node, addr, d);

    if (d[0] != '\0')
    {
        addr->Domain = add_domain (d);
    }
    else
        addr->Domain = NULL;

    return (ret);
}

int 
find_address (char *node, ADDRP addr)
{
    int i;
    int ret;
    char d[255];
    char *p;

    d[0] = '\0';
    p = skip_blanks (node);
    if (!isdigit (*p) && (*p != '.'))
    {
        (*userfunc) (p, addr);
        if ((addr->Net == 0xffff) || (addr->Node == 0xffff) || (addr->Zone == 0xffff))
        {
            ret = 0;
        }
        else
        {
            ret = 1;
        }
    }
    else
    {
        ret = find_addr (p, addr, d);
    }

    if (d[0] != '\0' && domain_name[0] != NULL)
    {
        addr->Domain = add_domain (d);
    }
    else
    {
        addr->Domain = domain_name[0];
        for (i = 0; i < num_domain_kludge; i++)
        {
            if (addr->Zone == domainkludge[i].zone)
            {
                addr->Domain = domain_name[domainkludge[i].domain];
                break;
            }
        }
    }

    return (ret);
}

char *
add_domain (char *d)
{
    char *p, *q;
    int i;

    for (i = 0; (p = domain_name[i]) != NULL; i++)
    {
        if (strnicmp (d, p, strlen (p)) == 0)
            return (p);
        q = domain_abbrev[i];
        if (q == NULL)          /* This can happen if no "Domain" line... */
            continue;
        if (strnicmp (d, q, strlen (q)) == 0)
            return (p);
    }

    if (i >= 49)
        return (NULL);

    domain_name[i] = strdup (d);
    return (domain_name[i]);
}

char *
find_domain (char *d)
{
    char *p, *q, *s;
    char c;
    int i, j, k;

    j = strlen (d);
    /* First see if we can find the whole domain name at the right */
    for (i = 0; (p = domain_name[i]) != NULL; i++)
    {
        k = strlen (p);
        if (k > j)
            continue;

        q = &(d[j - k]);
        if (strnicmp (q, p, k) == 0)
        {
            return (p);
        }
    }

    /* Ok, now see if we can find the abbreviated name at the right */
    for (i = 0; (p = domain_abbrev[i]) != NULL; i++)
    {
        k = strlen (p);
        if (k > j)
            continue;

        q = &(d[j - k]);
        if (strnicmp (q, p, k) == 0)
        {
            return (domain_name[i]);
        }
    }

    /* If there is a period in it, see if we can match the abbreviated name
      just before the period */
    if ((s = strrchr (d, '.')) != NULL)
    {
        c = *s;
        *s = '\0';
        j = strlen (d);
        for (i = 0; (p = domain_abbrev[i]) != NULL; i++)
        {
            k = strlen (p);
            if (k > j)
                continue;

            q = &(d[j - k]);
            if (strnicmp (q, p, k) == 0)
            {
                *s = c;
                return (domain_name[i]);
            }
        }
        *s = c;
    }

    return (NULL);
}

static char addr_str[100];

char *
Hex_Addr_Str (ADDRP a)
{
    if (a->Point != 0)
    {
        (void) sprintf (addr_str, "%04hx%04hx.PNT\\%08hx", a->Net, a->Node, a->Point);
    }
    else
        (void) sprintf (addr_str, "%04hx%04hx", a->Net, a->Node);
    return (addr_str);
}

char *
Full_Addr_Str (ADDRP a)
{
    char t1[10];
    char t2[30];
    char t3[10];
    char t4[50];

    if (a->Zone && !no_zones)
        (void) sprintf (t1, "%hu:", a->Zone);
    else
        t1[0] = '\0';

    (void) sprintf (t2, "%hu/%hu", a->Net, a->Node);

    if (a->Point)
        (void) sprintf (t3, ".%hu", a->Point);
    else
        t3[0] = '\0';

    if (a->Domain != NULL)
        (void) sprintf (t4, "@%s", a->Domain);
    else
        t4[0] = '\0';

    (void) strcpy (addr_str, t1);
    (void) strcat (addr_str, t2);
    (void) strcat (addr_str, t3);
    (void) strcat (addr_str, t4);
    return (addr_str);
}

unsigned short
crc_block (unsigned char *ptr, int count)
{
    register word crc;
    int i;

    for (crc = 0, i = 0; i < count; i++, ptr++)
    {
        crc = xcrc (crc, (byte) * ptr);
    }
    return (crc);
}

void 
Data_Check (XMDATAP xtmp, int mode)
{
    int i;
    unsigned char cs;
    unsigned char *cp;
    unsigned short cs1;

    /* If we are in checksum mode, just do it */
    if (mode == CHECKSUM)
    {
        cp = xtmp->data_bytes;
        cs = 0;
        for (i = 0; i < 128; i++)
        {
            cs += *cp++;
        }
        xtmp->data_check[0] = cs;
    }
    /* If we are in CRC mode, run the characters through the CRC calculator */
    else
    {
        cs1 = crc_block (xtmp->data_bytes, 128);
        xtmp->data_check[0] = (unsigned char) (cs1 >> 8);
        xtmp->data_check[1] = (unsigned char) (cs1 & 0xff);
    }
}

void 
wait_for_key (int ts)
{
    long t_idle;

    sb_show ();
    t_idle = timerset (ts);
    while (!timeup (t_idle) && (PEEKBYTE () == -1) && !(KEYPRESS ()))
        time_release ();
    if (KEYPRESS ())
        ts = (int) FOSSIL_CHAR ();
}

/* MMP 960419 timesync() should be called at the end of a session,
 * and if diff_trx variable is set, the the called_addr matches the
 * TimeSync address in the configuration file, we adjust the clock
 */

void timesync(void)
{
    time_t diff;

    diff = diff_trx;
    diff_trx = 0;

    if ( diff ) {

#ifdef OS_2
        DATETIME dt;
#endif /* OS_2 */
#ifdef DOS16
        struct dostime_t dtime;
        struct dosdate_t ddate;
#endif /* DOS16 */
        time_t newtime;
        struct tm t;

        if (max_diff_trx)
            if ( labs((long)diff) > max_diff_trx )
                return;

        unix_time (&newtime);
        newtime += diff;

        /* Important! Make a copy of the tm struct - status_line() will read
         * the current time into it the static buffer used by localtime()
         */

        t = *unix_localtime (&newtime);
        status_line(MSG_TXT (M_SYNCMYCLOCK),
            t.tm_year, t.tm_mon+1, t.tm_mday,
            t.tm_hour, t.tm_min, t.tm_sec);
#ifdef OS_2
#define TIMESET
        DosGetDateTime(&dt);        /* Get timezone right */
        dt.hours      = (unsigned char)t.tm_hour;
        dt.minutes    = (unsigned char)t.tm_min;
        dt.seconds    = (unsigned char)t.tm_sec;
        dt.hundredths = 0;
        dt.day        = (unsigned char)t.tm_mday;
        dt.month      = (unsigned char)(t.tm_mon+1);
        dt.year       = (unsigned short)(1900+t.tm_year);
        dt.weekday    = (unsigned char)t.tm_wday;
        DosSetDateTime(&dt);
#endif /* OS_2 */
#ifdef DOS16
#define TIMESET
        dtime.hour      = (unsigned char)t.tm_hour;
        dtime.minute    = (unsigned char)t.tm_min;
        dtime.second    = (unsigned char)t.tm_sec;
        dtime.hsecond   = 0;
        ddate.day       = (unsigned char)t.tm_mday;
        ddate.month     = (unsigned char)(t.tm_mon+1);
        ddate.year      = (unsigned short)(80+t.tm_year);
        ddate.dayofweek = (unsigned char)t.tm_wday;
        _dos_settime(&dtime);
        _dos_setdate(&ddate);
#endif /* DOS16 */
#ifdef _WIN32
#define TIMESET
/* NOTE:
   there is no support for timesync in WIN32 Version yet */
#endif /* _WIN32 */

#ifndef TIMESET
#       error Need some tweaking here in timesync() to set the date/time
#endif
    }
}

void cleanup_req(char *fname){ /* cleanup *.REQ file from size information */
   FILE *fpi, *fpo;
   char tmpfname[PATHLEN];
   char *p1,*p2;
   char line[1024];
   
   strcpy(tmpfname,fname);
   strcpy(tmpfname+strlen(tmpfname)-3,"RQQ"); /* use *.RQQ as tempfile */
   
   unlink(tmpfname);
   
   if(rename(fname, tmpfname)){
      status_line("!Unable to rename %s to %s.", fname, tmpfname);
      return;
   }

   if((fpo=fopen(fname, write_ascii))==NULL){
      status_line("!Can't open %s.", fname);
      return;
   }
   if((fpi=fopen(tmpfname, read_ascii))==NULL){
      status_line("!Can't open %s.", tmpfname);
      fclose(fpo);
      return;
   }

   while(fgets(line, sizeof(line), fpi)){ /* read line from input file */
      /* status_line("!Read RQQ:'%s'", line); */
      p2=line;
      p1=skip_to_blank(p2);            /* skip to position after filename */
      p2=skip_blanks(p1);              /* skip to begin of next word      */
      while(*p2 && *p2!='$'){
         p1=skip_to_blank(p2);
         p2=skip_blanks(p1);
      }
      if(*p2=='$'){                    /* do we have size information ? */
          p2=skip_to_blank(p2);        /* skip it. */
          strcpy(p1,p2);               /* copy the stuff at the end over it */
      }
      /* status_line("!Write REQ:'%s'", line); */
      fputs(line, fpo);                /* write line to output file         */
   }
   fclose(fpi);
   fclose(fpo);
   unlink(tmpfname);
}

static void killblanks(char *s){
  char *p=s;
  while(*p){
    if(isspace(*p))
      strcpy(p,p+1);
    else
      p++;
  }
}

int eval(char *e){ /* evaluate expression -> true = 1, false = 0, invalid=-1 */
  int ret;
  char *op, *left, *right;
  killblanks(e);                                  /* remove blanks, tabs.. */
  strlwr(e);                                      /* make all lowercase */
  if((op=strstr(e,"=="))!=NULL){                  /* x == y */
    left=e;
    *op++ = '\0';
    right=++op;
    ret= (strcmp(left,right)==0);
  }else
  if((op=strstr(e,"!="))!=NULL){                  /* x != y */
    left=e;
    *op++ = '\0';
    right=++op;
    ret= (strcmp(left,right)!=0);
  }else
  if(!strcmp(e,"true") || !strcmp(e,"common")){   /* true / common */
    ret=1;
  }else
  if(!strcmp(e,"false") || !strcmp(e,"ignore")){  /* false / ignore */
    ret=0;
  }else
    ret=-1;                                       /* syntax error */ 
  return ret;
}

/* TJW 960928: some compilers use ANSI style time_t, which begins at 1.1.1900 */
/* Binkley wants time_t beginning at 1.1.1970 (UNIX style).                   */
/* So Binkley will ever call unix_* functions and they will be remapped via   */
/* #define either directly to the library functions (if they are UNIX style)  */
/* or to the adaption functions below for ANSI style libraries...             */

#if defined(ANSI_TIME_T)

int
ansi_stat(const char *fname, struct stat *buffer){
  int ret;
  ret = stat(fname, buffer);
  buffer->st_ctime -= ANSI_TIME_T_DELTA;
  buffer->st_mtime -= ANSI_TIME_T_DELTA;
  buffer->st_atime -= ANSI_TIME_T_DELTA;
  return ret;
}

int
ansi_fstat(int handle, struct stat *buffer){
  int ret;
  ret = fstat(handle, buffer);
  buffer->st_ctime -= ANSI_TIME_T_DELTA;
  buffer->st_mtime -= ANSI_TIME_T_DELTA;
  buffer->st_atime -= ANSI_TIME_T_DELTA;
  return ret;
}

time_t
ansi_time(time_t *timeptr){
  time_t t;
  t = time(timeptr);
  if(t!=(time_t)-1){
    if(timeptr)
      *timeptr -= ANSI_TIME_T_DELTA;
    t -= ANSI_TIME_T_DELTA;
  }
  return t;
}

time_t
ansi_mktime(struct tm *time){
  time_t t;
  t = mktime(time);
  if(t!=(time_t)-1)
    t -= ANSI_TIME_T_DELTA;
  return t;
}

struct tm *
ansi_localtime(const time_t *timeval){
  time_t t;
  t = *timeval + ANSI_TIME_T_DELTA;
  return localtime(&t);
}

struct tm *
ansi_gmtime(const time_t *timeval){
  time_t t;
  t = *timeval + ANSI_TIME_T_DELTA;
  return gmtime(&t);
}

char *
ansi_ctime(const time_t *timeval){
  time_t t;
  t = *timeval + ANSI_TIME_T_DELTA;
  return ctime(&t);
}

int
ansi_utime(char *fname, struct utimbuf *buf){
  struct utimbuf abuf, *p;
  p = buf;
  if(p){
    abuf = *p;
    p = &abuf;
    p->modtime   += ANSI_TIME_T_DELTA;
    p->UT_ACTIME += ANSI_TIME_T_DELTA;
  }
  return utime(fname, p);
}

#endif
