/****************************************************************************/
/*  this module was written by tom schlangen                                */
/****************************************************************************/
#ifdef  PROMETER

/* Include this file before any other includes or defines! */
#include "includes.h"

/****************************************************************************/
/*  local constants                                                         */
/****************************************************************************/

#define PROMETER_WIDTH  23
#define PROMETER_STRLEN PROMETER_WIDTH + 3
#define PROMETER_BAR_X 12                   /*  column of prometer bars     */
#define PROMETER_MSG_X 0                    /*  column of explanatory msgs  */
#define PROMETER_CHARS_ASC  "[* ]"          /*  bar chars for boxtype==0    */
#define PROMETER_CHARS_GRA  "[۰]"          /*  bar chars for boxtype!=0    */

#define PROMETER_INVALID    -1L

#define PROMETER_ROW_HDR    0               /*  row for header / percentage */
#define PROMETER_ROW_TX_CUR 1               /*  row for tx current file     */
#define PROMETER_ROW_RX_CUR 3               /*  row for tx current file     */
#define PROMETER_ROW_TX_MOH 2               /*  row for tx mail on hold     */
#define PROMETER_ROW_RX_MOH 4               /*  row for rx mail on hold     */

/****************************************************************************/
/*  local static data                                                       */
/****************************************************************************/
static  BINK_SAVEP holdwinsave;
static  char  acbarchars[5];
static  char  acprometer[PROMETER_STRLEN];
static  long  txmohtotal, txmohtogo, rxmohtotal, rxmohtogo;
static  int   otxc, orxc,otxm,orxm;

/****************************************************************************/
/*  local prototypes                                                        */
/****************************************************************************/
char *prometer_clr(void);
int prometer_comp(long total, long done);
char *prometer_str(int i);

/****************************************************************************/
/*  functions                                                               */
/****************************************************************************/
char *prometer_clr(void)
{
    /*  simply blank out the string */
    register int i;
    register char *p=acprometer;

    for (i=PROMETER_WIDTH+2;i>0;i--)
        *p++=' ';
    *p='\0';
    return (acprometer);
}

void prometer_show(long total, long done, int what)
{
    int row=0;
    int i;
    if (un_attended && fullscreen)
    {
        switch (what)
        {
            case  PROMETER_TX_CUR:
                row = PROMETER_ROW_TX_CUR;
                if (done < total)
                {
                    if (otxc!=(i=prometer_comp(total, done)))
                    {
                        otxc=i;
                        sb_move (holdwin, row, PROMETER_BAR_X);
                        sb_puts (holdwin, prometer_str(i));
                    }
                }
                else
                {
                    sb_move (holdwin, row, PROMETER_BAR_X);
                    sb_puts (holdwin, prometer_clr());
                }
                break;
            case  PROMETER_RX_CUR:
                row = PROMETER_ROW_RX_CUR;
                if (done < total)
                {
                    if (orxc!=(i=prometer_comp(total, done)))
                    {
                        orxc=i;
                        sb_move (holdwin, row, PROMETER_BAR_X);
                        sb_puts (holdwin, prometer_str(i));
                    }
                }
                else
                {
                    sb_move (holdwin, row, PROMETER_BAR_X);
                    sb_puts (holdwin, prometer_clr());
                }
                break;
            case  PROMETER_TX_MOH:
                row = PROMETER_ROW_TX_MOH;
                sb_move (holdwin, row, PROMETER_BAR_X);
                if (total == PROMETER_INVALID || done == PROMETER_INVALID)
                    sb_puts (holdwin, MSG_TXT (M_NOT_AVAIL));
                else if (done <= 0L)
                    sb_puts (holdwin, prometer_clr());
                else
                {
                    if (otxm!=(i=prometer_comp(total, done)))
                    {
                        otxm=i;
                        sb_puts (holdwin, prometer_str(i));
                    }
                }
                break;
            case  PROMETER_RX_MOH:
                row = PROMETER_ROW_RX_MOH;
                sb_move (holdwin, row, PROMETER_BAR_X);
                if (total == PROMETER_INVALID || done == PROMETER_INVALID)
                    sb_puts (holdwin, MSG_TXT (M_NOT_AVAIL));
                else if (done <= 0L)
                    sb_puts (holdwin, prometer_clr());
                else
                {
                    if (orxm!=(i=prometer_comp(total, done)))
                    {
                        orxm=i;
                        sb_puts (holdwin, prometer_str(i));
                    }
                }
                break;
        }
    }
}

void prometer_set_moh(long moh, int what)
{
    if (un_attended && fullscreen)
    {
        switch (what)
        {
            case  PROMETER_TX_MOH:
                txmohtotal=txmohtogo=moh;
                prometer_show(txmohtotal, txmohtogo, PROMETER_TX_MOH);
                break;
            case  PROMETER_RX_MOH:
                rxmohtotal=rxmohtogo=moh;
                prometer_show(rxmohtotal, rxmohtogo, PROMETER_RX_MOH);
                break;
        }
    }
}

void prometer_dec_moh(long l, int what)
{
    if (un_attended && fullscreen)
    {
        switch (what)
        {
            case  PROMETER_TX_MOH:
                txmohtogo-=l;
                prometer_show(txmohtotal, txmohtogo, PROMETER_TX_MOH);
                break;
            case  PROMETER_RX_MOH:
                rxmohtogo-=l;
                prometer_show(rxmohtotal, rxmohtogo, PROMETER_RX_MOH);
                break;
        }
    }
}

int prometer_comp(long total, long done)
{
    int i;

    if(done>total)                      /* TJW 960707 catch prob with FREQs */
       total=done;                      /* until there is a better solution */

    if (total == 0)                         
        i=0;                                /*  catch division by zero      */
    else
        i=abs((int)((done*PROMETER_WIDTH)/total));  /*  compute `done' chars*/
    return i;
}

char    *prometer_str(int i)
{
    register char *p=acprometer;
    int j=PROMETER_WIDTH-i;                 /*  remaining `todo' chars      */
    *p++=acbarchars[0];                     /*  write left bar delimiter    */
    for(;i>0;i--)                           /*  write chars for `done'      */
        *p++=acbarchars[1];
    for(;j>0;j--)                           /*  write chars for `todo'      */
        *p++=acbarchars[2];
    *p++=acbarchars[3];                     /*  write right bar delimiter   */
    *p='\0';                                /*  terminate string            */
    return (acprometer);                    /*  return string adress        */
}

void prometer_init(void)
{
    if (un_attended && fullscreen)
    {
        /*  initialize some local data                                      */
        txmohtotal=txmohtogo=rxmohtotal=rxmohtogo=PROMETER_INVALID;
        orxc=otxc=orxm=otxm= (int) PROMETER_INVALID;

        /*  set up bar characters according to frametype in use             */
        strcpy(acbarchars, boxtype==0 ? PROMETER_CHARS_ASC:PROMETER_CHARS_GRA);

        /*  save hold window contents                                       */
        holdwinsave=sb_save (1, 41, TOP_PARTS_ROWS-1, (short) (SB_COLS - 41));

        /*  draw the box                                                    */
        sb_box (holdwin, boxtype, colors.frames);

        /*  give it a title                                                 */
        sb_caption (holdwin, MSG_TXT (M_PROMETER_CAPTION));

        /*  blank out window contents                                       */
        sb_fill (holdwin, ' ', colors.hold);

        sb_move (holdwin, PROMETER_ROW_HDR, 0);
        sb_puts (holdwin, MSG_TXT (M_PROMETER_PERCENT));

        sb_move (holdwin, PROMETER_ROW_TX_CUR, PROMETER_MSG_X);
        sb_puts (holdwin, MSG_TXT (M_PROMETER_TX_CUR));

        sb_move (holdwin, PROMETER_ROW_TX_MOH, PROMETER_MSG_X);
        sb_puts (holdwin, MSG_TXT (M_PROMETER_TX_MOH));

        sb_move (holdwin, PROMETER_ROW_RX_CUR, PROMETER_MSG_X);
        sb_puts (holdwin, MSG_TXT (M_PROMETER_RX_CUR));

        sb_move (holdwin, PROMETER_ROW_RX_MOH, PROMETER_MSG_X);
        sb_puts (holdwin, MSG_TXT (M_PROMETER_RX_MOH));
    }
}

void prometer_done(void)
{
    if (un_attended && fullscreen)
    {
        /*  restore old holdwin contents    */
        sb_restore(holdwinsave);
    }
}

#endif
