/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                                                          */
/*      ------------         Bit-Bucket Software, Co.                       */
/*      \ 10001101 /         Writers and Distributors of                    */
/*       \ 011110 /          Freely Available<tm> Software.                 */
/*        \ 1011 /                                                          */
/*         ------                                                           */
/*                                                                          */
/*              (C) Copyright 1987-96, Bit Bucket Software Co.              */
/*                                                                          */
/*           This module was originally written by Bob Hartman              */
/*                                                                          */
/*                     BinkleyTerm Scheduler Routines                       */
/*                                                                          */
/*                                                                          */
/*    For complete  details  of the licensing restrictions, please refer    */
/*    to the License  agreement,  which  is published in its entirety in    */
/*    the MAKEFILE and BT.C, and also contained in the file LICENSE.260.    */
/*                                                                          */
/*    USE  OF THIS FILE IS SUBJECT TO THE  RESTRICTIONS CONTAINED IN THE    */
/*    BINKLEYTERM  LICENSING  AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF    */
/*    THIS  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES,  OR IF YOU DO    */
/*    NOT HAVE THESE FILES,  YOU  SHOULD  IMMEDIATELY CONTACT BIT BUCKET    */
/*    SOFTWARE CO.  AT ONE OF THE  ADDRESSES  LISTED BELOW.  IN NO EVENT    */
/*    SHOULD YOU  PROCEED TO USE THIS FILE  WITHOUT HAVING  ACCEPTED THE    */
/*    TERMS  OF  THE  BINKLEYTERM  LICENSING  AGREEMENT,  OR  SUCH OTHER    */
/*    AGREEMENT AS YOU ARE ABLE TO REACH WITH BIT BUCKET SOFTWARE, CO.      */
/*                                                                          */
/*                                                                          */
/* You can contact Bit Bucket Software Co. at any one of the following      */
/* addresses:                                                               */
/*                                                                          */
/* Bit Bucket Software Co.        FidoNet  1:104/501, 1:343/491             */
/* P.O. Box 460398                AlterNet 7:42/1491                        */
/* Aurora, CO 80046               BBS-Net  86:2030/1                        */
/*                                Internet f491.n343.z1.fidonet.org         */
/*                                                                          */
/* Please feel free to contact us at any time to share your comments about  */
/* our software and/or licensing policies.                                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/

/* Include this file before any other includes or defines! */

#include "includes.h"

/**************************************************************************/
/*** This MUST be exactly 16 total bytes including the terminating null ***/
/*** or the routines read_sched() and write_sched() must be changed!!!! ***/
/**************************************************************************/
static char *BinkSched = PRDCT_SHRT "Schedule 11";  /* Version of scheduler   */

long custom_date(struct tm *tm){
   return 366L*(tm->tm_year-80) + tm->tm_yday;
}

void 
find_event (void)
{
    time_t long_time;
    struct tm *tm;

    int cur_day;
    int our_time;
    int i;
    char cmds[PATHLEN];
    long customdate;
    
    BINK_EVENT evt;

    /* Get the current time into a structure */

    (void) unix_time (&long_time);
    tm = unix_localtime (&long_time);
    customdate = custom_date(tm);
    
    cur_day = 1 << tm->tm_wday;

    our_time = (tm->tm_hour % 24) * 60 + (tm->tm_min % 60);

    cur_event = -1;

    if (customdate != hist.which_date)
    {
        write_stats ();
        (void) memset (&hist, 0, sizeof (HISTORY));
        start_hist = hist;
        hist.which_date = customdate;
        do_today (1);
    }

    /* Go through the events from top to bottom */
    for (i = 0; i < num_events; i++)
    {
        if (our_time >= e_ptrs[i].minute)
        {
            evt = e_ptrs[i];

            if ((cur_day & evt.days) && !(evt.behavior&MAT_COSTEVENT) &&
                ((!evt.day) || (evt.day == (char) tm->tm_mday)) &&
                ((!evt.month) || (evt.month == (char) tm->tm_mon+1)))
            {
                if (((our_time - evt.minute) < evt.length) ||
                    ((our_time == evt.minute) && (evt.length == 0)) ||
                    ((evt.behavior & MAT_FORCED) && (evt.last_ran_date != customdate)))
                {
                    /* Are we not supposed to force old events */
                    if (((our_time - evt.minute) > evt.length) && (noforce))
                    {
                        e_ptrs[i].last_ran_date = customdate;
                        continue;
                    }

                    if (evt.last_ran_date != customdate)
                    {
                        cur_event = i;
                        do_ready (M_READY_WAITING);
                        status_line (MSG_TXT (M_STARTING_EVENT), i + 1);

                        if (!blank_on_key)
                            screen_blank = 0;

                        more_mail = 1;

                        /* Mark that this one is running */
                        e_ptrs[i].last_ran_date = customdate;

                        /*
                         * Mark that we have not yet skipped it. After all, it just
                         * started!
                         */
                        e_ptrs[i].behavior &= ~MAT_SKIP;

                        /* Write out the schedule */
                        write_sched ();

                        /* Do this first ... if any $.. to kill, do it */
                        if (e_ptrs[i].behavior & MAT_KILLBAD)
                            kill_bad ();


                        /* If we are supposed to exit, then do it */
                        if (evt.errlevel[0])
                        {
                            if (errlvlshell[evt.errlevel[0] - 1] == NULL)
                            {
                                status_line (MSG_TXT (M_EVENT_EXIT), evt.errlevel[0]);
                                screen_blank = 0;
                                errl_exit (evt.errlevel[0]);
                            }
                            else
                            {
                                status_line (MSG_TXT (M_EVENT_SHELL), evt.errlevel[0]);
                                errl_shell (evt.errlevel[0]);
                            }
                        }
                        else if (packer != NULL || cleanup != NULL)
                        {
                            if (!blank_on_key)
                                screen_blank = 0;
                            status_line (MSG_TXT (M_CLEAN_PACK));
                            mdm_init (modem_busy);
                            exit_DTR ();
                            screen_clear ();
                            vfossil_cursor (1);
                            if (cleanup != NULL)
                            {
                                (void) strcpy (cmds, cleanup);
                                if (i >= 0)
                                    (void) strcat (cmds, evt.cmd);
                                b_spawn (cmds);
                            }
                            if (packer != NULL)
                            {
                                (void) strcpy (cmds, packer);
                                if (i >= 0)
                                    (void) strcat (cmds, evt.cmd);
                                b_spawn (cmds);
                            }
                            if (fullscreen)
                            {
                                screen_clear ();
                                sb_dirty ();
                                opening_banner ();
                                mailer_banner ();
                            }
                            RAISE_DTR ();
                            mdm_init (modem_init);
                            status_line (MSG_TXT (M_AFTER_CLEAN_PACK));
                            waitfor_line = timerset ((unsigned int) 6000);
                        }

                        cur_event = i;
                        max_connects = evt.with_connect;
                        max_noconnects = evt.no_connect;
                        set_up_outbound (0);
                    }
                    else
                    {
                        /* Don't do events that have been exited already */
                        if (evt.behavior & MAT_SKIP)
                            continue;
                    }

                    cur_event = i;

                    if (evt.behavior & MAT_NOREQ)
                    {
                        matrix_mask &= ~TAKE_REQ;
                        no_requests = 1;
                    }
                    else
                    {
                        matrix_mask |= TAKE_REQ;
                        no_requests = 0;
                    }

                    if (evt.behavior & MAT_NOOUTREQ)
                    {
                        requests_ok = 0;
                    }
                    else
                    {
                        requests_ok = 1;
                    }

                    max_connects = evt.with_connect;
                    max_noconnects = evt.no_connect;

                    break;
                }
            }
        }
    }
}

BINK_EVENTP 
find_costevent (long t)
{
    struct tm *tm;
    time_t long_time=t;
    
    int cur_day;
    int our_time;
    int i;

    BINK_EVENTP evt, ret;

    tm = unix_localtime (&long_time);

    cur_day = 1 << tm->tm_wday;

    our_time = (tm->tm_hour % 24) * 60 + (tm->tm_min % 60);

    ret = NULL;
    /* Go through the events from top to bottom */
    for (i = 0; i < num_events; i++)
    {
        if (our_time >= e_ptrs[i].minute)
        {
            evt = &e_ptrs[i];

            if ((cur_day & evt->days) && (evt->behavior&MAT_COSTEVENT) &&
                ((!evt->day) || (evt->day == (char) tm->tm_mday)) &&
                ((!evt->month) || (evt->month == (char) tm->tm_mon+1)))
            {
                if (((our_time - evt->minute) < evt->length)){
                    ret = evt;
                    break;
                }
            }
        }
    }
    return ret;
}

void 
read_sched (void)
{
    char temp1[PATHLEN], temp2[PATHLEN];
    struct stat buffer1, buffer2;

    HFILE stream;
    unsigned long got;

    sprintf(temp1, "%s%s.S%02X", task_dir, PRDCT_PRFX, TaskNumber); /* TJW 960414 */
    sprintf(temp2, "%s%s.EVT",   BINKpath, PRDCT_PRFX);             /* TJW 960414 */

    if (unix_stat (temp1, &buffer1) == -1)
    {
        return;
    }

    if (unix_stat (temp2, &buffer2) == -1)
    {
        (void) strcpy (temp2, BINKpath);
        (void) strcat (temp2, config_name);
        if (unix_stat (temp2, &buffer2) == -1)
        {
            return;
        }
    }

    if ((buffer1.st_mtime < buffer2.st_mtime) || /* TJW 960420: mtime, not atime !!! */
        (buffer1.st_size < sizeof (BINK_EVENT)))
    {
        printf("Re-compiling events, %s is older than %s!\n", temp1, temp2); 
        return;
    }

    if ((stream = (HFILE) open (temp1, O_RDONLY | O_BINARY)) == (HFILE) - 1)
    {
        printf("Can't open compiled events %s!\n", temp1); 
        return;
    }

    temp1[0] = '\0';
    (void) read (stream, temp1, 16);
    if (strcmp (temp1, BinkSched) != 0)
    {
        (void) close (stream);
        printf("Invalid scheduler version in %s!\n", temp1); 
        return;
    }

    (void) read (stream, (char *) &hist, (int) sizeof (HISTORY));
    start_hist = hist;

    /* Note, we do not use "got". Some platforms want it to be a long
       and some want it to be a int. We don't care, so we make it a
       long (see above) so there's enough room, then use a void * so
       the compiler won't complain about an argument we don't even want */

    (void) _dos_read (stream, (char far *) e_ptrs,
        (USHORT) (buffer1.st_size - 16 - sizeof (HISTORY)), (void *)&got);

    got_sched = 1;

    num_events = (int) ((buffer1.st_size - 16 - sizeof (HISTORY)) / sizeof (BINK_EVENT));

    if(close(stream)!=0)                      /* TJW 960420 */ 
       printf("Error closing event file *.S%02X!\n", TaskNumber); 
    return;
}

void 
write_sched (void)
{
    char temp1[PATHLEN], temp2[PATHLEN];
    int i;
    struct stat buffer1;
    struct utimbuf times;
    long t;

    HFILE stream;

    unsigned long got;

    /* Get the current time */
    t = unix_time (NULL);

    sprintf(temp1, "%s%s.S%02X", task_dir, PRDCT_PRFX, TaskNumber); /* TJW 960414 */
    sprintf(temp2, "%s%s.EVT",   BINKpath, PRDCT_PRFX);             /* TJW 960414 */

    /* Get the current stat for .Evt file */

    if (unix_stat (temp2, &buffer1) == -1)
    {
        (void) strcpy (temp2, BINKpath);
        (void) strcat (temp2, config_name);
        if (unix_stat (temp2, &buffer1) == -1)
        {
            return;
        }
    }

    /*
    * If it is newer than current time, we have a problem and we must
    * reset the file date - yucky, but it will probably work
    */
    if ((time_t) t < buffer1.st_mtime)
    {
        times.UT_ACTIME = times.modtime = buffer1.st_mtime;
        status_line (MSG_TXT (M_DATE_PROBLEM));
    }
    else
        times.UT_ACTIME = times.modtime = t;

    if ((stream = (HFILE) open (temp1, O_CREAT | O_RDWR | O_BINARY, S_IREAD | S_IWRITE)) == (HFILE) - 1)
    {
        return;
    }

    (void) write (stream, BinkSched, 16);
    (void) write (stream, (char *) &hist, (int) sizeof (HISTORY));

    for (i = 0; i < num_events; i++)
    {
        /* If it is skipped, but not dynamic, reset it */
        if ((!(e_ptrs[i].behavior & MAT_COSTEVENT)) &&
            (e_ptrs[i].behavior & MAT_SKIP) &&
            (!(e_ptrs[i].behavior & MAT_DYNAM)))
        {
            e_ptrs[i].behavior &= ~MAT_SKIP;
        }

    }

    /* Note, we do not use "got". Some platforms want it to be a long
       and some want it to be a int. We don't care, so we make it a
       long (see above) so there's enough room, then use a void * so
       the compiler won't complain about an argument we don't even want */

    (void) _dos_write (stream, (char far *) e_ptrs,
        (USHORT) (num_events * sizeof (BINK_EVENT)), (void *)&got);

    (void) close (stream);

    (void) unix_utime (temp1, (UTIMBUF *) & times);

    return;
}

void 
write_stats (void)
{
    char temp1[PATHLEN];
    FILE *fp;

    sprintf(temp1, "%s%s.D%02X", task_dir, PRDCT_PRFX, TaskNumber); /* TJW 960414 */

    if ((fp=fopen (temp1, write_binary)) != NULL){
      (void) fwrite (&hist, (int) sizeof (HISTORY), 1, fp);
      (void) fclose (fp);
    }
    return;
}

int 
time_to_next (int skip_bbs)
{
    time_t long_time;
    struct tm *tm;

    int cur_day;
    int our_time;
    int i;
    int time_to;
    int guess;
    int nmin;

    /* Get the current time into a structure */

    (void) unix_time (&long_time);
    tm = unix_localtime (&long_time);

    our_time = tm->tm_hour * 60 + tm->tm_min;

    next_event = -1;
    cur_day = 1 << tm->tm_wday;

    /* A ridiculous number */
    time_to = 3000;

    /* Go through the events from top to bottom */
    for (i = 0; i < num_events; i++)
    {
        /* If it is a costevent, skip it */
        if (e_ptrs[i].behavior & MAT_COSTEVENT)
            continue;

        /* If it is the current event, skip it */
        if (cur_event == i)
            continue;

        /* If it is a BBS event, skip it */
        if (skip_bbs && e_ptrs[i].behavior & MAT_BBS)
            continue;

        /* If it was already run today, skip it */
        if (e_ptrs[i].last_ran_date == custom_date(tm))
            continue;

        /* If it doesn't happen today, skip it */
        if (!(e_ptrs[i].days & cur_day))
            continue;

        /* If not this day of the month, skip it */
        if (e_ptrs[i].day && (e_ptrs[i].day != (char) tm->tm_mday))
            continue;

        /* If not this month of the year, skip it */
        if (e_ptrs[i].month && (e_ptrs[i].month != (char) tm->tm_mon+1))
            continue;

        /* If it is earlier than now, skip it unless it is forced */
        if (e_ptrs[i].minute <= our_time)
        {
            if (!(e_ptrs[i].behavior & MAT_FORCED))
                continue;

            /* Hmm, found a forced event that has not executed yet */
            /* Give the guy 2 minutes and call it quits */
            guess = 2;
        }
        else
        {
            /* Calculate how far it is from now */
            guess = e_ptrs[i].minute - our_time;
        }

        /* If less than closest so far, keep it */
        if (time_to > guess)
        {
            time_to = guess;
            next_event = i;
        }
    }

    /* If we still have nothing, then do it again, starting at midnight */
    if (time_to > 1440)
    {
        /* Calculate here to midnight */
        nmin = 1440 - our_time;

        /* Go to midnight */
        our_time = 0;

        /* Go to the next possible day */
        cur_day = (int) (((unsigned) cur_day) << 1);
        if (cur_day > DAY_SATURDAY)
            cur_day = DAY_SUNDAY;

        /* Go through the events from top to bottom */
        for (i = 0; i < num_events; i++)
        {
            /* If it is a costevent, skip it */
            if (e_ptrs[i].behavior & MAT_COSTEVENT)
               continue;

            /* If it is a BBS event, skip it */
            if (skip_bbs && e_ptrs[i].behavior & MAT_BBS)
                continue;

            /* If it doesn't happen today, skip it */
            if (!(e_ptrs[i].days & cur_day))
                continue;

            /* If not this day of the month, skip it */
            if (e_ptrs[i].day && (e_ptrs[i].day != (char) tm->tm_mday))
                continue;

            /* If not this month of the year, skip it */
            if (e_ptrs[i].month && (e_ptrs[i].month != (char) tm->tm_mon+1))
                continue;

            /* Calculate how far it is from now */
            guess = e_ptrs[i].minute + nmin;

            /* If less than closest so far, keep it */
            if (time_to > guess)
            {
                time_to = guess;
                next_event = i;
            }
        }
    }

    if (time_to > 1440)
        time_to = 1440;

    if (skip_bbs && (time_to < 1))
        time_to = 1;

    return (time_to);
}
