/*--------------------------------------------------------------------------*/
/*                                                                          */
/*                                                                          */
/*      ------------         Bit-Bucket Software, Co.                       */
/*      \ 10001101 /         Writers and Distributors of                    */
/*       \ 011110 /          Freely Available<tm> Software.                 */
/*        \ 1011 /                                                          */
/*         ------                                                           */
/*                                                                          */
/*              (C) Copyright 1987-96, Bit Bucket Software Co.              */
/*                                                                          */
/*                  This module was adapted by Bill Andrus                  */
/*                With the usual kibitzing by Vince Perriello               */
/*              BinkleyTerm Version7 Nodelist processing module             */
/*                                                                          */
/*                                                                          */
/*    For complete  details  of the licensing restrictions, please refer    */
/*    to the License  agreement,  which  is published in its entirety in    */
/*    the MAKEFILE and BT.C, and also contained in the file LICENSE.260.    */
/*                                                                          */
/*    USE  OF THIS FILE IS SUBJECT TO THE  RESTRICTIONS CONTAINED IN THE    */
/*    BINKLEYTERM  LICENSING  AGREEMENT.  IF YOU DO NOT FIND THE TEXT OF    */
/*    THIS  AGREEMENT IN ANY OF THE  AFOREMENTIONED FILES,  OR IF YOU DO    */
/*    NOT HAVE THESE FILES,  YOU  SHOULD  IMMEDIATELY CONTACT BIT BUCKET    */
/*    SOFTWARE CO.  AT ONE OF THE  ADDRESSES  LISTED BELOW.  IN NO EVENT    */
/*    SHOULD YOU  PROCEED TO USE THIS FILE  WITHOUT HAVING  ACCEPTED THE    */
/*    TERMS  OF  THE  BINKLEYTERM  LICENSING  AGREEMENT,  OR  SUCH OTHER    */
/*    AGREEMENT AS YOU ARE ABLE TO REACH WITH BIT BUCKET SOFTWARE, CO.      */
/*                                                                          */
/*                                                                          */
/* You can contact Bit Bucket Software Co. at any one of the following      */
/* addresses:                                                               */
/*                                                                          */
/* Bit Bucket Software Co.        FidoNet  1:104/501, 1:343/491             */
/* P.O. Box 460398                AlterNet 7:42/1491                        */
/* Aurora, CO 80046               BBS-Net  86:2030/1                        */
/*                                Internet f491.n343.z1.fidonet.org         */
/*                                                                          */
/* Please feel free to contact us at any time to share your comments about  */
/* our software and/or licensing policies.                                  */
/*                                                                          */
/*--------------------------------------------------------------------------*/

/* Include this file before any other includes or defines! */

#include "includes.h"

static int LOCALFUNC btree (void);
static int LOCALFUNC _btree (long);
static struct _ndx far *LOCALFUNC get7node (FILE *, long, struct _ndx far *);
static void LOCALFUNC unpk (char *instr, char *outp, int many);
static int LOCALFUNC addr_compare (void *, void *, int);
static int LOCALFUNC name_compare (void *, void *, int);
static void LOCALFUNC set_vars(PV7NODE, ADDRP);

extern int get_nodelist_name (ADDRP);

extern char *nodelist_base;
extern char *nodelist_name;

static struct _ndx *noderef = NULL;    /* index file             */
static char aline[160];

static FILE *idx_fp, *dat_fp;

static int find_first;         /* is this findfirst or findnext ? */

static ADDR searchaddr;
static char searchname[40];

static void *desired;
static int desired_len;
static int (LOCALFUNC * compare_fn) (void *key, void *desired, int len);

static PV7RCB pv7rcb;

static PV7NODEL beginp, lastp, actp;


int  V7Initialize(void){         /* allocate memory */

   if (node_index == NULL)
       node_index = _fmalloc (sizeof (struct _ndx));

   if (noderef == NULL)
       noderef = _fmalloc (sizeof (struct _ndx));

   if (!node_index || !noderef)
   {
       status_line (MSG_TXT(M_NODELIST_MEM));
       return V7NL_ERROR;
   }

   return V7NL_SUCCESS;
}

void V7Finish(void){             /* free memory */

   _ffree (node_index);
   node_index = NULL;

   _ffree (noderef);
   noderef = NULL;
}

int  V7FindFirstNode(PV7RCB p){    /* open node idx/dat and find first match */
   
   if ((idx_fp = share_fopen (p->pszIndexName, read_binary, DENY_WRITE)) == NULL){
       status_line (MSG_TXT(M_UNABLE_TO_OPEN), p->pszIndexName);
       return (V7NL_NODEINDEX_OPENFAILED);   /* no file, no work to do */
   }
   if ((dat_fp = share_fopen (p->pszDataName, read_binary, DENY_WRITE)) == NULL){
      status_line (MSG_TXT(M_UNABLE_TO_OPEN), p->pszDataName);
      return (V7NL_NODEDAT_OPENFAILED);      /* no file, no work to do */
   }

   searchaddr.Zone  = p->pV7Node->ZoneNumber;
   searchaddr.Net   = p->pV7Node->NetNumber;
   searchaddr.Node  = p->pV7Node->NodeNumber;
   searchaddr.Point = p->pV7Node->HubNode;
   
   compare_fn = addr_compare;
   desired = &searchaddr;
   desired_len = p->wLen;

   find_first = 1;

   return V7FindNext(p);
}

int  V7FindFirstSysOp(PV7RCB p){   /* open sysop idx/dat and find first match */  
   char last_name_first[26], midname[26], *c, *p2, *m;
   int namelen;

   if ((idx_fp = share_fopen (p->pszSysOpIndex, read_binary, DENY_WRITE)) == NULL){
       status_line (MSG_TXT(M_UNABLE_TO_OPEN), p->pszSysOpIndex);
       return (V7NL_SYSOPINDEX_OPENFAILED);     /* no file, no work to do */
   }
   if ((dat_fp = share_fopen (p->pszDataName, read_binary, DENY_WRITE)) == NULL){
       status_line (MSG_TXT(M_UNABLE_TO_OPEN), p->pszDataName);
       return (V7NL_NODEDAT_OPENFAILED);       /* no file, no work to do */
   }

   c = midname;                                /* Start of temp name buff   */
   p2 = p->pV7Node->SysopName;                 /* Point to start of name    */
   m = NULL;                                   /* Init pointer to space     */

   *c = *p2++;
   while (*c)                                  /* Go entire length of name  */
   {
      if (*c == ' ')                           /* Look for space            */
         m = c;                                /* Save location             */
      c++;
      *c = *p2++;
   }

   if (m != NULL)                               /* If we have a pointer,     */
   {
      *m++ = '\0';                              /* Terminate the first half  */
      (void) strcpy (last_name_first, m);       /* Now copy the last name    */
      (void) strcat (last_name_first, ", ");    /* Insert a comma and space  */
      (void) strcat (last_name_first, midname); /* Finally copy first half */
   }
   else
      (void) strcpy (last_name_first, midname); /* Use whole name otherwise*/

   (void) fancy_str (last_name_first);         /* Get caps in where needed  */
   namelen = (int) strlen (last_name_first);   /* Calc length now           */
   namelen = min(25,namelen);
   
   strncpy(searchname, last_name_first, namelen);
   searchname[namelen]='\0';
   
   compare_fn = name_compare;
   desired = &searchname;
   desired_len = namelen;
   
   find_first = 1;
   
   return V7FindNext(p);
}

int  V7FindNext(PV7RCB p){         /* find next match */
   int ret;
   PV7NODEL p1,p2;

   pv7rcb = p;
      
   if(find_first){
      beginp = lastp = actp = NULL;
      if(btree()!=0)
         return V7NL_FINDNEXT_END;
   }

   if(find_first)
      actp = beginp;
   else
      actp = actp->next;
      
   if(actp==NULL){               /* if this is the end, free memory */
      for(p1=beginp;p1;p1=p2){
         p2=p1->next;
         free(p1);
      }
      beginp=NULL;
      ret = V7NL_FINDNEXT_END;
   }else{
      ret = V7NL_SUCCESS;
      *(p->pV7Node) = actp->v7n;
   }
   find_first = 0;
   return ret;
}

void V7FindClose(PV7RCB p){        /* close files */
   PV7NODEL p1,p2;

   p=p;   /* Make compiler happy */

   if(beginp){
      for(p1=beginp;p1;p1=p2){
         p2=p1->next;
         free(p1);
      }
   }
   beginp=NULL;
   
   if(idx_fp){
      fclose (idx_fp);
      idx_fp = NULL;
   }

   if(dat_fp){
      fclose (dat_fp);
      dat_fp = NULL;
   }
}

int V7FindNode(PV7RCB p){         /* open files, find node, close files */
   int ret;
   (void) V7Initialize();
   ret=V7FindFirstNode(p);
   (void) V7FindClose(p);
   /* (void) V7Finish(); don't call this and avoid many malloc/free calls */
   return ret;
};


int 
ver7find (ADDRP opus_addr, int have_boss_data)
{
   char idx_fname[PATHLEN], dat_fname[PATHLEN];
   ADDR faddr;
   V7RCB v7rcb;
   V7NODE v7n;
   
   have_boss_data = have_boss_data;     /* make compiler happy */

   (void) memset (&newnodedes, 0, sizeof (struct _newnode));
   
   memset (&v7rcb, 0, sizeof(v7rcb));
   memset (&v7n, 0, sizeof(v7n));
   v7rcb.pV7Node = &v7n;
   
   if (!get_nodelist_name (opus_addr))
       return 0;

   (void) strcpy (idx_fname, net_info);   /* take nodelist path     */
   (void) strcat (idx_fname, nodelist_base); /* add in the file name*/
   (void) strcat (idx_fname, ".NDX");     /* add in the file ext    */
   v7rcb.pszIndexName = idx_fname;
    
   (void) strcpy (dat_fname, net_info);             /* take nodelist path */
   (void) strcat (dat_fname, nodelist_base);        /* add in the filename*/
   (void) strcat (dat_fname, ".DAT");               /* then the extension */
   v7rcb.pszDataName = dat_fname;
   
   v7n.ZoneNumber = opus_addr->Zone;
   v7n.NetNumber  = opus_addr->Net;
   v7n.NodeNumber = opus_addr->Node;
   v7n.HubNode    = opus_addr->Point;
   v7rcb.wLen     = 8;
   
   if(V7FindNode(&v7rcb)==V7NL_SUCCESS){
      set_vars(&v7n, &faddr);
      return 1;
   }else
      return 0;
}

void 
opususer (char *name, ADDRP faddr)
{
   char idx_fname[PATHLEN], dat_fname[PATHLEN];
   V7RCB v7rcb;
   V7NODE v7n;
   int namelen;
    
   faddr->Zone = faddr->Net = faddr->Node = faddr->Point = (unsigned short) -1;
   faddr->Domain = NULL;

   (void) memset (&newnodedes, 0, sizeof (struct _newnode));
   
   memset (&v7rcb, 0, sizeof(v7rcb));
   memset (&v7n, 0, sizeof(v7n));
   v7rcb.pV7Node = &v7n;
   
   if (nodelist_base == NULL)
       get_nodelist_name (&my_addr);

   (void) strcpy (idx_fname, net_info);   /* take nodelist path        */
   (void) strcat (idx_fname, "SYSOP.NDX"); /* add in the file name     */
   v7rcb.pszSysOpIndex = idx_fname;
   
   (void) strcpy (dat_fname, net_info);             /* take nodelist path */
   (void) strcat (dat_fname, nodelist_base);        /* add in the filename*/
   (void) strcat (dat_fname, ".DAT");               /* then the extension */
   v7rcb.pszDataName = dat_fname;

   namelen = min(25,strlen(name));
   strncpy(v7n.SysopName, name, namelen);
   v7n.SysopName[namelen]='\0';
   v7rcb.wLen = namelen;
   
   if(V7FindNode(&v7rcb)==V7NL_SUCCESS){
      set_vars(&v7n, faddr);
   }
}

static void LOCALFUNC
set_vars(PV7NODE p, ADDRP faddr){

   newnodedes.NetNumber  = p->NetNumber;
   newnodedes.NodeNumber = p->NodeNumber;
   newnodedes.HubNode    = p->HubNode;

   newnodedes.BaudRate   = (byte)(p->BaudRate/300);
   newnodedes.ModemType  = p->ModemType;
   newnodedes.NodeFlags  = p->NodeFlags;

   newnodedes.Cost       = p->RealCost;
   newnodedes.RealCost   = p->RealCost; 
   
   (void) strcpy (newnodedes.SystemName, p->SystemName);
   (void) fancy_str (newnodedes.SystemName);

   (void) strcpy (newnodedes.PhoneNumber, p->PhoneNumber);

   (void) strcpy (newnodedes.MiscInfo, p->MiscInfo);
   (void) fancy_str (newnodedes.MiscInfo);

   (void) memcpy (newnodedes.Password, p->Password, 8);

   found_zone  = p->ZoneNumber;
   found_net   = p->NetNumber;

   faddr->Zone = p->ZoneNumber;
   faddr->Net  = p->NetNumber; 
   faddr->Node = p->NodeNumber;
   if (p->NodeFlags & B_point)
      faddr->Point = p->HubNode;
   else
      faddr->Point = 0;
   faddr->Domain = NULL;
}

static int LOCALFUNC
foundmatch(long datpos)
{
   struct _vers7 vers7;
   char a_line[160], a_line2[160];
   char *cp;

   PV7NODEL p;

   if((p = (PV7NODEL) malloc(sizeof(V7NODEL)))==NULL)
      return 1;
   
   if(beginp==NULL)
      beginp=p;
   else
      lastp->next=p;
   lastp=p;
   lastp->next=NULL;

   if (fseek (dat_fp, datpos, SEEK_SET)){           /* point at record  */
      status_line (MSG_TXT(M_NODELIST_SEEK_ERR), pv7rcb->pszDataName);
      return 1;
   }

   if (fread ((char *) &vers7, sizeof (struct _vers7), 1, dat_fp)!=1){
      status_line (MSG_TXT(M_NODELIST_REC_ERR), pv7rcb->pszDataName);
      return 1;
   }

   p->v7n.ulOffset   = datpos;
   
   p->v7n.ZoneNumber = vers7.Zone;
   p->v7n.NetNumber  = vers7.Net;
   p->v7n.NodeNumber = vers7.Node;
   p->v7n.HubNode    = vers7.HubNode;
   p->v7n.BaudRate   = 300*vers7.BaudRate;
   p->v7n.ModemType  = vers7.ModemType;
   p->v7n.NodeFlags  = vers7.NodeFlags;
   p->v7n.RealCost   = vers7.CallCost;
   p->v7n.Cost       = vers7.CallCost;      /* ????? */

   (void) memset (p->v7n.PhoneNumber, 0, 40);
   (void) fread (p->v7n.PhoneNumber, vers7.Phone_len, 1, dat_fp);

   (void) memset (p->v7n.Password, 0, 9);
   (void) fread (p->v7n.Password,    vers7.Password_len, 1, dat_fp);

   (void) memset (a_line,  '\0', 160);
   (void) memset (a_line2, '\0', 160);

   if (fread (a_line2, vers7.pack_len, 1, dat_fp)!=1){
      status_line (MSG_TXT(M_NODELIST_REC_ERR), pv7rcb->pszDataName);
      return 1;
   }

   unpk (a_line2, a_line, vers7.pack_len);

   cp = &a_line[0];
   
   (void) memset (p->v7n.SystemName, 0, 34);
   (void) memcpy (p->v7n.SystemName, cp, min (33, vers7.Bname_len));
   cp += vers7.Bname_len;
   
   (void) memset (p->v7n.SysopName, 0, 26);
   (void) memcpy (p->v7n.SysopName, cp, min (25, vers7.Sname_len));
   cp += vers7.Sname_len;
   
   (void) memset (p->v7n.MiscInfo, 0, 30);
   (void) memcpy (p->v7n.MiscInfo, cp, min (29, vers7.Cname_len));

   return 0;
}

static int LOCALFUNC 
name_compare (void *key, void *desired, int _len)
{
    return (strnicmp ((char *) key, (char *) desired, (unsigned int) min (desired_len, _len)));
}

static int LOCALFUNC 
addr_compare (void *key, void *desired, int _len)
{
    int k, len = desired_len; /* min(desired_len, _len); */

    if(len>=2){
       k = ((ADDRP) key)->Zone - ((ADDRP) desired)->Zone;
       if (k)
          return (k);
    }else return 0;
    
    if(len>=4){
       k = ((ADDRP) key)->Net - ((ADDRP) desired)->Net;
       if (k)
          return (k);
    }else return 0;
    
    if(len>=6){
       k = ((ADDRP) key)->Node - ((ADDRP) desired)->Node;
       if (k)
          return (k);
    }else return 0;

    if (_len == 6)                  /* spezial treatment for key point */
        ((ADDRP) key)->Point = 0;
        
    if(len>=8){
       k = ((ADDRP) key)->Point - ((ADDRP) desired)->Point;
       if (k)
          return (k);
    }else return 0;
    
    return 0;
}

/*
 * General V7 nodelist engine. Used by both the by-node and by-sysop
 * lookups.
 *
 * Thanks to Phil Becker for showing me how nice it looks in assembler.
 * It helped me see how nice it could be in C.
 *
 * (I know, Phil, it's still nicer in assembler!)
 *
 */

static int LOCALFUNC 
btree (void)
{
    /* Get CtlRec */
    if (get7node (idx_fp, 0L, noderef) == NULL)
       return 1;

    return _btree(noderef->ndx.CtlBlk.CtlRoot);     /* start recursion */
}

static int LOCALFUNC 
_btree (long record)
{
    struct _ndx nodeidx;
    int j, j_done=-1, k=0, l, ret=0, count;
    struct _IndxRef far *ip = NULL;
    struct _LeafRef far *lp = NULL;
    char far *tp;
    char *np;

    if (get7node (idx_fp, record * noderef->ndx.CtlBlk.CtlBlkSize, &nodeidx) == NULL)
          return 1;
       
/*
 * Follow the node tree until we either match a key right in the index
 * node, or locate the leaf node which must contain the entry.
 */
    if (nodeidx.ndx.INodeBlk.IndxFirst != -1) /* INodeBlk */
    {
        count = nodeidx.ndx.INodeBlk.IndxCnt; /* count may be 0 */
        if (count == 0){
           /* status_line(">Down 1"); */
           ret=_btree(nodeidx.ndx.INodeBlk.IndxFirst);
           /* status_line(">Up   1"); */
        }else{
           for (j = 0; j < count; j++)         /* check 20 or less */
           {
               ip = &(nodeidx.ndx.INodeBlk.IndxRef[j]);
               tp = (char far *) &nodeidx + ip->IndxOfs;
   
               k = l = ip->IndxLen;               /* binary length in bytes */
                                    /* addresses are stored as ADDR structs */
                                  /* names are stored as "2ndName, 1stName" */
   
               for (np = aline; k > 0; k--)
                   *np++ = *tp++;
   
               k = (*compare_fn) ((void *) aline, desired, l); /* k=key-desired */
   
               if (k > 0)
                  break;
                  
               if (k == 0)
               {
   
/* Key matches in the index node. Since we're just doing lookup, we
 * can assume its pointer is valid. If we were doing updates, that
 * assumption would not work, because leaf nodes update first. So in
 * an update environment, the entire code segment relating to k == 0
 * should not execute, and we should walk the tree all the way down.
 */
                   if(j == 0){
                      /* status_line(">Down 2"); */
                      ret=ret+_btree(nodeidx.ndx.INodeBlk.IndxFirst);
                      /* status_line(">Up   2"); */
                   }else if(j-1!=j_done){
                      /* status_line(">Down 3"); */
                      ret=ret+_btree(nodeidx.ndx.INodeBlk.IndxRef[j-1].IndxPtr);
                      /* status_line(">Up   3"); */
                   }
                   /* foundmatch (nodeidx.ndx.INodeBlk.IndxRef[j].IndxData); */
                   /* status_line(">Down 4"); */
                   ret=ret+_btree(nodeidx.ndx.INodeBlk.IndxRef[j].IndxPtr);
                   /* status_line(">Up   4"); */
                   j_done = j;
               }
           }
           if (k != 0){
              if(j == 0){
                 /* status_line(">Down 5"); */
                 ret=ret+_btree(nodeidx.ndx.INodeBlk.IndxFirst);
                 /* status_line(">Up   5"); */
              }else if(j-1 != j_done){
                 /* status_line(">Down 6"); */
                 ret=ret+_btree(nodeidx.ndx.INodeBlk.IndxRef[j-1].IndxPtr);
                 /* status_line(">Up   6"); */
              }
           }
        }

    }else{ /* LNodeBlk */ 

       count = nodeidx.ndx.LNodeBlk.IndxCnt; /* count may be 0 */
    
       /* Search for a higher key */

       for (j = 0; j < count; j++) /* check 30 or less */
       {
           lp = &(nodeidx.ndx.LNodeBlk.LeafRef[j]);
           tp = (char far *) &nodeidx + lp->KeyOfs;

           k = l = lp->KeyLen;

           for (np = aline; k > 0; k--)
               *np++ = *tp++;

           k = (*compare_fn) ((void *) aline, desired, l);

           if (k > 0)
              break;
           else
           if (k == 0)
              ret=ret+foundmatch (nodeidx.ndx.LNodeBlk.LeafRef[j].KeyVal);
       }
    }
    return ret;
}

static struct _ndx far *LOCALFUNC 
get7node (FILE *stream, long pos, struct _ndx far * ndx)
{
    (void) fseek (stream, pos, SEEK_SET);

    if (fread (ndx, sizeof (struct _ndx), 1, stream) != 1)
    {
        status_line (MSG_TXT(M_NODELIST_READ_ERR));
        return (NULL);
    }
    return (ndx);
}

/* ====================================================================
 * unpack a dense version of a symbol (base 40 polynomial)
 * ====================================================================
 */

static void LOCALFUNC 
unpk (char *instr, char *outp, int count)
{
    /* This table has been modified to minimize searches */
    static char unwrk[] = " EANROSTILCHBDMUGPKYWFVJXZQ-'0123456789";

    struct chars
    {
        unsigned char c1;
        unsigned char c2;
    };

    union
    {
        unsigned short w1;
        struct chars d;
    } u;

    register short i, j;        /* MB int */
    char obuf[4];

    outp[0] = '\0';

    while (count)
    {
        u.d.c1 = *instr++;
        u.d.c2 = *instr++;
        count -= 2;
        for (j = 2; j >= 0; j--)
        {
            i = u.w1 % 40;
            u.w1 /= 40;
            obuf[j] = unwrk[i];
        }
        obuf[3] = '\0';
        (void) strcat (outp, obuf);
    }
}

